//===========================================================================
//
// Analyzer.C
//
// By Poman So
//
//===========================================================================

#ifdef __ActiveCC
#include <string.h>
#endif

#include "Scope.h"
#include "Analyzer.h"
#include "3d-tlm.h"
#include "Const.h"

#include <iostream>
using namespace std;


//---------------------------------------------------------------------------
//
//	Analyzer Class
//
//---------------------------------------------------------------------------
Boolean Analyzer::RefOkay(){
int j;
  for (j=0;j<=nf;j++)
    if (rFr[j]==0 && rFi[j]==0) return False;
  return True;
}//--- Analyzer::RefOkay ---//


Complex Analyzer::S11(int i){
Complex V_inc(rFr[i],rFi[i]),
	V_tot(oFr[i],oFi[i]);
  return (V_tot-V_inc)/V_inc;
}//--- Analyzer::S11 ---//


Complex Analyzer::S21(int i){
Complex V_inc(rFr[i],rFi[i]),
	V_tot(oFr[i],oFi[i]);
  return V_tot/V_inc;
}//--- Analyzer::S21 ---//


float Analyzer::GraphValue(int i){
double gv;
  if (s21)
    switch (mod_statebits){
      case 1: gv = abs(S21(i)); break;
      case 2: gv = -arg(S21(i))*180.0/PI; break;
      case 4: gv = abs(S21(i));
	      gv = (gv>1.0e-100)?20.0*log10(gv):-200.0; break;
    }
  else
    switch (mod_statebits){
      case 1: gv = abs(S11(i)); break;
      case 2: gv = -arg(S11(i))*180.0/PI; break;
      case 4: gv = abs(S11(i));
	      gv = (gv>1.0e-100)?20.0*log10(gv):-200.0; break;
    }
  return gv;
}//--- Analyzer::GraphValue ---//


void Analyzer::GetGraph(){
int i=(mod_statebits==1)?0:(mod_statebits==2)?1:2;
float f1,f2;
  graphbox.GetTitle(title[i]);
  graphbox.GetHLabel(xlabel[i]);
  graphbox.GetVLabel(ylabel[i]);
  graphbox.GetTicks(xticks,yticks[i]);
  graphbox.GetWorld(f1,ymin[i],f2,ymax[i]);
  Parent()->ChildCB(this,DataChangeT);
}//--- Analyzer::GetGraph ---//


void Analyzer::SetGraph(){
int i=(mod_statebits==1)?0:(mod_statebits==2)?1:2;
  graphbox.SetTitle(title[i]);
  graphbox.SetHLabel(xlabel[i]);
  graphbox.SetVLabel(ylabel[i]);
  graphbox.SetTicks(xticks,yticks[i]);
  graphbox.SetWorld(xmin,ymin[i],xmax,ymax[i],True);
}//--- Analyzer::SetGraph ---//


void Analyzer::DrawGraph(){
  if (MeshPtr_ && connected && Mapped()){
  Cursor parentcursor=Parent()->SetCursor(Watch);
    graphbox.Clear();
    graphbox.ToLineGraph();
    if (rFr && rFi && oFr && oFi && RefOkay()){
    int   j;
    float fi,fj,df,ai,aj;
      fi=xmin;  df=(xmax-fi)/nf;  ai=GraphValue(0);
      for (fj=fi+df,j=1;j<=nf;j++,fj+=df){
        aj = GraphValue(j);
        graphbox.Line(fi,ai,fj,aj);
        fi=fj;  ai=aj;
      }
    }
    Parent()->SetCursor(parentcursor);
  }
}//--- Analyzer::DrawGraph ---//

static char *scolist[]={"Scope 1","Scope 2","Scope 3"},
	    *modlist[]={"|S|","<S>","|S| dB"},
	    *freqinGHz="Frequency in GHz",
	    *ylabels[]={"Amplitude","Phase","dB"};
Analyzer::Analyzer(Window &w,char *label,Boolean s_21, Analyzer *A):
DTaskWindow(w,100,50,450,350,label,A),
ref_label(*this,16,titlebar.Y2()+10,100,titlebar.Y2()+20,0,0,"Reference",Small,
	  LeftJ),
ref_box(*this,0,ref_label.Y2()+2,ref_label.X2(),ref_label.Y2()+62,3,scolist,4,
	True),
out_label(*this,16,ref_box.Y2()+5,ref_label.X2(),ref_box.Y2()+15,0,0,"Output",
	Small,LeftJ),
out_box(*this,0,out_label.Y2()+2,ref_label.X2(),out_label.Y2()+62,3,scolist,
	(s_21)?2:1,True),
mod_box(*this,0,out_box.Y2()+5,ref_label.X2(),out_box.Y2()+65,3,modlist,1,
	True),
graphbox(*this,ref_label.X2()+2,titlebar.Y2()+3,Width(),Height()){
int i;
  minW=Width();  maxW=MaxWidth();
  minH=Height(); maxH=MaxHeight();
  ref_statebits=ref_box.StateBits();
  out_statebits=out_box.StateBits();
  mod_statebits=mod_box.StateBits();
  s21 = s_21;
  graphbox.SetBorder(0,0);
  graphbox.SetRubber(AllSideR);
  title[0][0]=title[1][0]=title[2][0]=0;
  for (i=0;i<3;i++){
    strcpy(xlabel[i],freqinGHz);
    strcpy(ylabel[i],ylabels[i]);
  }  
  rFr = rFi = oFr = oFi = 0;
  xmin=0; xmax=10;  xticks=3;
  ymin[0]=0;  ymin[1]=-180; ymin[2]=-60;
  ymax[0]=1;  ymax[1]=180;  ymax[2]=0;
  yticks[0]=yticks[1]=yticks[2]=3;
  connected = False;
  SetGraph();
}//--- Analyzer::Analyzer ---//


Status Analyzer::WindowCB(CBTag tag){
  if (tag==ExposeT){
  int x=ref_box.X2()+1;
    Line(x,0,x,Height(),Black);
  }
  else
    return DTaskWindow::WindowCB(tag);
  return Continue;
}//--- Analyzer::WindowCB ---//


Status Analyzer::ChildCB(Window *w, CBTag tag){
int statebits;
  if (w==&deletebutton && ActivatedWindow()==w && tag==ButtonReleaseT){
    Unmap(Watch);
  }
  else if (w==&ref_box){
    if (tag==ActivateT && ref_statebits!=(statebits=ref_box.StateBits())){
      ref_statebits = statebits;
      Parent()->ChildCB(this,DataChangeT);
      Parent()->ChildCB(this,UserDef0);
    }
  }
  else if (w==&out_box){
    if (tag==ActivateT && out_statebits!=(statebits=out_box.StateBits())){
      out_statebits = statebits;
      Parent()->ChildCB(this,DataChangeT);
      Parent()->ChildCB(this,UserDef0);
    }
  }
  else if (w==&mod_box){
    if (tag==ActivateT && mod_statebits!=(statebits=mod_box.StateBits())){
      mod_statebits = statebits;
      Parent()->ChildCB(this,DataChangeT);
      Parent()->ChildCB(this,UserDef0);
    }
  }
  else if (w==&graphbox){
    switch (tag){
      case DataChangeT: GetGraph();  break;
      case ExposeT:     DrawGraph(); break;
    }
  }
  return Continue;
}//--- Analyzer::ChildCB ---//


fstream &Analyzer::IO(fstream &fs){
int i,MOD_STATEBITS,XLEN,YLEN,TLEN;
  ref_box.IO(fs);
  out_box.IO(fs);
  mod_box.IO(fs);
  ref_statebits=ref_box.StateBits();
  out_statebits=out_box.StateBits();
  mod_statebits=mod_box.StateBits();
  if (!(fs|nf|xticks|xmin|xmax))
    cerr << "Analyzer::IO Error!" << endl;
  for (i=0,MOD_STATEBITS=1;i<3;i++,MOD_STATEBITS<<=1){
    if (mod_statebits!=MOD_STATEBITS){
      if (iostate_==ios::out){
	XLEN=strlen(xlabel[i]);
	YLEN=strlen(ylabel[i]);
	TLEN=strlen(title[i]);
      }
      if (!(fs|ymin[i]|ymax[i]|yticks[i]|XLEN|YLEN|TLEN))
	cerr << "Analyzer::IO Error!" << endl;
      DataIO(fs,xlabel[i],XLEN);
      DataIO(fs,ylabel[i],YLEN);
      DataIO(fs,title[i], TLEN);
      xlabel[i][XLEN]=ylabel[i][YLEN]=title[i][TLEN]=0;
    }
  }
  graphbox.IO(fs);
  if (iostate_==ios::in){
    GetGraph();
    Parent()->ChildCB(this,UserDef0);
  }
  return (Next())?Next()->IO(fs):fs;
}//--- Analyzer::IO ---//


void Analyzer::Init(){
  Parent()->ChildCB(this,UserDef0);
  if (Next()) Next()->Init();
}//--- Analyzer::Init ---//


void Analyzer::Update(Boolean must_refresh){
  if (connected && Mapped() && visible_ && (NumIts_%Step_==0 || must_refresh))
    DrawGraph();
  if (Next()) Next()->Update(must_refresh);
}//--- Analyzer::Update ---//


//===========================================================================
//
// End of Analyzer.C
//
//===========================================================================
