//===========================================================================
//
// Editor.C
//
// By Poman So
//
//===========================================================================

#include "Editor.h"
#include <iostream>
using namespace std;

typedef int (*fcmp)(const void*, const void*);
#include <math.h>
#include <stdlib.h>
//#include <values.h>

Boolean Overlap(float x1, float x2, float X1, float X2){
float t1=Max(x1,X1),t2=Min(x2,X2);
  return (t1<t2)?True:False;
}//--- Overlap ---//


Boolean Enclosed(float *c, float *C){
int i,j,count;
  for (i=count=0,j=3;i<3;i++,j++)
    if ((c[i]>C[i] && c[j]<=C[j]) || (c[i]>=C[i] && c[j]<C[j])) count++;
  return (count>1)?True:False;
}//--- Enclosed ---//

int CompareX(const float *e1, const float *e2){
  return (e1[3]<e2[3])?-1:(e1[3]>e2[3])?1:(e1[0]<e2[0])?-1:(e1[0]>e2[0])?1:0;
}//--- CompareX ---//

int CompareY(const float *e1, const float *e2){
  return (e1[1]<e2[1])?1:(e1[1]>e2[1])?-1:(e1[4]<e2[4])?1:(e1[4]>e2[4])?-1:0;
}//--- CompareY ---//

int CompareZ(const float *e1, const float *e2){
  return (e1[5]<e2[5])?-1:(e1[5]>e2[5])?1:(e1[2]<e2[2])?-1:(e1[2]>e2[2])?1:0;
}//--- CompareZ ---//

int CompareE(Float8 e1, Float8 e2){
   return (Enclosed(e1,e2))?-1:(Enclosed(e2,e1))?1:0;
}//--- CompareE ---//

//---------------------------------------------------------------------------
//
//	Editor Class
//
//---------------------------------------------------------------------------

FillStyle Editor::ReflFill(float refl){
FillStyle fs;
  if (radio2 & 1){
    if (refl==-1.0)
      fs=XHatchF;
    else if (refl==1.0)
      fs=InterLeaveF;
    else
      fs=LTBKSlashF;
  }
  else
    fs=SolidF;
/*
  switch (int(refl)){
    case  0: fs=SolidF;      break;
    case  1: fs=LineF;	     break;
    case  2: fs=LTSlashF;    break;
    case  3: fs=SlashF;      break;
    case  4: fs=BKSlashF;    break;
    case  5: fs=LTBKSlashF;  break;
    case  6: fs=HatchF;      break;
    case  7: fs=XHatchF;     break;
    case  8: fs=InterLeaveF; break;
    case  9: fs=WideDotF;    break;
    default: fs=CloseDotF;
  }
*/
  return fs;
}//--- Editor::ReflFill ---//


Color Editor::ReflColor(float refl){
Color fc;
  if (!Mono() && (radio2 & 1)){
    if (refl==-1.0)
      fc=Red;
    else if (refl==1.0)
      fc=Blue;
    else
      fc=background;
  }
  else
    fc=background;
  return fc;
}//--- Editor::ReflColor ---//


Boolean Editor::Push(int code){
int	i;
float	dz2,dy2,dx2,dz1,dy1,dx1,Tx,Ty,Tz;
  if (NumIts_>0 && radio1!=ANIMATOR){
    return False;
  }
  if (NumIts_==0) Parent()->ChildCB(this,DataChangeT);
  switch (radio1){
    case ELECTRIC:
    case MAGNETIC:
    case GENERAL:
      switch (code){
        case XY_WINDOW:
	  dx1=Dx1();  dy1=Dy1();  dx2=Dx2();  dy2=Dy2();  dz1=dz2=az2;
	  if (dx1==dx2 || dy1==dy2) return False;
	  for (i=0;i<StkPtr && !(Stack[i][6]==WALL && dz1==Stack[i][2] &&
			   Overlap(dx1,dx2,Stack[i][0],Stack[i][3])&&
			   Overlap(dy1,dy2,Stack[i][1],Stack[i][4])); i++);
	  break;
        case XZ_WINDOW:
	  dx1=Dx1();  dz1=Dz1();  dx2=Dx2();  dz2=Dz2();  dy1=dy2=ay2;
	  if (dx1==dx2 || dz1==dz2) return False;
	  for (i=0;i<StkPtr && !(Stack[i][6]==WALL && dy1==Stack[i][1] &&
			   Overlap(dx1,dx2,Stack[i][0],Stack[i][3])&&
			   Overlap(dz1,dz2,Stack[i][2],Stack[i][5])); i++);
	  break;
        case YZ_WINDOW:
	  dy1=Dy1();  dz1=Dz1();  dy2=Dy2();  dz2=Dz2();  dx1=dx2=ax2;
	  if (dy1==dy2 || dz1==dz2) return False;
	  for (i=0;i<StkPtr && !(Stack[i][6]==WALL && dx1==Stack[i][0] &&
			   Overlap(dy1,dy2,Stack[i][1],Stack[i][4])&&
			   Overlap(dz1,dz2,Stack[i][2],Stack[i][5])); i++);
	  break;
      }
      break;
    case ER_N_UR:
    case GENERATOR:
    case SCOPE:
    case ANIMATOR:
      if (radio1!=ER_N_UR){
        Tx=0.25*Dx; Ty=0.25*Dy; Tz=0.25*Dz;
      }
      else
        Tx=Ty=Tz=0.0;
      switch (code){
        case XY_WINDOW:
	  dx1=Dx1(); dy1=Dy1(); dx2=Dx2(); dy2=Dy2();
	  dz1=Min(az0,az2); dz2=Max(az0,az2); break;
        case XZ_WINDOW:
	  dx1=Dx1(); dz1=Dz1(); dx2=Dx2(); dz2=Dz2();
	  dy1=Min(ay0,ay2); dy2=Max(ay0,ay2); break;
        case YZ_WINDOW:
	  dy1=Dy1(); dz1=Dz1(); dy2=Dy2(); dz2=Dz2();
	  dx1=Min(ax0,ax2); dx2=Max(ax0,ax2); break;
      }
      if (dx1==dx2 || dy1==dy2 || dz1==dz2) return False;
      dx1+=Tx;  dx2-=Tx;  dy1+=Ty;  dy2-=Ty;  dz1+=Tz;  dz2-=Tz;
      if (radio1==SCOPE){
        dx2=dx1+Dx-2*Tx; dy2=dy1+Dy-2*Ty; dz2=dz1+Dz-2*Tz;
      }
      else if (radio1==ANIMATOR){
	if (!((dx2-dx1<Dx) || (dy2-dy1<Dy) || (dz2-dz1<Dz)))
          switch (code){
	    case XY_WINDOW: dz1=Min(az0,az2)+Tz;  dz2=dz1+Dz-2*Tz;  break;
	    case XZ_WINDOW: dy1=Min(ay0,ay2)+Ty;  dy2=dy1+Dy-2*Ty;  break;
	    case YZ_WINDOW: dx1=Min(ax0,ax2)+Tx;  dx2=dx1+Dx-2*Tx;  break;
          }
        if (NumIts_>0 && !((dx2<(Xmax_/Cf_+0.5*Dx)) &&
	    (dy2<(Ymax_/Cf_+0.5*Dy)) && (dz2<(Zmax_/Cf_+0.5*Dz))))
	  return False;
      }
      switch(radio1){
	case ER_N_UR:
	  for (i=0;i<StkPtr && !(Stack[i][6]==ER_N_UR &&
				  Overlap(dx1,dx2,Stack[i][0],Stack[i][3]) &&
				  Overlap(dy1,dy2,Stack[i][1],Stack[i][4]) &&
				  Overlap(dz1,dz2,Stack[i][2],Stack[i][5]) &&
				  !Enclosed(&dx1,Stack[i]) &&
				  !Enclosed(Stack[i],&dx1)
				); i++);
	  break;
	case GENERATOR:
	case SCOPE:
	  for (i=0;i<StkPtr && !(Stack[i][6]==radio1 &&
			     Overlap(dx1,dx2,Stack[i][0],Stack[i][3]) &&
			     Overlap(dy1,dy2,Stack[i][1],Stack[i][4]) &&
			     Overlap(dz1,dz2,Stack[i][2],Stack[i][5])); i++);
	  break;
	case ANIMATOR:
	  for (i=0;i<StkPtr && !(Stack[i][6]==ANIMATOR &&
			     dx1==Stack[i][0] && dx2==Stack[i][3] &&
			     dy1==Stack[i][1] && dy2==Stack[i][4] &&
                             dz1==Stack[i][2] && dz2==Stack[i][5]); i++);
	  break;
      }
      break;
  }
  if (i==StkPtr){
    if (StkPtr<StkSize){
      Stack[i][0]=dx1;  Stack[i][1]=dy1;  Stack[i][2]=dz1;
      Stack[i][3]=dx2;  Stack[i][4]=dy2;  Stack[i][5]=dz2;
      return True;
    }
    else
      Message("Stack is full!");
  }
  else if (i<0)
    Message("Element is to be implemented!");
  else if (dx1==Stack[i][0] && dy1==Stack[i][1] && dz1==Stack[i][2] &&
	   dx2==Stack[i][3] && dy2==Stack[i][4] && dz2==Stack[i][5]){
  int j;
    if (i != --StkPtr)
      for (j=0;j<9;j++) Stack[i][j]=Stack[StkPtr][j];
    Refresh();
    if (radio1==ANIMATOR && NumIts_>0)
      Parent()->ChildCB(this,UserDef1);
  }
  else
    Message("The element that has been just entered overlaps illegally with \
at least one of the elements that are already in the stack!");
  return False;
}//--- Editor::Push ---//


void Editor::PromptData(){
static char *title="Editor: Data";

static char *prompt1[]={"Impulse Reflection Coefficient%g",""};
static char *prompt2[]={"Relative Permittivity%g",
			"Relative Permeability%g",""};
int  x1=Width()/4,y1=Height()/4,x2=x1*3,y2=y1*3;
char *prompt3="Pick a generator.",
     *generator1="Generator 1",
     *generator2="Generator 2",
     *generator3="Generator 3",
     *generator;
char *prompt4="Pick a scope.",
     *prompt5="Pick an animator.";
static char *sco[3]={"Scope 1","Scope 2","Scope 3"};
static char *ani[3]={"Animator 1","Animator 2","Animator 3"};

  switch (radio1){
    case ELECTRIC:  Stack[StkPtr][6]=WALL;  Stack[StkPtr++][7]=-1.0; break;
    case MAGNETIC:  Stack[StkPtr][6]=WALL;  Stack[StkPtr++][7]= 1.0; break;
    case GENERAL:   if (GetData(*Root(),0,x1,y1,x2,y2,title,prompt1,
		        &Refl_in)==Okay){
	              Stack[StkPtr][6]=WALL;  Stack[StkPtr++][7]=Refl_in;
		    }
		    break;
    case ER_N_UR:   if (GetData(*Root(),0,x1,y1,x2,y2,title,prompt2,
			&Er_in,&Ur_in)==Okay){
		      Stack[StkPtr  ][6]=ER_N_UR;
		      Stack[StkPtr  ][7]=Er_in;
		      Stack[StkPtr++][8]=Ur_in;
		    }
		    break;
    case GENERATOR:    generator=Message(*Root(),prompt3,generator1,generator2,generator3);
		    Stack[StkPtr  ][6]=GENERATOR;
		    Stack[StkPtr++][7]=(generator==generator1)?0:
				       (generator==generator2)?1:2;
	            break;
    case SCOPE:  GetEquipment(SCOPE,prompt4,sco);
		    break;
    case ANIMATOR:  GetEquipment(ANIMATOR,prompt5,ani);
		    if (NumIts_>0)
		      Parent()->ChildCB(this,UserDef1);
		    break;
  }
  Refresh();
}//--- Editor::PromptData ---//


void Editor::GetEquipment(int eq_num, char *prompt, char **eq_list){
int i;
char *eq[3],*equipment;
  for (i=0;i<3;i++) eq[i]=eq_list[i];
  for (i=0;i<StkPtr;i++)
    if (Stack[i][6]==eq_num) eq[int(Stack[i][7])]=0;
  if (eq[2])
    if (eq[1])
      if (eq[0])
	equipment=Message(*Root(),prompt,eq_list[0],eq_list[1],eq_list[2]);
      else
	equipment=Message(*Root(),prompt,eq_list[1],eq_list[2]);
      else
	if (eq[0])
	  equipment=Message(*Root(),prompt,eq_list[0],eq_list[2]);
	else
	  equipment=eq_list[2];
  else
    if (eq[1])
      if (eq[0])
	equipment=Message(*Root(),prompt,eq_list[0],eq_list[1]);
      else
	equipment=eq_list[1];
    else
      if (eq[0])
	equipment=eq_list[0];
      else
	equipment=0;
  if (equipment){
    Stack[StkPtr  ][6]=eq_num;
    Stack[StkPtr++][7]=(equipment==eq_list[0])?0:(equipment==eq_list[1])?1:2;
  }
}//--- Editor::GetEquipment ---//


void Editor::Refresh(Boolean clearbk){
Cursor parentcursor=Parent()->SetCursor(Watch);
  if (clearbk){
    XY.Clear();  XZ.Clear();  YZ.Clear();  XYZ.Clear();
  }
  XY_ExposeCB();
  XZ_ExposeCB();
  YZ_ExposeCB();
  XYZ_ExposeCB();
  Parent()->SetCursor(parentcursor);
}//--- Editor::Refresh ---//


void Editor::RadioBox1_ExposeCB(){
int i,t,ty,dy,x1,y1,x2,y2;
Color fc;
  radiobox1.LabelXYs(0,x1,t,t,y1);
  radiobox1.LabelXYs(1,t,y2,t,ty);
  dy = 13;//(y2-y1)/2;
  x2 = radiobox1.Width()-10;

  y1+=4;   y2=y1+dy;
  fc = ReflColor(-1);
  radiobox1.SetFillStyle(ReflFill(-1),fc);
  radiobox1.FillRectangle(x1,y1,x2,y2,fc,Black);

  y1=ty+4; y2=y1+dy;
  fc = ReflColor(1);
  radiobox1.SetFillStyle(ReflFill(1),fc);
  radiobox1.FillRectangle(x1,y1,x2,y2,fc,Black);

  radiobox1.LabelXYs(2,t,t,t,y1);
  y1+=4;   y2=y1+dy;
  fc = ReflColor(0);
  radiobox1.SetFillStyle(ReflFill(0),fc);
  radiobox1.FillRectangle(x1,y1,x2,y2,fc,Black);

  SetTextJustify(CenterJ,TopJ);
  for (i=3;i<7;i++){
    radiobox1.LabelXYs(i,t,t,t,y1);
    y1+=4;   y2=y1+dy;
    radiobox1.SetFillStyle(SolidF,background);
    radiobox1.FillRectangle(x1,y1,x2,y2,background,Black);
    if (i>3 && i<7)
      radiobox1.OutTextXY((x1+x2)/2,y1,(i==4)?"G":(i==5)?"S":"A",Black);
  }
}//--- Editor::RadioBox1_ExposeCB ---//


void Editor::DataButton_CB(){
static char *prompt[]={"Relative Permittivity%g",
		"Relative Permeability%g",
		"Dimension Conversion Factor%g",
		"Step Size - X%g",
		"         - Y%g",
		"         - Z%g",
		"Pixels Per Step - X%d",
		"               - Y%d",
		"               - Z%d",
		""
	       };
int x1,y1,x2,y2,dpx,dpy,dpz,idpx,idpy,idpz;
float cf,er,ur,iDx,iDy,iDz;
Boolean active[9];
  active[0]=active[1]=(NumIts_>0)?False:True;
  active[2]=active[3]=active[4]=active[5]=(StkPtr>0)?False:True;
  active[6]=active[7]=active[8]=True;
  x1=X1()+20;  y1=Y1()+20;  x2=x1+360;  y2=y1+400;
  dpx=2*(int(Dx/dx+0.5)/2);
  dpy=2*(int(Dy/dy+0.5)/2);
  dpz=2*(int(Dz/dz+0.5)/2);
  idpx=dpx;  idpy=dpy;  idpz=dpz;
  
  cf=Cf_;  er=Er_;  ur=Ur_;  iDx=Dx;  iDy=Dy;  iDz=Dz;
  if (GetActiveData(*Root(),0,x1,y1,x2,y2,"Mesh Data",prompt,active,
	      &er,&ur,&cf,&iDx,&iDy,&iDz,&idpx,&idpy,&idpz)==Okay){
    if (er!=Er_ || ur!=Ur_ || cf!=Cf_ || iDx!=Dx || iDy!=Dy || iDz!=Dz ||
	idpx!=dpx || idpy!=dpy || idpz!=dpz){
      if (NumIts_==0){
	Er_=Max(1.0f,er);
	Ur_=Max(1.0f,ur);
      }
      if (StkPtr==0){
	Cf_=(cf>1.0e-10)?pow(10,int(log10(cf*((cf>1.0)?2.0:0.5)))):Cf_;
	if (iDx>0.0) Dx=iDx;
	if (iDy>0.0) Dy=iDy;
	if (iDz>0.0) Dz=iDz;
      }
      idpx = 2*(idpx/2);
      idpy = 2*(idpy/2);
      idpz = 2*(idpz/2);
      dx = (idpx<4)?Dx/4:Dx/idpx;
      dy = (idpy<4)?Dy/4:Dy/idpy;
      dz = (idpz<4)?Dz/4:Dz/idpz;
      SetDrawParameters();
      Parent()->ChildCB(this,DataChangeT);
    }
  }
}//--- Editor::DataButton_CB ---//


void Editor::WXYZ_ResizeCB(){
double w=XY.Width() +XYZ.Width()+1,W=WXYZ.Width()-1,
       h=XY.Height()+XZ.Height()+1,H=WXYZ.Height()-1;
int    x2=int(Min(MW.X2()*W/w,W)),y2=int(Min(MW.Y2()*H/h,H)),
       x1=x2-MW_w,y1=y2-MW_h,d;
Boolean wxyzMapped=WXYZ.Mapped();
  if (wxyzMapped) WXYZ.Unmap();
  if (x1<50)  {d=50-x1;        x1+=d; x2+=d;}
  if (x2>W-50){d=int(x2-W+50); x1-=d; x2-=d;}
  if (y1<50)  {d=50-y1;        y1+=d; y2+=d;}
  if (y2>H-50){d=int(y2-H+50); y1-=d; y2-=d;}
  MW.SetX1Y1X2Y2(x1,y1,x2,y2);
  x1=0;y1=0;
  XY.SetX1Y1X2Y2(x1,y1,x2,y2);
  y1=y2+1;  y2=int(H); 
  XZ.SetX1Y1X2Y2(x1,y1,x2,y2);
  x1=x2+1;  x2=int(W);
  YZ.SetX1Y1X2Y2(x1,y1,x2,y2);
  y2=y1-1;  y1=XY.Y1();
  XYZ.SetX1Y1X2Y2(x1,y1,x2,y2);
  if (wxyzMapped) WXYZ.Map();
}//--- Editor::WXYZ_ResizeCB ---//


void Editor::XY_ActivateCB(){
  MW.Unmap();
  XZ.SetWriteMode(XorMode);
  XZ.SetLineStyle(DottedL);
  YZ.SetWriteMode(XorMode);
  YZ.SetLineStyle(DottedL);
  ax0=ax1; ay0=ay1;
  XY_ButtonPressCB();
}//--- Editor::XY_ActivateCB ---//


void Editor::XY_ButtonPressCB(){
  XY.GetData(ax1,ay1,ax2,ay2);
  switch (radio1){
    case ELECTRIC:
    case MAGNETIC:
    case GENERAL:
      XZ.Line(ax1,az2,ax2,az2,background);
      YZ.Line(ay1,az2,ay2,az2,background);
      break;
    case ER_N_UR:
    case GENERATOR:
    case SCOPE:
    case ANIMATOR:
      XZ.Rectangle(ax1,az0,ax2,az2,background);
      YZ.Rectangle(ay1,az0,ay2,az2,background);
      break;
  }
}//--- Editor::XY_ButtonPressCB ---//


void Editor::XY_ButtonReleaseCB(){
  MW.Map();
  XY_ButtonPressCB();
  if (Push(XY_WINDOW)) PromptData();
}//--- Editor::XY_ButtonReleaseCB ---//


void Editor::XY_ExposeCB(){
int   i,code1,code2;
float x1,y1,x2,y2,T,Dx,Dy;
FillStyle fs;
Color	  fc;
const char *label;
Cursor oldcursor;
  if (Delay) oldcursor=XY.SetCursor(Hand);
  XY.SetWriteMode(SetMode);
  XY.SetLineStyle(SolidL);
  XY.DrawGrid();
  XY.GetDrawXY(T,T,Dx,Dy);
  Dx*=0.2; Dy*=0.2;
  qsort(Stack,StkPtr,sizeof(Stack[0]),fcmp(CompareZ));
  for (i=0;i<StkPtr;i++){
    code1 = int(Stack[i][6]);
    if (code1!=WALL && code1!=ER_N_UR && !(radio2&2)) continue;
    x1=Stack[i][0]; y1=Stack[i][1]; x2=Stack[i][3]; y2=Stack[i][4];
    if ((radio2&1) || (code1!=WALL && code1!=ER_N_UR)){
      switch (code1){
	case WALL:     fs=ReflFill(Stack[i][7]);
		       fc=ReflColor(Stack[i][7]);
		       break;
	case ER_N_UR:
	case GENERATOR:
	case SCOPE:
	case ANIMATOR: fs=SolidF;
		       fc=background;
		       break;
      }
      XY.SetFillStyle(fs,fc);
      XY.FillRectangle(x1,y1,x2,y2,fc,Black);
      if (code1==GENERATOR || code1==SCOPE || code1==ANIMATOR){
	code2=int(Stack[i][7]);
	switch (code1){
	  case GENERATOR:   label=(code2==0)?"G1":(code2==1)?"G2":"G3"; break;
	  case SCOPE: label=(code2==0)?"S1":(code2==1)?"S2":"S3"; break;
	  case ANIMATOR: label=(code2==0)?"A1":(code2==1)?"A2":"A3"; break;
        }
        XY.OutTextXY((x1+x2)/2-Dx,(y1+y2)/2+Dy,label);
      }
    }
    XY.Rectangle(x1,y1,x2,y2,Black);
    if (Delay && i<(StkPtr-1) && x1<x2 && y1<y2) Wait(ButtonRelease);
  }
  if (Delay) XY.SetCursor(oldcursor);
  Flush();
}//--- Editor::XY_ExposeCB ---//


void Editor::XY_SelectCB(){
  if (radio2&4){
    Delay=True;
    Wait(ButtonRelease);
    XY.Clear();
    XY_ExposeCB();
    Delay=False;
  }
}//--- Editor::XY_SelectCB ---//


void Editor::XY_ScrollCB(int tag){
float t;
  switch (tag){
    case HDecT:
    case HIncT:
    case HPageDecT:
    case HPageIncT:
    case HScrollT:
	   XY.GetDrawW(t,yo,dx,dy);
	   if (t!=xo) XZ.SetDrawW(xo=t,zo,dx,dz);
	   break;
    case VDecT:
    case VIncT:
    case VPageDecT:
    case VPageIncT:
    case VScrollT:
	   XY.GetDrawW(xo,t,dx,dy);
	   if (t!=yo) YZ.SetDrawW(yo=t,zo,dy,dz);
    	   break;
  }
}//--- Editor::XY_ScrollCB ---//


void Editor::XZ_ActivateCB(){
  XY.SetWriteMode(XorMode);
  XY.SetLineStyle(DottedL);
  YZ.SetWriteMode(XorMode);
  YZ.SetLineStyle(DottedL);
  ax0=ax1; az0=az1;
  XZ_ButtonPressCB();
}//--- Editor::XZ_ActivateCB ---//


void Editor::XZ_ButtonPressCB(){
  XZ.GetData(ax1,az1,ax2,az2);
  switch (radio1){
    case ELECTRIC:
    case MAGNETIC:
    case GENERAL:
      XY.Line(ax1,ay2,ax2,ay2,background);
      YZ.Line(ay2,az1,ay2,az2,background);
      break;
    case ER_N_UR:
    case GENERATOR:
    case SCOPE:
    case ANIMATOR:
      XY.Rectangle(ax1,ay0,ax2,ay2,background);
      YZ.Rectangle(ay0,az1,ay2,az2,background);
      break;
  }
}//--- Editor::XZ_ButtonPressCB ---//


void Editor::XZ_ButtonReleaseCB(){
  XZ_ButtonPressCB();
  if (Push(XZ_WINDOW)) PromptData();
}//--- Editor::XZ_ButtonReleaseCB ---//


void Editor::XZ_ExposeCB(){
int   i,code1,code2;
float x1,z1,x2,z2,T,Dx,Dz;
FillStyle fs;
Color	  fc;
const char  *label;
Cursor oldcursor;
  if (Delay) oldcursor=XZ.SetCursor(Hand);
  XZ.SetWriteMode(SetMode);
  XZ.SetLineStyle(SolidL);
  XZ.DrawGrid();
  XZ.GetDrawXY(T,T,Dx,Dz);
  Dx*=0.2; Dz*=0.2;
  qsort(Stack,StkPtr,sizeof(Stack[0]),fcmp(CompareY));
  for (i=0;i<StkPtr;i++){
    code1 = int(Stack[i][6]);
    if (code1!=WALL && code1!=ER_N_UR && !(radio2&2)) continue;
    x1=Stack[i][0]; z1=Stack[i][2]; x2=Stack[i][3]; z2=Stack[i][5];
    if ((radio2&1) || (code1!=WALL && code1!=ER_N_UR)){
      switch (code1){
	case WALL:     fs=ReflFill(Stack[i][7]);
		       fc=ReflColor(Stack[i][7]);
		       break;
	case ER_N_UR:
	case GENERATOR:
	case SCOPE:
	case ANIMATOR: fs=SolidF;
		       fc=background;
		       break;
      }
      XZ.SetFillStyle(fs,fc);
      XZ.FillRectangle(x1,z1,x2,z2,fc,Black);
      if (code1==GENERATOR || code1==SCOPE || code1==ANIMATOR){
	code2=int(Stack[i][7]);
	switch (code1){
	  case GENERATOR:   label=(code2==0)?"G1":(code2==1)?"G2":"G3"; break;
	  case SCOPE: label=(code2==0)?"S1":(code2==1)?"S2":"S3"; break;
	  case ANIMATOR: label=(code2==0)?"A1":(code2==1)?"A2":"A3"; break;
        }
        XZ.OutTextXY((x1+x2)/2-Dx,(z1+z2)/2+Dz,label);
      }
    }
    XZ.Rectangle(x1,z1,x2,z2,Black);
    if (Delay && i<(StkPtr-1) && x1<x2 && z1<z2) Wait(ButtonRelease);
  }
  if (Delay) XZ.SetCursor(oldcursor);
  Flush();
}//--- Editor::XZ_ExposeCB ---//


void Editor::XZ_SelectCB(){
  if (radio2&4){
    Delay=True;
    Wait(ButtonRelease);
    XZ.Clear();
    XZ_ExposeCB();
    Delay=False;
  }
}//--- Editor::XZ_SelectCB ---//


void Editor::XZ_ScrollCB(int tag){
float t;
  switch (tag){
    case HDecT:
    case HIncT:
    case HPageDecT:
    case HPageIncT:
    case HScrollT:
	   XZ.GetDrawW(t,zo,dx,dz);
	   if (t!=xo) XY.SetDrawW(xo=t,yo,dx,dy);
	   break;
    case VDecT:
    case VIncT:
    case VPageDecT:
    case VPageIncT:
    case VScrollT:
	   XZ.GetDrawW(xo,t,dx,dz);
	   if (t!=zo) YZ.SetDrawW(yo,zo=t,dy,dz);
    	   break;
  }
}//--- Editor::XZ_ScrollCB ---//


void Editor::YZ_ActivateCB(){
  XY.SetWriteMode(XorMode);
  XY.SetLineStyle(DottedL);
  XZ.SetWriteMode(XorMode);
  XZ.SetLineStyle(DottedL);
  ay0=ay1; az0=az1;
  YZ_ButtonPressCB();
}//--- Editor::YZ_ActivateCB ---//


void Editor::YZ_ButtonPressCB(){
  YZ.GetData(ay1,az1,ay2,az2);
  switch (radio1){
    case ELECTRIC:
    case MAGNETIC:
    case GENERAL:
      XY.Line(ax2,ay1,ax2,ay2,background);
      XZ.Line(ax2,az1,ax2,az2,background);
      break;
    case ER_N_UR:
    case GENERATOR:
    case SCOPE:
    case ANIMATOR:
      XY.Rectangle(ax0,ay1,ax2,ay2,background);
      XZ.Rectangle(ax0,az1,ax2,az2,background);
      break;
  }
}//--- Editor::YZ_ButtonPressCB ---//


void Editor::YZ_ButtonReleaseCB(){
  YZ_ButtonPressCB();
  if (Push(YZ_WINDOW)) PromptData();
}//--- Editor::YZ_ButtonReleaseCB ---//


void Editor::YZ_ExposeCB(){
int   i,code1,code2;
float y1,z1,y2,z2,T,Dy,Dz;
FillStyle fs;
Color     fc;
const char *label;
Cursor oldcursor;
  if (Delay) oldcursor=YZ.SetCursor(Hand);
  YZ.SetWriteMode(SetMode);
  YZ.SetLineStyle(SolidL);
  YZ.DrawGrid();
  YZ.GetDrawXY(T,T,Dy,Dz);
  Dy*=0.2; Dz*=0.2;
  qsort(Stack,StkPtr,sizeof(Stack[0]),fcmp(CompareX));
  for (i=0;i<StkPtr;i++){
    code1 = int(Stack[i][6]);
    if (code1!=WALL && code1!=ER_N_UR && !(radio2&2)) continue;
    y1=Stack[i][1]; z1=Stack[i][2]; y2=Stack[i][4]; z2=Stack[i][5];
    if ((radio2&1) || (code1!=WALL && code1!=ER_N_UR)){
      switch (code1){
	case WALL:     fs=ReflFill(Stack[i][7]);
		       fc=ReflColor(Stack[i][7]);
		       break;
	case ER_N_UR:
	case GENERATOR:
	case SCOPE:
	case ANIMATOR: fs=SolidF;
		       fc=background;
		       break;
      }
      YZ.SetFillStyle(fs,fc);
      YZ.FillRectangle(y1,z1,y2,z2,fc,Black);
      if (code1==GENERATOR || code1==SCOPE || code1==ANIMATOR){
	code2=int(Stack[i][7]);
	switch (code1){
	  case GENERATOR:   label=(code2==0)?"G1":(code2==1)?"G2":"G3"; break;
	  case SCOPE: label=(code2==0)?"S1":(code2==1)?"S2":"S3"; break;
	  case ANIMATOR: label=(code2==0)?"A1":(code2==1)?"A2":"A3"; break;
	}
	YZ.OutTextXY((y1+y2)/2-Dy,(z1+z2)/2+Dz,label);
      }
    }
    YZ.Rectangle(y1,z1,y2,z2,Black);
    if (Delay && i<(StkPtr-1) && y1<y2 && z1<z2) Wait(ButtonRelease);
  }
  if (Delay) YZ.SetCursor(oldcursor);
  Flush();
}//--- Editor::YZ_ExposeCB ---//


void Editor::YZ_SelectCB(){
  if (radio2&4){
    Delay=True;
    Wait(ButtonRelease);
    YZ.Clear();
    YZ_ExposeCB();
    Delay=False;
  }
}//--- Editor::YZ_SelectCB ---//


void Editor::YZ_ScrollCB(int tag){
float t;
  switch (tag){
    case HDecT:
    case HIncT:
    case HPageDecT:
    case HPageIncT:
    case HScrollT:
	   YZ.GetDrawW(t,zo,dy,dz);
	   if (t!=yo) XY.SetDrawW(xo,yo=t,dx,dy);
	   break;
    case VDecT:
    case VIncT:
    case VPageDecT:
    case VPageIncT:
    case VScrollT:
	   YZ.GetDrawW(yo,t,dy,dz);
	   if (t!=zo) XZ.SetDrawW(xo,zo=t,dx,dz);
    	   break;
  }
}//--- Editor::YZ_ScrollCB ---//


void Editor::XYZ_ExposeCB(){
int   i;
float p[15],x1,y1,z1,x2,y2,z2;
Color dc;
  XYZ.SetLineStyle(SolidL);
  XYZ.SetWriteMode(SetMode);
  for (i=0;i<StkPtr;i++){
    if (Stack[i][6]!=WALL && Stack[i][6]!=ER_N_UR && !(radio2&2)) continue;
    x1=Stack[i][0]; y1=Stack[i][1]; z1=Stack[i][2];
    x2=Stack[i][3]; y2=Stack[i][4]; z2=Stack[i][5];
    if (z1<z2 && y1<y2 && x1<x2)
      XYZ.Box(x1,y1,z1,x2,y2,z2);
    else{
      if (z1==z2){
	p[0] =x1; p[1] =y1; p[2] =z1;
	p[3] =x1; p[4] =y2; p[5] =z1;
	p[6] =x2; p[7] =y2; p[8] =z1;
	p[9] =x2; p[10]=y1; p[11]=z1;
	p[12]=x1; p[13]=y1; p[14]=z1;
      }
      else if (y1==y2){
	p[0] =x1; p[1] =y1; p[2] =z1;
	p[3] =x1; p[4] =y1; p[5] =z2;
	p[6] =x2; p[7] =y1; p[8] =z2;
	p[9] =x2; p[10]=y1; p[11]=z1;
	p[12]=x1; p[13]=y1; p[14]=z1;
      } 
      else{
	p[0] =x1; p[1] =y1; p[2] =z1;
	p[3] =x1; p[4] =y1; p[5] =z2;
	p[6] =x1; p[7] =y2; p[8] =z2;
	p[9] =x1; p[10]=y2; p[11]=z1;
	p[12]=x1; p[13]=y1; p[14]=z1;
      }
      if (radio2 & 1){
	if (int(Stack[i][6])==WALL){
	  dc = ReflColor(Stack[i][7]);
 	  if (dc==background) dc=Black;
	}
	else
	  dc = background;
	XYZ.SetLineStyle(DottedL,Normal);
	XYZ.Line(p[0],p[1],p[2],p[6],p[7],p[8],dc);
	XYZ.Line(p[3],p[4],p[5],p[9],p[10],p[11],dc);
	XYZ.SetLineStyle(SolidL);
      }
      XYZ.Polygon(5,p);
    }
  }
  Flush();
}//--- Editor::XYZ_ExposeCB ---//


void Editor::MW_ExposeCB(){
int x1=2,y1=x1,x2=MW.Width()-x1-1,y2=MW.Height()-y1-1;
  MW.Rectangle(x1,y1,x2,y2);
  MW.Rectangle(x1+3,y1+3,x2-3,y2-3);
}//--- Editor::MW_ExposeCB ---//


void Editor::MW_MoveCB(){
int x1=0,y1=0,x2=MW.X2(),y2=MW.Y2();
Cursor oldcursor=SetCursor(Watch);
Boolean wxyzMapped=WXYZ.Mapped();
  if (wxyzMapped) WXYZ.Unmap();
  XY.Resize(x2,y2);
  y1=y2+1;  y2=WXYZ.Height()-1; 
  XZ.MoveResize(x1,y1,x2,y2);
  x1=x2+1;  x2=WXYZ.Width()-1;
  YZ.MoveResize(x1,y1,x2,y2);
  y2=y1-1;  y1=XY.Y1();
  XYZ.MoveResize(x1,y1,x2,y2);
  if (wxyzMapped) WXYZ.Map();
  SetCursor(oldcursor);
}//--- Editor::MW_MoveCB ---//


char *element[]={"Electric","Magnetic","General","Er & Ur","Generator","Scope",
		 "Animator"};
char *style[]={"Opaque","Tools","Click N Show"};

Editor::Editor(Window &w, int stkptr, int stksize, StkElm *stack):
DTaskWindow(w,20,50,w.Width()-20,w.Height()-20,"Editor",0),
radiobox1(*this,-1,titlebar.Y2()+2,110,titlebar.Y2()+280,7,element,0,True),
radiobox2(*this,-1,radiobox1.Y2()+2,radiobox1.X2(),radiobox1.Y2()+80,3,style,
	  2,False),
databutton(*this,PB_Margin,Height()-PB_Y1fromB,radiobox2.X2()-PB_Margin,
	   Height()-PB_Margin,"Mesh Data"),
WXYZ(*this,radiobox1.X2()+1,radiobox1.Y1()-1,Width(),Height()),
XY(WXYZ,0,0,WXYZ.Width()/2,WXYZ.Height()/2,0,0,1,1),
XZ(WXYZ,0,XY.Y2()+1,XY.X2(),WXYZ.Height()-1,0,0,1,1),
YZ(WXYZ,XY.X2()+1,XZ.Y1(),WXYZ.Width()-1,XZ.Y2(),0,0,1,1),
XYZ(WXYZ,YZ.X1(),XY.Y1(),YZ.X2(),XY.Y2(),100,1,0,0,0,30,0,-30,True),
MW(WXYZ,XY.X2()-SB_Width,XY.Y2()-SB_Width,XY.X2(),XY.Y2()){
  Delay=False;
  minH=radiobox1.Height()*3/2;
  minW=minH*4/3;
  maxW=MaxWidth();
  maxH=MaxHeight();
  databutton.SetRubber(LeftBottomR);
  WXYZ.SetBorder(0,0);
  WXYZ.SetRubber(AllSideR);
  MW_w = MW_h = SB_Width;
  MW.SetMovement(DefaultDirection,50,50,50,50);
  radio1=radiobox1.StateBits();
  radio2=radiobox2.StateBits();
  xo=yo=zo=0;
  dx=dy=dz=0.05;
  Dx=Dy=Dz=20*dx;
  Xmin=Ymin=Zmin=0;
  Xmax=Ymax=Zmax=10;
  rxo=ryo=rzo=0;
  Er_in=Ur_in=1.0;
  Refl_in=0.0;
  SetStack(stkptr,stksize,stack);
  SetMeshData(1e-3,1,1,10,10,10);
  MW.Map();
  WXYZ.Map();
}//--- Editor::Editor ---//


Status Editor::WindowCB(CBTag tag){
    return DTaskWindow::WindowCB(tag);
}//--- Editor::WindowCB ---//


Status Editor::ChildCB(Window *w, CBTag tag){
    if (w==&deletebutton && ActivatedWindow()==w && tag==ButtonReleaseT){
        Unmap(Watch);
    }
    else if (w==&MW)
        switch(tag){
            case MoveT:   MW_MoveCB();   break;
            case ExposeT: MW_ExposeCB(); break;
        }
    else if (w==&WXYZ)
        switch(tag){
            case ResizeT: WXYZ_ResizeCB(); break;
        }
    else if (w==&XY)
        switch(tag){
            case SelectT:	   XY_SelectCB();	 break;
            case ExposeT:        XY_ExposeCB();        break;
            case ActivateT:      XY_ActivateCB();      break;
            case ButtonPressT:   XY_ButtonPressCB();   break;
            case ButtonReleaseT: XY_ButtonReleaseCB(); break;
            default:	           XY_ScrollCB(tag);
        }
    else if (w==&XZ)
        switch(tag){
            case SelectT:	   XZ_SelectCB();	 break;
            case ExposeT:        XZ_ExposeCB();        break;
            case ActivateT:      XZ_ActivateCB();      break;
            case ButtonPressT:   XZ_ButtonPressCB();   break;
            case ButtonReleaseT: XZ_ButtonReleaseCB(); break;
            default:	           XZ_ScrollCB(tag);
        }
    else if (w==&YZ)
        switch(tag){
            case SelectT:	   YZ_SelectCB();	 break;
            case ExposeT:        YZ_ExposeCB();        break;
            case ActivateT:      YZ_ActivateCB();      break;
            case ButtonPressT:   YZ_ButtonPressCB();   break;
            case ButtonReleaseT: YZ_ButtonReleaseCB(); break;
            default:	           YZ_ScrollCB(tag);
        }
    else if (w==&XYZ)
        switch(tag){
            case ExposeT: XYZ_ExposeCB(); break;
            case DataChangeT: Parent()->ChildCB(w, DataChangeT); break;
        }
    else if (w==&radiobox1){
        switch(tag){
            case ExposeT:   RadioBox1_ExposeCB();         break;
            case ActivateT: radio1=radiobox1.StateBits(); break;
        }
    }
    else if (w==&radiobox2){
        switch (tag){
            case ActivateT:
                radio2=radiobox2.StateBits();
                RadioBox1_ExposeCB();
                Refresh();
                break;
        }
    }
    else if (w==&databutton && w==ActivatedWindow()){
        switch(tag){
            case ButtonReleaseT: {
                DataButton_CB();
            } break;
        }
    }
    return Continue;
}//--- Editor::ChildCB ---//


fstream &Editor::IO(fstream &fs){
  if (!(fs|StkPtr|xo|yo|zo|dx|dy|dz|Dx|Dy|Dz))
    cerr << "Editor::IO Error!" << endl;
  DataIO(fs,(char*)Stack,sizeof(StkElm)*StkPtr);
  if (iostate_==ios::in) SetStack(StkPtr,StkSize,Stack);
  if (!(fs|rxo|ryo|rzo|Xmin|Ymin|Zmin|Xmax|Ymax|Zmax
          |ax0|ay0|az0|ax1|ay1|az1|ax2|ay2|az2|Er_in|Ur_in|Refl_in))
    cerr << "Editor::IO Error!" << endl;
  XY.IO(fs);
  XZ.IO(fs);
  YZ.IO(fs);
  XYZ.IO(fs);
  if (iostate_==ios::in){
    Top();
    Map();
    SetDrawParameters();
  }
  return fs;
}//--- Editor::IO ---//


void Editor::SetDrawParameters(){
  XY.SetDrawXY(Dx/2,Dy/2,Dx,Dy);
  XY.SetDrawW(xo,yo,dx,dy);
  XY.DrawXY(ax1,ay1);
  XZ.SetDrawXY(Dx/2,Dz/2,Dx,Dz);
  XZ.SetDrawW(xo,zo,dx,dz);
  XZ.DrawXY(ax2,az1);
  YZ.SetDrawXY(Dy/2,Dz/2,Dy,Dz);
  YZ.SetDrawW(yo,zo,dy,dz);
  YZ.DrawXY(az2,ay2);
}//--- Editor::SetDrawParameters ---//


void Editor::SetStack(int stkptr, int stksize, StkElm *stack){
float t;
Cursor parent_cursor=Parent()->SetCursor(Watch);
  StkPtr=stkptr;  StkSize=stksize;  Stack=stack;
  if (StkPtr>0){
  int   i;
    Xmin=Ymin=Zmin= MAXFLOAT;
    Xmax=Ymax=Zmax=-MAXFLOAT;
    for (i=0;i<StkPtr;i++){
      Xmin=Min(Xmin,Stack[i][0]);  Xmax=Max(Xmax,Stack[i][3]);
      Ymin=Min(Ymin,Stack[i][1]);  Ymax=Max(Ymax,Stack[i][4]);
      Zmin=Min(Zmin,Stack[i][2]);  Zmax=Max(Zmax,Stack[i][5]);
    }
    t=Xmin+Xmax; rxo=t/2; t/=10; Xmin=Min(Xmin-t,0.0f); Xmax+=t;
    t=Ymin+Ymax; ryo=t/2; t/=10; Ymin=Min(Ymin-t,0.0f); Ymax+=t;
    t=Zmin+Zmax; rzo=t/2; t/=10; Zmin=Min(Zmin-t,0.0f); Zmax+=t;
  }
  t=(Xmax-Xmin)/4;  ax1=t+Xmin;  ax2=ax1+2*t;
  t=(Ymax-Ymin)/4;  ay1=t+Ymin;  ay2=ay1+2*t;
  t=(Zmax-Zmin)/4;  az1=t+Zmin;  az2=az1+2*t;
  XY.SetDrawXY(Dx/2,Dz/2,Dx,Dy); XY.DrawXY(ax1,ay1);
  XZ.SetDrawXY(Dx/2,Dz/2,Dx,Dz); XZ.DrawXY(ax2,az1);
  YZ.SetDrawXY(Dy/2,Dz/2,Dy,Dz); YZ.DrawXY(az2,ay2);
  ax0=ax1; ay0=ay1; az0=az1;
  XY.SetDrawW(xo,yo,dx,dy);
  XY.SetWorld(Xmin,Ymin,Xmax,Ymax);
  XZ.SetDrawW(xo,zo,dx,dz);
  XZ.SetWorld(Xmin,Zmin,Xmax,Zmax);
  YZ.SetDrawW(yo,zo,dy,dz);
  YZ.SetWorld(Ymin,Zmin,Ymax,Zmax);
  XYZ.SetOrigin(rxo,ryo,rzo);
  t=sqrt(pow(Xmax-rxo,2)+pow(Ymax-ryo,2)+pow(Zmax-rzo,2));
  XYZ.SetWorld(rxo-t,ryo-t,rxo+t,ryo+t);
  XYZ.Clear();
  XYZ.WindowCB(ExposeT);
  Parent()->SetCursor(parent_cursor);
}//--- Editor::SetStack ---//


void Editor::GetMeshData(float &cf,float &er,float &ur,int &nx,int &ny,
			 int &nz){
float t;
  cf=Cf_; er=Er_; ur=Ur_;
  MeshData(Nx_,Ny_,Nz_,t,t,t);
  nx=Nx_; ny=Ny_; nz=Nz_;
}//--- Editor::GetMeshData ---//


void Editor::SetMeshData(float cf,float er,float ur,int nx,int ny,int nz){
  Cf_=cf; Er_=er; Ur_=ur; Nx_=nx; Ny_=ny; Nz_=nz;
}//--- Editro::SetMeshData ---//


void Editor::MeshData(int &nx,int &ny,int &nz,float &x,float &y,float &z){
int i;
  x=y=z=0.0f;
  for (i=0;i<StkPtr;i++){
    if (x<Stack[i][3]) x=Stack[i][3];
    if (y<Stack[i][4]) y=Stack[i][4];
    if (z<Stack[i][5]) z=Stack[i][5];
  }
  nx=int(x/Dx);  x=Cf_*(x-0.5*Dx);
  ny=int(y/Dy);  y=Cf_*(y-0.5*Dy);
  nz=int(z/Dz);  z=Cf_*(z-0.5*Dz);
}//--- Editor::MeshData ---//


int Editor::NumElements(int code){
int i,n;
  for (n=i=0;i<StkPtr;i++) if (int(Stack[i][6])==code) n++;
  return (code==ER_N_UR)?(n+1):n;
}//--- Editor::NumElements ---//


void Editor::GetWallStack(Float7 *stack){
int i,n;
float ddx=0.5*Dx,ddy=0.5*Dy,ddz=0.5*Dz;
  for (n=i=0;i<StkPtr;i++)
    if (int(Stack[i][6])==WALL){
      stack[n][0]=Cf_*(Stack[i][0]-ddx);  stack[n][3]=Cf_*(Stack[i][3]-ddx);
      stack[n][1]=Cf_*(Stack[i][1]-ddy);  stack[n][4]=Cf_*(Stack[i][4]-ddy);
      stack[n][2]=Cf_*(Stack[i][2]-ddz);  stack[n][5]=Cf_*(Stack[i][5]-ddz);
      stack[n][6]=Stack[i][7];  n++;
    }
}//--- Editor::GetWallStack ---//


void Editor::GetErUrStack(Float8 *stack){
float ddx=0.5*Dx,ddy=0.5*Dy,ddz=0.5*Dz;
int   i,n;
  stack[0][0]=stack[0][1]=stack[0][2]=0;
  stack[0][3]=Xmax_;  stack[0][4]=Ymax_;  stack[0][5]=Zmax_;
  stack[0][6]=Er_;    stack[0][7]=Ur_;
  for (n=1,i=0;i<StkPtr;i++)
    if (int(Stack[i][6]==ER_N_UR)){
      stack[n][0]=Cf_*(Stack[i][0]-ddx);  stack[n][3]=Cf_*(Stack[i][3]-ddx);
      stack[n][1]=Cf_*(Stack[i][1]-ddy);  stack[n][4]=Cf_*(Stack[i][4]-ddy);
      stack[n][2]=Cf_*(Stack[i][2]-ddz);  stack[n][5]=Cf_*(Stack[i][5]-ddz);
      stack[n][6]=Stack[i][7];	stack[n][7]=Stack[i][8];  n++;
    }
  qsort(stack+1,n-1,sizeof(stack[0]),fcmp(CompareE));
}//--- Editor::GetErUrStack ---//


void Editor::GetGeneratorStack(Int8 *stack){
int i,n;
  for (n=i=0;i<StkPtr;i++)
    if (int(Stack[i][6])==GENERATOR){
      stack[n][0]=Xnode(Stack[i][0]);  stack[n][3]=Xnode(Stack[i][3]);
      stack[n][1]=Ynode(Stack[i][1]);  stack[n][4]=Ynode(Stack[i][4]);
      stack[n][2]=Znode(Stack[i][2]);  stack[n][5]=Znode(Stack[i][5]);
      stack[n][6]=int(Stack[i][7]);    stack[n][7]=CONSTANT;	  n++;
    }
}//--- Editor::GetGeneratorStack ---//


void Editor::GetScopeStack(Int4 *stack){
int i,n;
  for (n=i=0;i<StkPtr;i++)
    if (int(Stack[i][6])==SCOPE){
      stack[n][0]=Xnode(0.5*(Stack[i][0]+Stack[i][3]));
      stack[n][1]=Ynode(0.5*(Stack[i][1]+Stack[i][4]));
      stack[n][2]=Znode(0.5*(Stack[i][2]+Stack[i][5]));
      stack[n][3]=int(Stack[i][7]);		   n++;
    }
}//--- Editor::GetScopeStack ---//


void Editor::GetAnimatorStack(Int7 *stack){
int i,n;
  for (n=i=0;i<StkPtr;i++)
    if (int(Stack[i][6])==ANIMATOR){
      stack[n][0]=Xnode(Stack[i][0]);  stack[n][3]=Xnode(Stack[i][3]);
      stack[n][1]=Ynode(Stack[i][1]);  stack[n][4]=Ynode(Stack[i][4]);
      stack[n][2]=Znode(Stack[i][2]);  stack[n][5]=Znode(Stack[i][5]);
      stack[n][6]=int(Stack[i][7]);   n++;
    }
}//--- Editor::GetAnimatorStack ---//


//===========================================================================
//
// End of Editor.C
//
//===========================================================================
