//===========================================================================
//
// GUI.C
//
// By Poman So
//
//===========================================================================
#ifdef __ActiveCC
#include <string.h>
#endif

#define INT_MAX 65535
#define INT_MIN (-65536)
#include <ctype.h>
#include <stdarg.h>
#include <stdio.h>
#include <time.h>
#include <math.h>
extern "C"{
#include "XWindow.h"
}
#include "GUI.h"
#include <iostream>
using namespace std;

//---------------------------------------------------------------------------
//
//	Label Class
//
//---------------------------------------------------------------------------
void Label::ExposeCB(){
int x,y;
  switch (hjustify){
    case LeftJ:  x=5;         break;
    case RightJ: x=Width()-5; break;
    default:	 x=Width()/2;
  }
  switch (vjustify){
    case LeftJ:  y=5;          break;
    case RightJ: y=Height()-5; break;
    default:	 y=Height()/2;
  }
  OutTextXY(x,y,label,border);
}//--- Label::ExposeCB() ---//


Label::Label(Window &w, int x1, int y1, int x2, int y2, int bw, int tdl,
	     char *s,Font f,Justify hj,Justify vj,Rubber rub,Boolean popup):
Window(w,x1,y1,(x2>x1)?x2:x1+w.TextWidth(s)+6,(y2>y1)?y2:y1+w.TextHeight(s)+6,
       DefaultColor,DefaultColor,DefaultColor,DefaultCursor,f,rub,bw){
  label       = s;
  SetBorder(bw,tdl);
  hjustify    = hj;
  vjustify    = vj;
  Popup(popup);
  Map();
}//--- Label::Label ---//


void Label::HighLight(Boolean rect){
int X1=0,Y1=0,X2=Width()-1,Y2=Height()-1,x,y;
  switch (hjustify){
    case LeftJ:  x=5;    break;
    case RightJ: x=X2-5; break;
    default:	 x=X2/2;
  }
  switch (vjustify){
    case LeftJ:  y=5;    break;
    case RightJ: y=Y2-5; break;
    default:	 y=Y2/2;
  }
  FillRectangle(X1,Y1,X2,Y2,border);
  OutTextXY(x,y,label,background);
  if (rect) Rectangle(X1,Y1,X2,Y2,background);
}//--- Label::HighLight ---//


Status Label::WindowCB(CBTag tag){
  if (tag==ExposeT) ExposeCB();
  return Continue;
}//--- Label::WindowCB ---//


void Label::SetLabel(char *s){
  label = s;
  FillRectangle(0,0,Width(),Height(),background);
  ExposeCB();
}//--- Label::SetLabel ---//



//---------------------------------------------------------------------------
//
//	PushButton Class
//
//---------------------------------------------------------------------------
PushButton::PushButton(Window &w, int x1, int y1, int x2, int y2, char *s,
		       Font f,Rubber rub):
Label(w,(x1>0 )?x1:w.Width()-PB_X1fromR,(y1> 0)?y1:w.Height()-PB_Y1fromB,
	(x2>x1)?x2:w.Width()-PB_Margin, (y2>y1)?y2:w.Height()-PB_Margin,
	1,1,s,f,CenterJ,CenterJ,rub){
}//--- PushButton::PushButton ---//


Status PushButton::WindowCB(CBTag tag){
    switch(tag){
        case ActivateT: {
            UngrabPointer();
        }
        break;
        case ButtonReleaseT: {
            if (ActivatedWindow()==this)
                return Parent()->ChildCB(this, ButtonReleaseT);
        }
        break;
        case EnterT: {
            if (ActivatedWindow()==this)
                HighLight();
        }
        break;
        case LeaveT: {
            if (ActivatedWindow()==this){
                Clear();
                Expose();
            }
        }
        break;
        case ExposeT: {
            ExposeCB();
        }
        break;
    }
    return Continue;
}//--- PushButton::Window
 


//---------------------------------------------------------------------------
//
//	EventButton Class
//
//---------------------------------------------------------------------------
void EventButton::ExposeCB(){
int X2=Width()-3,Y2=Height()-3;
  FillRectangle(2,2,X2,Y2,White,Black);
}//--- EventButton::ExposeCB ---//


EventButton::EventButton(Window &w, int x1, int y1, int x2, int y2,
			   Rubber rub):
Window(w,x1,y1,x2,y2,DefaultColor,DefaultColor,DefaultColor,DefaultCursor,
       DefaultFont,rub){
  Map();
}//--- EventButton::EventButton ---//


Status EventButton::WindowCB(CBTag tag){
  if (tag==ExposeT){
    ExposeCB();
    return Continue;
  }
  else
    return Parent()->ChildCB(this,tag);
}//--- EventButton::WindowCB ---//



//---------------------------------------------------------------------------
//
//	BottomButton Class
//
//---------------------------------------------------------------------------
BottomButton::BottomButton(Window &w, int x1, int y1, int x2, int y2,
			   Rubber rub):
EventButton(w,x1,y1,x2,y2,rub){
}//--- BottomButton::ButtomButton ---//


Status BottomButton::WindowCB(CBTag tag){
  switch (tag){
    case ButtonReleaseT:
      if (ActivatedWindow()==this){
      Cursor ppc=Parent()->Parent()->SetCursor(Watch);
	Parent()->Bottom();
        Parent()->Parent()->SetCursor(ppc);
      }
      break;
    case ExposeT:
      ExposeCB();
      break;
  }
  return Continue;
}//--- BottomButton::WindowCB ---//



//---------------------------------------------------------------------------
//
//	ResizeButton Class
//
//---------------------------------------------------------------------------
ResizeButton::ResizeButton(Window &w, int x1, int y1, int x2, int y2,
			   Color hc, Rubber rub):
Window(w,x1,y1,x2,y2,DefaultColor,DefaultColor,DefaultColor,DefaultCursor,
       DefaultFont,rub){
  highlightcolor = (hc==DefaultColor)?background:hc;
  Map();
}//--- ResizeButton::ResizeButton ---//


Status ResizeButton::WindowCB(CBTag tag){
  switch (tag){
    case ButtonPressT:
      if (ActivatedWindow()==this)
	if (Parent()->WindowCB(DResizeT)!=Okay)
	  Parent()->DragResize(100,100,MaxWidth(),MaxHeight());
      break;
    case ExposeT:{
      int X2=Width()-3,Y2=Height()-3;
        FillRectangle(2,2,X2,Y2,background,Black);
      }
      break;
  }
  return Continue;
}//--- ResizeButton::WindowCB ---//



//---------------------------------------------------------------------------
//
//	TitleBar Class
//
//---------------------------------------------------------------------------
TitleBar::TitleBar(Window &w, Boolean rsb, Boolean bb, char *left,
		   char *center, char *right, Font ft, int height):
Window(w,-1,-1,w.Width()-1,height,Green,Black,Black,DefaultCursor,ft,
       LeftTopRightR){
int X1,Y1,X2,Y2,bw;
  left_title  =left;
  center_title=center;
  right_title =right;
  Y1 = 2;
  Y2 = Height()-Y1-2;
  bw = Y2-Y1;
  X1 = Width()-Y2-2;
  X2 = X1+bw;
  if (rsb){
    resizebutton = new ResizeButton(w,X1,Y1,X2,Y2,White,TopRightR);
    X1-= bw+5;
    X2 = X1+bw;
  }
  else
    resizebutton = 0;
  if (bb)
    bottombutton = new BottomButton(w,X1,Y1,X2,Y2,TopRightR);
  else
    bottombutton = 0;
  Map();
}//--- TitleBar::TitleBar ---//


TitleBar::~TitleBar(){
  if (bottombutton) delete bottombutton;
  if (resizebutton) delete resizebutton;
}//--- TitleBar::~TitleBar ---//


Status TitleBar::WindowCB(CBTag tag){
  switch (tag){
    case ButtonReleaseT:
      if (ActivatedWindow()==this)
	Parent()->Top();
      break;
    case ButtonPressT:
      if (ActivatedWindow()==this){
      int cx1,cy1,cx2,cy2;
 	cx1=TB_Height-Width();
	cy1=2*TB_Height+3;
	cx2=TB_Height;
	cy2=TB_Height;
	Parent()->DragMove(DefaultDirection,cx1,cy1,cx2,cy2);
      }
      break;
    case ExposeT:
      LeftTitle(left_title);
      CenterTitle(center_title);
      RightTitle(right_title);
      break;
  }
  return Continue;
}//--- TitleBar::WindowCB ---//


void TitleBar::LeftTitle(char *s){
int w=5,y=Height()/2+((textfont==Small)?-1:1);
  textdir   = Horizontal;
  writemode = SetMode;
  vjustify  = CenterJ;
  hjustify  = LeftJ;
  OutTextXY(w,y,left_title,background);
  OutTextXY(w,y,left_title=s,Black);
 }//--- TitleBar::LeftTitle ---//


void TitleBar::CenterTitle(char *s){
int w=Width()/2,y=Height()/2+((textfont==Small)?-1:1);
  textdir   = Horizontal;
  writemode = SetMode;
  vjustify  = hjustify = CenterJ;
  OutTextXY(w,y,center_title,background);
  OutTextXY(w,y,center_title=s,Black);
}//--- TitleBar::CenterTitle ---//


void TitleBar::RightTitle(char *s){
int t,w,y=Height()/2+((textfont==Small)?-1:1);
  textdir   = Horizontal;
  writemode = SetMode;
  vjustify  = CenterJ;
  hjustify  = RightJ;
  if (bottombutton)
    bottombutton->Position(w,t,t,t);
  else if (resizebutton)
    resizebutton->Position(w,t,t,t);
  else
    w=Width();
  w -= 5;
  OutTextXY(w,y,right_title,background);
  OutTextXY(w,y,right_title=s,Black);
  Flush();
}//--- TitleBar::RightTitle ---//



//---------------------------------------------------------------------------
//
//	MenuBar Class
//
//---------------------------------------------------------------------------
int MenuBar::temp;

MenuBar::MenuBar(Window &w, int y, int height):
Window(w,-1,y,w.Width()-1,y+height,DefaultColor,DefaultColor,DefaultColor,
       DefaultCursor,DefaultFont,LeftTopRightR){
  nextx=5;
  Map();
}//--- MenuBar::MenuBar ---//


int MenuBar::ItemX1(){
  return nextx;
}//--- MenuBar::ItemX1 ---//


int MenuBar::ItemX2(char *s){
  return nextx+(temp=TextWidth(s)+15);
}//--- MenuBar::ItemX2 ---//


void MenuBar::SetItemX(){
  nextx += temp+15;
}//--- MenuBar::SetItemX ---//



//---------------------------------------------------------------------------
//
//	PulldownMenu Class
//
//---------------------------------------------------------------------------
Window *PulldownMenu::ActivatedMenuBar;
PulldownMenu *PulldownMenu::ActivatedPDM;
int PulldownMenu::temp;

void PulldownMenu::ButtonReleaseCB(){
  if (ActivatedMenuBar==Parent()){
    ActivatedPDM->Clear();
    ActivatedPDM->Expose();
    ActivatedPDM->itembox.Unmap();
    ActivatedPDM->itembox.Bottom();
    ActivatedMenuBar = ActivatedPDM = 0;
    UngrabPointer();
  }
}//--- PulldownMenu::ButtonReleaseCB ---//


PulldownMenu::PulldownMenu(MenuBar &mb, char *s):
Label(mb,mb.ItemX1(),0,mb.ItemX2(s),mb.Height()+1,0,0,s),
itembox(*mb.Parent(),X1(),mb.Y2()+1,X2(),mb.Y2()+2){
  mb.SetItemX();
  ActivatedMenuBar = ActivatedPDM = 0;
  itembox.Popup(True);
  itembox.SetBorder(1,1);
  max_item_width = Width();
}//--- PulldownMenu::PulldownMenu ---//


int PulldownMenu::ItemX1(){
  return 1;
}//--- PulldownMenu::ItemX1 ---//


int PulldownMenu::ItemY1(){
  return itembox.Height();
}//--- PulldownMenu::ItemY1 ---//


int PulldownMenu::ItemX2(char *s){
  max_item_width=Max(TextWidth(s)+16,max_item_width);
  return itembox.Width();
}//--- PulldownMenu::ItemX2 ---//


int PulldownMenu::ItemY2(char *s){
  temp=itembox.Height()+TextHeight(s)+7;
  return temp;
}//--- PulldownMenu::ItemY2 ---//


void PulldownMenu::SetItemXY(){
  itembox.Resize(itembox.X1()+max_item_width,itembox.Y1()+temp);
}//--- PulldownMenu::SetItemXY ---//


Status PulldownMenu::WindowCB(CBTag tag){
  switch (tag){
    case ActivateT:
      ActivatedMenuBar = Parent();
      ActivatedPDM     = this;
      HighLight(False);
      itembox.Top();
      itembox.Map();
      Parent()->Parent()->GrabPointer();
      break;
    case ButtonReleaseT:
      ButtonReleaseCB();
      break;
    case ExposeT:
      ExposeCB();
      break;
    case EnterT:
      if (ActivatedPDM!=this && ActivatedMenuBar==Parent()){
	ActivatedPDM->WindowCB(ButtonReleaseT);
	WindowCB(ActivateT);
      }
      break;
    case DeActivateT:
      ButtonReleaseCB();
      break;
  }
  return Continue;
}//--- PulldownMenu::WindowCB ---//



//---------------------------------------------------------------------------
//
//	PulldownItem Class
//
//---------------------------------------------------------------------------
PulldownItem::PulldownItem(PulldownMenu &pdm, char *s):
Label(pdm.ItemBox(),pdm.ItemX1(),pdm.ItemY1(),pdm.ItemX2(s),pdm.ItemY2(s),
      0,0,s,DefaultFont,LeftJ,CenterJ,LeftTopRightR){
  pdm.SetItemXY();
}//--- PulldownItem::PulldownItem ---//


Status PulldownItem::WindowCB(CBTag tag){
  switch (tag){
    case ButtonReleaseT: 
      ActivatedWindow()->WindowCB(DeActivateT);
      return Parent()->ChildCB(this,ButtonReleaseT);
    case ExposeT:ExposeCB();       break;
    case EnterT: HighLight(False); break;
    case LeaveT: Clear(); Expose();break;
  }
  return Continue;
}//--- PulldownItem::WindowCB ---//



//---------------------------------------------------------------------------
//
//	MovableWindow Class
//
//---------------------------------------------------------------------------
MovableWindow::MovableWindow(Window &w, int x1, int y1, int x2, int y2,
	       Color bk,Color bd,Color fg,Cursor cur, Font ft,Rubber rub):
Window(w,x1,y1,x2,y2,bk,bd,fg,cur,ft,rub){
  SetMovement(DefaultDirection,TB_Height,TB_Height,TB_Height,TB_Height);
}//--- MovableWindow::MovableWindow ---//


void MovableWindow::SetMovement(Direction d,int Mx1,int My1,int Mx2,int My2){
  direction=d;  mx1=Mx1;  my1=My1;  mx2=Mx2;  my2=My2;
}//--- MovableWindow::SetMovement ---//


Status MovableWindow::WindowCB(CBTag tag){
  switch (tag){
    case ButtonPressT:
      if (ActivatedWindow()==this){
      int cx1,cy1,cx2,cy2,pmx,pmy;
        Parent()->MouseXY(pmx,pmy);
	cx1=mx1-3;
	cy1=my1-3;
	cx2=mx2+Width()-2;
	cy2=my2+Height()-2;
	DragMove(direction,cx1,cy1,cx2,cy2);
	if (pmx!=Parent()->MouseX() || pmy!=Parent()->MouseY())
	  Parent()->ChildCB(this,MoveT);
      }
      break;
    default:
      return Parent()->ChildCB(this,tag);
  }
  return Continue;
}//--- MovableWindow::WindowCB ---//


//---------------------------------------------------------------------------
//
//	ScrollBar Class
//
//---------------------------------------------------------------------------
int ScrollBar::DecP(){
  return ((direction==Horizontal)?Height():Width())/2+SB_Margin;
}//--- ScrollBar::DecP ---//

int ScrollBar::IncP(){
  return ((direction==Horizontal)?Width():Height())-DecP()-1;
}//--- ScrollBar::IncP ---//

int ScrollBar::PageDecP(){
  return 2*DecP();
}//--- ScrollBar::PageDecP ---//

int ScrollBar::PageIncP(){
  return IncP()-DecP();
}//--- ScrollBar::PageIncP ---//

void ScrollBar::ExposeCB(){
int p[10],n,a,b,H=Height(),W=Width();
  if (W>H){
    a=movethumb.Height()+3;
    b=(H-a-1)/2;
    p[0]=p[3]=p[6]=p[8]=b;
    p[1]=p[2]=p[4]=p[9]=b+a/2;
    p[5]=b+a;
    p[7]=b+(a+1)/2;
    n = (p[7]==p[9])?4:5;
    Polygon(n,p,border);
    p[0]=p[6]=p[8]=W-b-1;
    p[2]=p[4]=W-p[1]-1;
    Polygon(n,p,border);
    p[0]=a/2+b+b+1;  p[1]=b+1;
  }
  else{
    a=movethumb.Width()+3;
    b=(W-a-1)/2;
    p[1]=p[2]=p[7]=p[9]=b;
    p[0]=p[3]=p[5]=p[8]=b+a/2;
    p[4]=b+a;
    p[6]=b+(a+1)/2;
    n = (p[6]==p[8])?4:5;
    Polygon(n,p,border);
    p[1]=p[7]=p[9]=H-b-1;
    p[3]=p[5]=H-p[0]-1;
    Polygon(n,p,border);
    p[0]=b+1;  p[1]=a/2+b+b+1;
  }
  p[2]=W-p[0]-1;   p[3]=H-p[1]-1;
  if (W>H)
    Rectangle(p[0]+2,p[1]+1,p[2]-2,p[3]-1,border);
  else
    Rectangle(p[0]+1,p[1]+2,p[2]-1,p[3]-2,border);
}//--- ScrollBar::ExposeCB ---//


ScrollBar::ScrollBar(Window &w,int x1,int y1,int x2,int y2):
Window(w,x1,y1,x2,y2),
movethumb(*this,0,0,10,10){
int	  mx1,my1,mx2,my2;
  if (Width()>Height()){
    direction = Horizontal;
    mx1 = mx2 = DecP()+3;
    my1 = my2 = 0;
    rubber = LeftRightBottomR;
  }
  else{
    direction = Vertical;
    mx1 = mx2 = 0;
    my1 = my2 = DecP()+3;
    rubber = TopRightBottomR;
  }
  SetP1P2(0,1,False);
  movethumb.SetMovement(direction,mx1,my1,mx2,my2);
  movethumb.Map();
  Map();
}//--- ScrollBar::ScrollBar ---//


float ScrollBar::P1(){
int   decp=DecP();
float f1,f=IncP()-decp;
  f1=(direction==Horizontal)?movethumb.X1():movethumb.Y1();
  return (f)?((f1-decp)/f):0;
}//--- ScrollBar::P1 ---//


float ScrollBar::P2(){
int   decp=DecP();
float f2,f=IncP()-decp;
  f2=(direction==Horizontal)?movethumb.X2():movethumb.Y2();
  return (f)?((f2-decp)/f):0;
}//--- ScrollBar::P2 ---//


void ScrollBar::SetP2(float p2){
int   X2,Y2,minvalue,decp=DecP(),incp=IncP();
float p1=P1(),f=incp-decp;
  if (p2<p1)p2=p1;
  if (p2>1) p2=1;
  minvalue=int(f*p1+decp);
  X2=Y2=int(f*p2+decp);
  if (direction==Horizontal){
    Y2=Height()-SB_Margin-1;
    X2=Min(Max(X2,minvalue+3),incp);
  }
  else{
    X2=Width()-SB_Margin-1;
    Y2=Min(Max(Y2,minvalue+3),incp);
  }
  movethumb.Resize(X2,Y2);
}//--- ScrollBar::SetP2 ---//


void ScrollBar::SetP1P2(float p1, float p2, Boolean moveresize){
int   X1,Y1,X2,Y2,decp=DecP(),incp=IncP();
float f=incp-decp;
  if (p1<0) p1=0;
  if (p1>1) p1=1;
  if (p2<p1)p2=p1;
  if (p2>1) p2=1;
  X1=Y1=int(f*p1+decp);
  X2=Y2=int(f*p2+decp);
  if (direction==Horizontal){
    Y1=SB_Margin-1;
    Y2=Height()-SB_Margin-1;
    X2=Min(Max(X2,X1+3),incp);
  }
  else{
    X1=SB_Margin-1;
    X2=Width()-SB_Margin-1;
    Y2=Min(Max(Y2,Y1+3),incp);
  }
  if (moveresize)
    movethumb.MoveResize(X1,Y1,X2,Y2);
  else
    movethumb.SetX1Y1X2Y2(X1,Y1,X2,Y2);
}//--- ScrollBar::SetP1P2 ---//


Status ScrollBar::WindowCB(CBTag tag){
  switch (tag){
    case ButtonReleaseT:
      if (ActivatedWindow()==this){
      int mx,my,mp;
	MouseXY(mx,my);
	mp = (direction==Horizontal)?mx:my;
	if (mp<PageDecP())
	  if (mp<DecP())
	    Parent()->ChildCB(this,DecT);
	  else
	    Parent()->ChildCB(this,PageDecT);
	else if (mp>PageIncP())
	  if (mp>IncP())
	    Parent()->ChildCB(this,IncT);
	  else
	    Parent()->ChildCB(this,PageIncT);
      }
      break;
    case ExposeT:
      ExposeCB();
      break;
    case ResizeT:
      return Parent()->ChildCB(this,ResizeT);
  }
  return Continue;
}//--- ScrollBar::WindowCB ---//


Status ScrollBar::ChildCB(Window *w, CBTag tag){
  if (w==&movethumb){
    switch (tag){
      case MoveT:
	return Parent()->ChildCB(this,ScrollT);
      case ResizeT:
	return Parent()->ChildCB(this,ResizeT);
    }
  }
  return Continue;
}//--- ScrollBar::ChildCB ---//



//---------------------------------------------------------------------------
//
//	ListBox Class
//
//---------------------------------------------------------------------------
float ListBox::P1(){
  return (num_elements>0)
		?float(first_show)/num_elements
		:0.0f;
}//--- ListBox::P1 ---//


float ListBox::P2(){
  return (num_elements>0)
		?Min(1.0f,float(ElementsPerPage()+first_show)/num_elements)
		:1.0;
}//--- ListBox::P2 ---//


void ListBox::HighLight(Boolean mark){
  if (highlight_element>=0 && highlight_element<num_elements){
  int   Y1=(highlight_element-first_show)*ItemHeight()+Margin(),
	Y2=Y1+ItemHeight()-1,
	W =Width()-vbar.Width()-3;
  Color bc,tc;
    if (mark){
      bc=drawcolor; tc=background;
    }
    else{
      bc=background; tc=drawcolor;
    }
    FillRectangle(1,Y1,W,Y2,bc);
    OutTextXY(Margin(),Y1,list[highlight_element],tc);
  }
}//--- ListBox::HighLight ---//


int ListBox::SelectedElement(){
int mx,my,selected;
  MouseXY(mx,my);
  selected = (my-Margin())/ItemHeight()+first_show;
  return (selected>=0 && selected<num_elements)?selected:-1;
}//--- ListBox::SelectedElement ---//


void ListBox::ShowList(){
  if (num_elements){
  int i,x=Margin(),y=x,BH=Height(),h=ItemHeight(),
      BW=Width()-vbar.Width()-1;
    for (i=first_show;i<num_elements;i++){
      if (i!=highlight_element){
	FillRectangle(0,y,BW,y+h-1,background);
	OutTextXY(x,y,list[i],border);
      }
      else
	HighLight();
      y+=h;
      if (y>BH) break;
    }
    if (y<BH) FillRectangle(0,y,BW,BH,background);
  }
}//--- ListBox::ShowList ---//


void ListBox::DecOneLine(){
  if (first_show>0){
    first_show--;
    ShowList();
  }
  vbar.SetP1P2(P1(),P2());
}//--- ListBox::DecOneLine ---//


void ListBox::IncOneLine(){
  if (first_show<MaxFirstShow()){
    first_show++;
    ShowList();
  }
  vbar.SetP1P2(P1(),P2());
}//--- ListBox::IncOneLine ---//


void ListBox::DecOnePage(){
  if (first_show>0){
    first_show-=ElementsPerPage();
    if (first_show<0) first_show=0;
    ShowList();
  }
  vbar.SetP1P2(P1(),P2());
}//--- ListBox::DecOnePage ---//


void ListBox::IncOnePage(){
int max_first_show=MaxFirstShow();
  if (first_show<max_first_show){
    first_show+=ElementsPerPage();
    if (first_show>max_first_show) first_show=max_first_show;
    ShowList();
  }
  vbar.SetP1P2(P1(),P2());
}//--- ListBox::IncOnePage ---//


void ListBox::ScrollList(){
int vbar_first_show=int(vbar.P1()*num_elements+0.5);
  if (first_show!=vbar_first_show){
    first_show = vbar_first_show;
    ShowList();
  }
}//--- ListBox::ScrollList ---//


ListBox::ListBox(Window &w, int x1, int y1, int x2, int y2, int num_e,
		 char **s, int f_s, int hl, Rubber rub):
Window(w,x1,y1,x2,y2,DefaultColor,DefaultColor,DefaultColor,DefaultCursor,
       DefaultFont,rub),
vbar(*this,Width()-SB_Width-1,-1,Width()-1,Height()-1){
  SetTextJustify(LeftJ,TopJ);
  SetList(num_e,s,f_s,hl);
  Map();
}//--- ListBox::ListBox ---//


Status ListBox::WindowCB(CBTag tag){
  if (num_elements<=0) return Continue;
  switch (tag){
    case ButtonReleaseT:
      if (ActivatedWindow()==this){
      int selected_element=SelectedElement();
	if (selected_element!=highlight_element){
	  HighLight(False);
	  highlight_element=selected_element;
	  HighLight();
	  Parent()->ChildCB(this,ButtonReleaseT);
	}
      }
      break;
    case SelectT:
      return Parent()->ChildCB(this,SelectT);
    case ExposeT: ShowList(); break;
  }
  return Continue;
}//--- ListBox::WindowCB ---//


Status ListBox::ChildCB(Window *w, CBTag tag){
  if (w==&vbar && num_elements){
    switch (tag){
      case DecT:     DecOneLine(); break;
      case IncT:     IncOneLine(); break;
      case PageDecT: DecOnePage(); break;
      case PageIncT: IncOnePage(); break;
      case ScrollT:  ScrollList(); break;
      case ResizeT:  vbar.SetP1P2(P1(),P2(),False); break;
    }
  }
  return Continue;
}//--- ListBox::ChildCB ---//


fstream &ListBox::IO(fstream &fs){
  return fs|first_show|highlight_element;
}//--- ListBox::IO ---//


void ListBox::SetList(int num_e,char **s,int f_s,int hl){
  num_elements      = num_e;
  list		    = s;
  highlight_element = hl;
  if (num_e>0 && highlight_element>=0)
    first_show = Min(Max(0,highlight_element-ElementsPerPage()/2),
		     MaxFirstShow());
  else
    first_show = f_s;
  ShowList();
  vbar.SetP1P2(P1(),P2());
}//--- ListBox::SetList ---//


void ListBox::HighLighted(int hl){
  if (hl>=0 && hl <num_elements){
    HighLight(False);
    highlight_element=hl;
    HighLight();
  }
}//--- ListBox::HighLighted ---//



//---------------------------------------------------------------------------
//
//	TextBox Class
//
//---------------------------------------------------------------------------
char TextBox::linebuffer[];

float TextBox::P1(){
  return (textlen>0)?float(firstshow)/textlen:0;
}//--- TextBox::P1 ---//


float TextBox::P2(Boolean findlastshow){
 if (findlastshow) ShowText(False);
 if (textlen>0){
 float p2=Max(P1(),Min(1.0f,float(lastshow-CharsPerLine())/textlen));
   return (lastshow<textlen)?p2:1;
 }
 else
   return 1;
}//--- TextBox::P2 ---//


int TextBox::FillBuffer(int start){
char C,*TextB,*TextB_max=text+textlen,*LineB,*LineB_max;
  for (TextB=text+start;TextB<TextB_max;TextB++){
    C=*TextB;
    if (!isspace(C) || C=='\t' || C=='\n') break;
    if (TextB>text && TextB[-1]=='\n') break;
  }
  LineB_max=linebuffer+Min(80,CharsPerLine());
  for (LineB=linebuffer;LineB<LineB_max && TextB<TextB_max;LineB++){
    *LineB=C=*TextB++;
    if (isspace(C)){
      if (C=='\n'){
        *LineB++=' ';
	break;
      }
      if (C=='\t'){
      int i;
	for (i=0,LineB;LineB<LineB_max && i<8;i++) *LineB++=' ';
	LineB--;
      }
      else
	*LineB=' ';
    }
  }
  *LineB=0;
  if (LineB[-1]!=' ' && !isspace(*TextB) && *TextB){
  char *sp=strrchr(linebuffer,' ');
    if (sp){
      TextB -= LineB-sp;
      LineB  = sp;
      *LineB = 0;
    }
    else{
      TextB--;
      LineB[-1]='-';
    }
  }
  return TextB-text;
}//--- TextBox::FillBuffer ---//

  
void TextBox::ShowText(Boolean show){
int firstc=firstshow,nextc,x=Margin(),y=x,h=LineHeight(),
    BH=vbar.Y2()-h,BW=vbar.X1()-2;
  while(y<BH){
    nextc=FillBuffer(firstc);
    if (nextc>firstc){
      if (show){
	FillRectangle(0,y,BW,y+h-1,background);
	OutTextXY(x,y,linebuffer,border);
      }
      firstc = nextc;
      y += h;
    }
    else
      break;
  }
  if (show) FillRectangle(0,y,BW,BH+h,background);
  lastshow=nextc;
}//--- TextBox::ShowText ---//


void TextBox::DecOneLine(){
  if (firstshow>0){
  int prev,next=0;
    do{
      prev=next;
      next=FillBuffer(prev);
    }while(next<firstshow);
    if (prev<firstshow){
      firstshow=prev;
      ShowText();
      vbar.SetP1P2(P1(),P2());
    }
  }
}//--- TextBox::DecOneLine ---//


Boolean TextBox::IncOneLine(Boolean showtext){
  if (lastshow<textlen){
    firstshow=FillBuffer(firstshow);
    ShowText(showtext);
    if (showtext) vbar.SetP1P2(P1(),P2());
    return True;
  }
  return False;
}//--- TextBox::IncOneLine ---//


void TextBox::DecOnePage(){
  if (firstshow>0 && LinesPerPage()>2){
  int *showrecord,prevptr,nextptr,lines=LinesPerPage(),target;
    showrecord = new int[lines];
    for (nextptr=0;nextptr<lines;nextptr++) showrecord[nextptr]=0;
    nextptr = showrecord[0] = 0;
    do{
      prevptr = nextptr++;
      nextptr %= lines;
      showrecord[nextptr]=FillBuffer(showrecord[prevptr]);
    }while(showrecord[nextptr]<firstshow);
    target = showrecord[(nextptr+1)%lines];
    delete showrecord;
    if (target<firstshow){
      firstshow=target;
      ShowText();
      vbar.SetP1P2(P1(),P2());
    }
  }
}//--- TextBox::DecOnePage ---//


void TextBox::IncOnePage(){
int i,oldfirst=firstshow,oldlast=lastshow,lines=LinesPerPage()-1;
  for (i=0;i<lines && IncOneLine(False);i++);
  if (oldfirst<firstshow){
    ShowText();
    vbar.SetP1P2(P1(),P2());
  }
  else
    lastshow=oldlast;
}//--- TextBox::IncOnePage ---//


void TextBox::ScrollText(){
float vp1=vbar.P1();
int   oldfirst=firstshow,oldlast=lastshow;
  if (vp1<P1() || vp1==0) firstshow=lastshow=0;
  while(P1()<vp1 && IncOneLine(False));
  if (oldfirst!=firstshow){
    ShowText();
    handle_resize=False;
    if (lastshow<textlen){
      if (P2()!=vbar.P2()) vbar.SetP2(P2());
    }
    else
      vbar.SetP1P2(P1(),P2());
    handle_resize=True;
  }
  else
    lastshow = oldlast;
}//--- TextBox::ScrollText ---//


TextBox::TextBox(Window &w, int x1, int y1, int x2, int y2, char *s):
Window(w,x1,y1,x2,y2),
vbar(*this,Width()-SB_Width-1,-1,Width()-1,Height()-1){
  rubber    = AllSideR;
  handle_resize = True;
  SetTextJustify(LeftJ,TopJ);
  SetText(s);
  Map();
}//--- TextBox::TextBox ---//


Status TextBox::WindowCB(CBTag tag){
  switch (tag){
    case ExposeT: ShowText(); break;
  }
  return Continue;
}//--- TextBox::WindowCB ---//


Status TextBox::ChildCB(Window *w, CBTag tag){
  if (w==&vbar){
    switch (tag){
      case DecT:     DecOneLine(); break;
      case IncT:     IncOneLine(); break;
      case PageDecT: DecOnePage(); break;
      case PageIncT: IncOnePage(); break;
      case ScrollT:  ScrollText(); break;
      case ResizeT:
	if (handle_resize)
	  vbar.SetP1P2(P1(),P2(True),False);
	break;
    }
  }
  return Continue;
}//--- TextBox::ChildCB ---//


void TextBox::SetText(char *s){
  text      = s;
  textlen   = (text!=0)?strlen(text):0;
  firstshow = 0;
  linebuffer[0] = 0;
  if (textlen>0) ShowText();
  vbar.SetP1P2(P1(),P2());
}//--- TextBox::SetText ---//



//---------------------------------------------------------------------------
//
//	TextInput Class
//
//---------------------------------------------------------------------------
char TextInput::BS;
char TextInput::CR;
char TextInput::WB[];
int  TextInput::cursor_w;
int  TextInput::cursor_h;
int  TextInput::text_x;

void TextInput::HideCursor(){
  if (cursor_show){
    FillRectangle(cursor_x,cursor_y,cursor_x+cursor_w,cursor_y+cursor_h,
	background);
    cursor_show = False;
    cursor_clock = EventClock();
  }
}//--- TextInput::HideCursor ---//


void TextInput::ShowCursor(){
  if (!cursor_show){
    FillRectangle(cursor_x,cursor_y,cursor_x+cursor_w,cursor_y+cursor_h,Black);
    cursor_show = True;
    cursor_clock = EventClock();
  }
}//--- TextInput::ShowCursor ---//


void TextInput::ActivateCB(){
  ShowCursor();
}//--- TextInput::ActivateCB ---//


void TextInput::ExposeCB(){
int h=Height()-1;
  OutTextXY(text_x,cursor_y+5,buffer);
  if (!active) SetLineStyle(DottedL);
  Line(0,0,0,h,border);
  Line(0,h,Width()-1,h,border);
  SetLineStyle(SolidL);
}//--- TextInput::ExposeCB ---//


void TextInput::NoEventCB(){
  if (TimeSec(cursor_clock,EventClock())>0.5)
    if (cursor_show)
      HideCursor();
    else
      ShowCursor();
}//--- TextInput::NoEventCB ---//


void TextInput::LastNoEventCB(){
  HideCursor();
}//--- TextInput::LastNoEventCB ---//

#define _Up_Key_    -1
#define _Down_Key_   13
#define _Return_Key_ 13
Status TextInput::KeyPressCB(){
int  KP=KeyPressed();
char inchar=char(KP);
  if ((inchar==CR || KP==_Up_Key_ || KP==_Down_Key_) && textlen<=maxindex)
    return Parent()->ChildCB(this,KeyPressT);
  else if(inchar==BS){
    if(textlen>0){
      HideCursor();
      WB[0]=buffer[--textlen];
      buffer[textlen]=0;
      cursor_x-=TextWidth(WB);
      OutTextXY(cursor_x,cursor_y+5,WB,background);
    }
    else
      Bell();
  }
  else if (textlen<maxindex && cursor_x<(Width()-10) && (inchar==' ' || 
  	   isalnum(inchar) || inchar=='.' || inchar=='-' || inchar=='+')){
    HideCursor();
    WB[0]=buffer[textlen++]=inchar;
    buffer[textlen]=0;
    OutTextXY(cursor_x,cursor_y+5,WB,Black);
    cursor_x+=TextWidth(WB);
  }
  else if (textlen>=maxindex || cursor_x>=(Width()-10))
    Bell();
  return Continue;
}//--- TextInput::KeyPressCB ---//


TextInput::TextInput(Window &w, int x1, int y1, int x2, int y2, int bs,
		     char *bf, int bw, Boolean act):
Window(w,x1,y1,x2,y2){
  SetBorder(bw,ThreeDLook());
  BS		= '\b';
  CR		= '\r';
  WB[1]		= 0;
  active	= act;
  buffer	= bf;
  maxindex	= bs-1;
  buffer[maxindex]= 0;
  textlen	= Min(int(maxindex),int(strlen(buffer)));
  text_x	= 5;
  cursor_show	= False;
  cursor_clock	= EventClock();
  cursor_w	= TextWidth("Z");
  cursor_h	= int(1.4*TextHeight("g"));
  cursor_x	= text_x+TextWidth(buffer);
  cursor_y	= (Height()-cursor_h)/2;
  hjustify	= LeftJ;
  vjustify	= TopJ;
  Map();
}//--- TextInput::TextInput ---//


Status TextInput::WindowCB(CBTag tag){
  switch (tag){
    case ActivateT:    if (active) ActivateCB();	break;
    case ExposeT:      ExposeCB();			break;
    case NoEventT:     if (active) NoEventCB();		break;
    case LastNoEventT: if (active) LastNoEventCB();	break;
    case KeyPressT:    if (active) return KeyPressCB(); break;
  }
  return Continue;
}//--- TextInput::WindowCB ---//


void TextInput::BufferContentChangedCB(){
  HideCursor();
  textlen  = Min(int(maxindex),int(strlen(buffer)));
  cursor_x = text_x+TextWidth(buffer);
  FillRectangle(0,0,Width(),Height(),background);
  ExposeCB();
}//--- TextInput::BufferContentChangeCB ---//



//---------------------------------------------------------------------------
//
//	RadioBox Class
//
//---------------------------------------------------------------------------
int RadioBox::button_x;
int RadioBox::button_y;
int RadioBox::button_w;
int RadioBox::button_h;
int RadioBox::margin;

void RadioBox::ButtonXY(int n){
  if (direction==Vertical){
    button_x = margin;
    button_y = int(margin+n*double(Height()-2*button_x)/num_buttons);
  }
  else{
  int i,j,*w=new int[num_buttons+1];
    for (w[0]=margin,j=1,i=0;i<num_buttons;i++,j++)
      w[j]=w[i]+TextWidth(labels[i]);
    button_x = int(w[n]+n*double(Width()-w[i]-margin)/num_buttons);
    button_y = margin;
    delete w;
  }
}//--- RadioBox::ButtonXY ---//


void RadioBox::DrawButton(int i){
Color fc=((1<<i)&state_bits)?border:background;
  SetLineStyle(SolidL);
  SetFillStyle(SolidF,fc);
  ButtonXY(i);
  if (OnlyOneOn){
  int xr=button_w/2,yr=button_h/2;
    button_x+=xr;  button_y+=yr;
    Ellipse(button_x,button_y,xr,yr,border);
    FillEllipse(button_x,button_y,xr-2,yr-2,fc);
  }
  else{
  int x2=button_x+button_w,y2=button_y+button_h;
    Rectangle(button_x,button_y,x2,y2,border);
    FillRectangle(button_x+3,button_y+3,x2-2,y2-2,fc);
  }
}//--- RadioBox::DrawButton ---//


void RadioBox::DrawLabel(int i){
int x,y,t;
  LabelXYs(i,x,y,t,t);
  OutTextXY(x,y,labels[i],border);
}//--- RadioBox::DrawLabel ---//


Boolean RadioBox::ButtonActivated(int i){
int mx,my;
  MouseXY(mx,my);
  ButtonXY(i);
  return (mx<button_x || my<button_y || mx>(button_x+button_w) ||
	  my>(button_y+button_h))
    ?False
    :True;
}//--- RadioBox::ButtonActivated ---//


void RadioBox::ExposeCB(){
int i;
  for (i=0;i<num_buttons;i++){
    DrawButton(i);
    DrawLabel(i);
  }
}//--- RadioBox::ExposeCB ---//


void RadioBox::ActivateCB(){
int i;
  for (i=0;i<num_buttons && !ButtonActivated(i);i++);
  if (i<num_buttons){
    if (OnlyOneOn)
      state_bits=(1<<i);
    else
      state_bits^=(1<<i);
    ExposeCB();
  }
}//--- RadioBox::ActivateCB ---//


RadioBox::RadioBox(Window &w, int x1, int y1, int x2, int y2, int num_b,
		   char **s, int state_b, Boolean Only_1_On, Direction dir):
Window(w,x1,y1,x2,y2,DefaultColor,DefaultColor,DefaultColor,DefaultCursor,
	DefaultFont,DefaultRubber,0){
  num_buttons = num_b;
  labels      = s;
  OnlyOneOn   = Only_1_On;
  state_bits  = (OnlyOneOn && !state_b)?1:state_b;
  direction   = dir;
  margin      = 5;
  button_w    = button_h = 11;
  hjustify    = LeftJ;
  vjustify    = TopJ;
  Map();
}//--- RadioBox::RadioBox ---//


Status RadioBox::WindowCB(CBTag tag){
  switch (tag){
    case ExposeT:
      ExposeCB();
      Parent()->ChildCB(this,ExposeT);
      break;
    case ActivateT:
      ActivateCB();
      return Parent()->ChildCB(this,ActivateT);
  }
  return Continue;
}//--- RadioBox::WindowCB ---//


fstream &RadioBox::IO(fstream &fs){
  return fs|state_bits;
}//--- RadioBox::IO ---//


void RadioBox::LabelXYs(int i, int &x1, int &y1, int &x2, int &y2){
int t;
  ButtonXY(i);
  x1 = t = button_x+button_w+margin;
  y1 = button_y;
  x2 = t+TextWidth(labels[i]);
  y2 = button_y+TextHeight(labels[i]);
}//--- RadioBox::LabelXYs ---//



//---------------------------------------------------------------------------
//
//	TaskWindow Class
//
//---------------------------------------------------------------------------
void TaskWindow::ResizeCB(){
Boolean reset=False;
int     W=Width(),H=Height(),x1=X1(),y1=Y1(),x2,y2;
  if (W<minW){      x2=x1+minW; reset=True;}
  else if (W>maxW){ x2=x1+maxW; reset=True;}
  else x2=X2();
  if (H<minH){      y2=y1+minH; reset=True;}
  else if (H>maxH){ y2=y1+maxH; reset=True;}
  else y2=Y2();
  if (reset) SetX1Y1X2Y2(x1,y1,x2,y2);
}//--- TaskWindow::ResizeCB ---//


TaskWindow::TaskWindow(Window &w, int x1, int y1, int x2, int y2, Boolean RB,
		       Boolean BB, char *left, char *center, char *right):
Window(w,x1,y1,x2,y2,DefaultColor,DefaultColor,DefaultColor,DefaultCursor,
	DefaultFont,DefaultRubber,2),
titlebar(*this,RB,BB,left,center,right){
int     W,H;
Boolean reset=False;
  minW = 86;
  minH = 100;
  maxW = MaxWidth();
  maxH = MaxHeight();
  W    = Width();
  H    = Height();
  x1=X1();  y1=Y1();  x2=X2();  y2=Y2();  
  if (W<minW)     {x2=x1+minW; reset=True;}
  else if (W>maxW){x2=x1+maxW; reset=True;}
  if (H<minH)     {y2=x1+minH; reset=True;}
  else if (H>maxH){y2=y1+maxH; reset=True;}
  if (reset) SetX1Y1X2Y2(x1,y1,x2,y2);
}//--- TaskWindow::TaskWindow ---//


Status TaskWindow::WindowCB(CBTag tag){
  switch (tag){
    case DResizeT: DragResize(minW,minH,maxW,maxH); return Okay;
    case ResizeT:  ResizeCB(); break;
  }
  return Continue;
}//--- TaskWindow::WindowCB ---//


void TaskWindow::SetMinMax(int minw, int minh, int maxw, int maxh){
  minW=minw;  minH=minh;  maxW=maxw;  maxH=maxh;
}//--- TaskWindow::SetMinMax ---//



//---------------------------------------------------------------------------
//
//	MessageBox Class
//
//---------------------------------------------------------------------------
int MessageBox::ButtonWidth(char *b_L){
  return Max(PB_Width,TextWidth(b_L)+8);
}//--- MessageBox::ButtonWidth ---//


void MessageBox::DeleteButtons(){
  if (b1){ delete b1; b1=0; }
  if (b2){ delete b2; b2=0; }
  if (b3){ delete b3; b3=0; }
}//--- MessageBox::DeleteButtons ---//


char *MessageBox::ShowMessage(char *message, char *b1_L, char *b2_L,
			      char *b3_L,Boolean beep){
  DeleteButtons();
  if (!b3_L) b3_L="Okay";
  b3 = new PushButton(*this,Width()-ButtonWidth(b3_L)-PB_Margin,
		      Height()-PB_Y1fromB,Width()-PB_Margin,
		      PB_Margin,b3_L);
  if (b2_L){
    b2 = new PushButton(*this,b3->X1()-ButtonWidth(b2_L)-PB_Margin-2,
		      b3->Y1(),b3->X2()-ButtonWidth(b2_L)-PB_Margin-2,
		      b3->Y2(),b2_L);
    if (b1_L)
      b1 = new PushButton(*this,b2->X1()-ButtonWidth(b1_L)-PB_Margin-2,
		      b2->Y1(),b2->X2()-ButtonWidth(b1_L)-PB_Margin-2,
		      b2->Y2(),b1_L);
  }
  textbox.SetText(message);
  if (beep) Bell();
  DispatchEvents();
  DeleteButtons();
  return button_pushed;
}//--- MessageBox::ShowMessage ---//


MessageBox::MessageBox(Window &w, int x1, int y1, int x2, int y2,char *title):
TaskWindow(w,x1,y1,x2,y2,True,False,"",title,""),
textbox(*this,5,titlebar.Y2()+6,Width()-5,Height()-PB_Y1fromB-PB_Margin){
  b1=b2=b3=0;
  button_pushed=0;
}//--- MessageBox::MessageBox ---//


MessageBox::~MessageBox(){
  DeleteButtons();
}//--- MessageBox::~MessageBox ---//


Status MessageBox::WindowCB(CBTag tag){
  return TaskWindow::WindowCB(tag);
}//--- MessageBox::WindowCB ---//


Status MessageBox::ChildCB(Window *w, CBTag tag){
  if (tag==ButtonReleaseT){
  Window *ActW=ActivatedWindow();
    if (w && ActW==w && (ActW==b1 || ActW==b2 || ActW==b3)){
      button_pushed = ((PushButton*)ActW)->GetLabel();
      Unmap(Watch);
      if (DispatchingWindow()==this)
	return Quit;
      else return Parent()->ChildCB(this,ButtonReleaseT);
    }
  }
  return Continue;
}//--- MessageBox::ChildCB ---//


void MessageBox::Message(char *message, char *b1_L, Boolean beep){
  ShowMessage(message,0,0,b1_L,beep);
}//--- MessageBox::Message ---//


char *MessageBox::Message(char *message,char *b1_L,char *b2_L,Boolean beep){
  return ShowMessage(message,0,b1_L,b2_L,beep);
}//--- MessageBox::Message ---//
  

char *MessageBox::Message(char *message, char *b1_L, char *b2_L, char *b3_L,
			  Boolean beep){
  return ShowMessage(message,b1_L,b2_L,b3_L,beep);
}//--- MessageBox::Message ---//



//---------------------------------------------------------------------------
//
//	FileBox
//
//---------------------------------------------------------------------------
void FileBox::FreeFileList(){
  if (listsize>0){
    ::FreeFileList(&filelist);
    filelist = 0;
    listsize = 0;
  }
}//--- FileBox::FreeFileList ---//


int FileBox::GetFileList(char *ext){
int i;
  FreeFileList();
  listsize = ::GetFileList(0,ext,&filelist);
  if (listsize>0){
    for(i=0;i<listsize;i++)
      if(!strcmp(filelist[i],filename))
        break;
    if (i==listsize) i=-1;
  }
  listbox.SetList(listsize,filelist,0,i);
  return listsize;
}//--- FileBox::GetFileList ---//

static char *OpenLabel="Open",
	    *SaveLabel="Save",
	    *CancelLabel="Cancel";

FileBox::FileBox(Window &w, int x, int y):
TaskWindow(w,x,y,x+200,y+307,True,False,"","File Box",""),
okaybutton(*this,Width()-Max(PB_Width,TextWidth(CancelLabel)+8)-PB_Margin,
		 Height()-PB_Y1fromB,Width()-PB_Margin,PB_Margin,OpenLabel),
cancelbutton(*this,okaybutton.X1()-okaybutton.Width()-PB_Margin,
	     okaybutton.Y1(),okaybutton.X1()-PB_Margin,okaybutton.Y2(),
	     CancelLabel),
textinput(*this,PB_Margin-1,okaybutton.Y1()-int(1.7*PB_Height),
	  okaybutton.X2()+1,okaybutton.Y1()-PB_Margin,DosFileNameSize,
	  buffer),
listbox(*this,PB_Margin,titlebar.Y2()+PB_Margin+2,okaybutton.X2(),
	textinput.Y1()-PB_Margin-2){
  minW = 200;
  maxW = 300;
  minH = 300;
  maxH = 400;
  buffer[0]= 0;
  listsize = 0;
  filelist = 0;
  textinput.SetRubber(LeftRightBottomR);
  listbox.SetRubber(AllSideR);
}//--- FileBox::FileBox ---//


FileBox::~FileBox(){
  FreeFileList();
}//--- FileBox::~FileBox ---//


Status FileBox::WindowCB(CBTag tag){
  return TaskWindow::WindowCB(tag);
}//--- FileBox::WindowCB ---//


Status FileBox::ChildCB(Window *w, CBTag tag){
  if (tag==ButtonReleaseT){
  Window *ActW=ActivatedWindow();
    if (w==&listbox && ActW==w){
    int i=listbox.HighLighted();
      if (i>=0)
        strcpy(buffer,filelist[i]);
      else
        buffer[0]=0;
      textinput.BufferContentChangedCB();
    }
    else if (w==&okaybutton && ActW==&okaybutton){
      if (strlen(buffer)){
	strcpy(filename,buffer);
	return Okay;
      }
      else
	return Null;
    }
    else if (w==&cancelbutton && ActW==&cancelbutton)
      return Cancel;
  }
  else if (w==&textinput && tag==KeyPressT && (char)KeyPressed()=='\r'){
    strcpy(filename,buffer);
    return Okay;
  }
  else if (w==&listbox && tag==SelectT){
    if (strlen(buffer)){
      strcpy(filename,buffer);
      return Okay;
    }
  }
  return Continue;
}//--- FileBox::ChildCB ---//


Status FileBox::GetFile(char *f, char *ext, Boolean ToOpen){
Status status;
char *dot;
  filename = f;
  strcpy(buffer,f);
  textinput.BufferContentChangedCB();
  GetFileList(ext);
  if (ToOpen)
    okaybutton.SetLabel(OpenLabel);
  else{
    okaybutton.SetLabel(SaveLabel);
    textinput.ToNoEventWindow();
  }
  status = DispatchEvents();
  if (status==Okay && ext && strlen(ext)>0){
    dot = strrchr(filename,'.');
    if (dot) dot[0]=0;
    strcat(filename,".");
    strcat(filename,ext);
  }
  Unmap(Watch);
  FreeFileList();
  return status;
}//--- FileBox::GetFile ---//



//---------------------------------------------------------------------------
//
//	DataPrompt Class
//
//---------------------------------------------------------------------------
int DataPrompt::prompt_x1;
int DataPrompt::prompt_y1;
int DataPrompt::lineheight;
int DataPrompt::spacing;
int DataPrompt::bsize;
int DataPrompt::csize;

void DataPrompt::ExposeCB(){
    int  i, y, dy=lineheight+spacing;
    char *temp;
    for (y=prompt_y1, i=0;i<num_textinputs; i++, y+=dy){
        char *buffer = new char[strlen(prompts[i])+1];
        strcpy(buffer,prompts[i]);
        temp = strrchr(buffer, '%');
        if (temp) *temp= 0;
        inputwindow.OutTextXY(prompt_x1, y+5, buffer, Black);
        delete buffer;
    }
}//--- DataPrompt::ExposeCB ---//


void DataPrompt::GetBuffer(int i){
char     *bp=buffer+i*bsize,*cp=code+i*csize;
long     ld,LD;
unsigned long lu,LU;
double   lg;
  LD = 65535;
  LU = 2*65536-1;
  if (!strcmp(cp,"%s"))
    strcpy((char*)data[i],bp);
  else if (!strcmp(cp,"%c")){
    if (bp[1]==0) *((char*)data[i])=bp[0];
  }
  else if (!strcmp(cp,"%d")){
    sscanf(bp,"%ld",&ld);
    if (ld>-LD && ld<LD)
      *((int*)data[i])=int(ld);
  }
  else if (!strcmp(cp,"%u")){
    sscanf(bp,"%lu",&lu);
    if (lu<LU)
      *((unsigned*)data[i])=unsigned(lu);
  }
  else if (!strcmp(cp,"%ld")){
    sscanf(bp,"%ld",&ld);
    *((long*)data[i])=ld;
  }
  else if (!strcmp(cp,"%lu")){
    sscanf(bp,"%lu",&lu);
    *((unsigned long*)data[i])=lu;
  }
  else if (!strcmp(cp,"%g") || !strcmp(cp,"%f") || !strcmp(cp,"%e")){
    sscanf(bp,"%lg",&lg);
    if (lg>double(-MAXFLOAT) && lg<double(MAXFLOAT))
      *((float*)data[i])=lg;
  }
  else if (!strcmp(cp,"%lg") || !strcmp(cp,"%lf") || !strcmp(cp,"%le")){
    sscanf(bp,"%lg",&lg);
    *((double*)data[i])=lg;
  }
}//--- DataPrompt::GetBuffer ---//


void DataPrompt::SetBuffer(int i){
char *bp=buffer+i*bsize,*cp=code+i*csize;
  if (!strcmp(cp,"%s"))
    sprintf(bp,cp,(char*)data[i]);
  else if (!strcmp(cp,"%c"))
    sprintf(bp,cp,*((char*)data[i]));
  else if (!strcmp(cp,"%d"))
    sprintf(bp,cp,*((int*)data[i]));
  else if (!strcmp(cp,"%u"))
    sprintf(bp,cp,*((unsigned*)data[i]));
  else if (!strcmp(cp,"%ld"))
    sprintf(bp,cp,*((long*)data[i]));
  else if (!strcmp(cp,"%lu"))
    sprintf(bp,cp,*((unsigned long*)data[i]));
  else if (!strcmp(cp,"%g") || !strcmp(cp,"%f") || !strcmp(cp,"%e"))
    sprintf(bp,cp,*((float*)data[i]));
  else if (!strcmp(cp,"%lg") || !strcmp(cp,"%lf") || !strcmp(cp,"%le"))
    sprintf(bp,cp,*((double*)data[i]));
  if (textinput[i]) textinput[i]->BufferContentChangedCB();
}//--- DataPrompt::SetBuffer ---//


DataPrompt::DataPrompt(Window &w, Window *cr_cb, int x1, int y1, int x2, int y2,
        char *title, char **in_prompts, Boolean *active,
        void **in_data):
        TaskWindow(w, x1, y1, x2, y2, True, False, "", title, ""),
                okaybutton(*this, 0, 0, 0, 0, "Okay"),
                cancelbutton(*this, okaybutton.X1()-PB_X1fromR-PB_Margin, okaybutton.Y1(),
                okaybutton.X2()-PB_X1fromR-PB_Margin, okaybutton.Y2(), "Cancel"),
                inputwindow(*this, PB_Margin-1, titlebar.Y2()+PB_Margin+2, Width()-PB_Margin+1,
                okaybutton.Y1()-PB_Margin-2){
            
            minW = Width();
            minH = Height();
            maxW = MaxWidth();
            maxH = MaxHeight();
            inputwindow.SetRubber(AllSideR);
            bsize	       = 80;
            csize        = 4;
            prompt_x1    = 5;
            prompt_y1    = titlebar.Y2()+2*prompt_x1;
            lineheight   = 2*TextHeight("Z");
            spacing      = 2;
            hjustify     = LeftJ;
            vjustify     = TopJ;
            textinput_x1 = 0;
            prompts      = in_prompts;
            data         = in_data;
            for (num_textinputs=0; prompts[num_textinputs][0]; num_textinputs++){
                char *buffer = new char[strlen(prompts[num_textinputs])+1];
                strcpy(buffer,prompts[num_textinputs]);
                char *temp=NULL;
                temp = strrchr(buffer, '%');
                if (temp) *temp=0;
                textinput_x1 = Max(textinput_x1, TextWidth(buffer));
                delete buffer;
            }
            textinput_x1 += 2*prompt_x1;
            if (num_textinputs){
                int i, dy=lineheight+spacing, x1=textinput_x1, y1=prompt_y1-2,
                        x2=inputwindow.Width()-prompt_x1, y2=y1+lineheight-1;
                char *bp, *cp;
                code  =cp = new char[num_textinputs*csize];
                buffer=bp = new char[num_textinputs*bsize];
                textinput = new TextInput*[num_textinputs];
                for (i=0;i<num_textinputs;i++, y1+=dy, y2+=dy, bp+=bsize, cp+=csize){
                    char *temp;
                    temp = strrchr(prompts[i], '%');
                    if (temp){
                        strncpy(cp, temp, 3);
                        cp[3] = 0;
                        // strlwr(cp);
                    }
                    else
                        cp[0] = 0;
                    bp[0]=0;
                    textinput[i] = new TextInput(inputwindow, x1, y1, x2, y2, bsize, bp, 0,
                            active[i]);
                    textinput[i]->SetRubber(LeftTopRightR);
                }
            }
            inputwindow.Map();
            CR_CB = cr_cb;
}//--- DataPrompt::DataPrompt ---//


DataPrompt::~DataPrompt(){
  if (num_textinputs){
  int i;
    for (i=num_textinputs-1;i>=0;i--) delete textinput[i];
    delete textinput;
    delete buffer;
    delete code;
  }
}//--- DataPrompt::~DataPrompt ---//


Status DataPrompt::WindowCB(CBTag tag){
int i;
  if (tag==MapT){
    for (i=0;i<num_textinputs;i++) SetBuffer(i);
    textinput[0]->ToNoEventWindow();
    return Continue;
  }
  else
    return TaskWindow::WindowCB(tag);
}//--- DataPrompt::WindowCB ---//



Status DataPrompt::ChildCB(Window  *w, CBTag tag){
int  KP=KeyPressed();
char inchar = char(KP);
  if (w==&inputwindow && tag==ExposeT) ExposeCB();
  if (tag==KeyPressT && (inchar=='\r' || KP==_Up_Key_ || KP==_Down_Key_)){
  int i;
    for (i=0;i<num_textinputs;i++) if (w==textinput[i]) break;
    if (i<num_textinputs){
      GetBuffer(i);
      SetBuffer(i);
      if (CR_CB) CR_CB->WindowCB(DataChangeT);
      if (num_textinputs==1){
        Unmap(Watch);
	return Okay;
      }
      else{
        i=(i+((KP==_Up_Key_)?(num_textinputs-1):1))%num_textinputs;
        textinput[i]->ToNoEventWindow();
      }
    }
  }
  else if (tag==ButtonReleaseT){
    if (w==&cancelbutton && ActivatedWindow()==&cancelbutton){
      Unmap(Watch);
      return Cancel;
    }
    else if (w==&okaybutton && ActivatedWindow()==&okaybutton){
    int i;
      for (i=0;i<num_textinputs;i++){
	GetBuffer(i);
	SetBuffer(i);
      }
      if (CR_CB) CR_CB->WindowCB(DataChangeT);
      Unmap(Watch);
      return Okay;
    }
  }
  return Continue;
}//--- DataPrompt::ChildCB ---//



//---------------------------------------------------------------------------
//
//	GraphicBox Class
//
//---------------------------------------------------------------------------
void GraphicBox::WorldToWindow(float xf, float yf, int &xi, int &yi){
#ifdef __ActiveCC
  xi=int((xf-xo)/dx+leftM+0.5);
  yi=int((yo-yf)/dy+Height()-bottomM);
#else
  xi=int(Max(Min((xf-xo)/dx+leftM+0.5,double(INT_MAX)),double(INT_MIN)));
  yi=int(Max(Min(double((yo-yf)/dy+Height()-bottomM),
	 double(INT_MAX)),double(INT_MIN)));
#endif
}//--- GraphicBox::WorldToWindow ---//

void GraphicBox::Refresh(Direction BarToSet){
  if (BarToSet==Horizontal || BarToSet==DefaultDirection) SetHBar();
  if (BarToSet==Vertical   || BarToSet==DefaultDirection) SetVBar();
  if (mapped){
  Cursor parentcursor=Parent()->SetCursor(Watch);
    Clear(leftM,topM,Width()-rightM,Height()-bottomM);
    WindowCB(ExposeT);
    Parent()->SetCursor(parentcursor);
  }
}//--- GraphicBox::Refresh ---//

void GraphicBox::DecOneXUnit(){
  if (xo>Xmin){
    xo -= Max(W()/10,1)*dx;
    if (xo<Xmin) xo=Xmin;
    Refresh(Horizontal);
  }
}//--- GraphicBox::DecOneXUnit ---//


void GraphicBox::IncOneXUnit(){
float max_xo=Xmax-W()*dx;
  if (xo<max_xo){
    xo+=Max(W()/10,1)*dx;
    if (xo>max_xo) xo=max_xo;
    Refresh(Horizontal);
  }
}//--- GraphicBox::IncOneXUnit ---//


void GraphicBox::DecOneXPage(){
  if (xo>Xmin){
    xo-=W()*dx;
    if (xo<Xmin) xo=Xmin;
    Refresh(Horizontal);
  }
}//--- GraphicBox::DecOneXPage ---//


void GraphicBox::IncOneXPage(){
float max_xo=Xmax-W()*dx;
  if (xo<max_xo){
    xo+=W()*dx;
    if (xo>max_xo) xo=max_xo;
    Refresh(Horizontal);
  }
}//--- GraphicBox::IncOneXPage ---//


void GraphicBox::ScrollXWorld(){
float new_xo=hbar->P1()*(Xmax-Xmin)+Xmin;
  if (new_xo!=xo){
    xo = new_xo;
    Refresh(Horizontal);
  }
}//--- GraphicBox::ScrollXWorld ---//


void GraphicBox::DecOneYUnit(){
  if (yo>Ymin){
    yo -= Max(H()/10,1)*dy;
    if (yo<Ymin) yo=Ymin;
    Refresh(Vertical);
  }
}//--- GraphicBox::DecOneYUnit ---//


void GraphicBox::IncOneYUnit(){
float max_yo=Ymax-H()*dy;
  if (yo<max_yo){
    yo+=Max(H()/10,1)*dy;
    if (yo>max_yo) yo=max_yo;
    Refresh(Vertical);
  }
}//--- GraphicBox::IncOneYUnit ---//


void GraphicBox::DecOneYPage(){
  if (yo>Ymin){
    yo-=H()*dx;
    if (yo<Ymin) yo=Ymin;
    Refresh(Vertical);
  }
}//--- GraphicBox::DecOneYPage ---//


void GraphicBox::IncOneYPage(){
float max_yo=Ymax-H()*dy;
  if (yo<max_yo){
    yo+=H()*dy;
    if (yo>max_yo) yo=max_yo;
    Refresh(Vertical);
  }
}//--- GraphicBox::IncOneYPage ---//


void GraphicBox::ScrollYWorld(){
float new_yo=(1-vbar->P2())*(Ymax-Ymin)+Ymin;
  if (new_yo!=yo){
    yo = new_yo;
    Refresh(Vertical);
  }
}//--- GraphicBox::ScrollYWorld ---//


GraphicBox::GraphicBox(Window  &w, int x1, int y1, int x2, int y2, int bw,
		       float Xo, float Yo, float Dx, float Dy,
		       Boolean scrollbar):
Window(w,x1,y1,x2,y2){
  SetBorder(bw,0);
  hbar=vbar=0;
  barW=0;
  leftM=topM=0;
  rightM=bottomM=(scrollbar)?SB_Width+2:0;
  SetDrawW(Xo,Yo,Dx,Dy);
  if (scrollbar){
    hbar=new ScrollBar(*this,-1,Height()-SB_Width-1,
			Width()-SB_Width-2,Height()-1);
    vbar=new ScrollBar(*this,Width()-SB_Width-1,-1,
			Width()-1,Height()-SB_Width-2);
    barW=new Window(*this,hbar->X2()+2,vbar->Y2()+2,Width()-1,Height()-1);
    barW->SetBorder(0,0);
    barW->SetRubber(RightBottomR);
    barW->Map();
    SetHBar(False);
    SetVBar(False);
  }
  Map();
}//--- GraphicBox::GraphicBox ---//


GraphicBox::~GraphicBox(){
  if (hbar) delete hbar;
  if (vbar) delete vbar;
  if (barW) delete barW;
}//--- GraphicBox::~GraphicBox ---//


Status GraphicBox::WindowCB(CBTag tag){
  switch (tag){
    case ResizeT:
      SetWorld(Xmin,Ymin,Xmax,Ymax,False);
      break;
    case ExposeT:
    case SelectT:
      return Parent()->ChildCB(this,tag);
    case ActivateT:
      if (ActivatedWindow()==this){
      char  cbuf[80];
      int   tx1,ty1,tx2,ty2,mx,my,tw,th;
      float x,y;
        MouseXY(mx,my);
	WindowToWorld(mx,my,x,y);
	sprintf(cbuf,"%g,%g",x,y);
	tw = TextWidth(cbuf)+5;
	th = TextHeight(cbuf)+5;
	tx2= Min(mx+tw+5,Width());
        tx1= Max(0,tx2-tw-2);
        ty1= Max(my-(th+5),0);
	ty2= Min(ty1+th,Height());
	SaveImage(tx1,ty1,tx2,ty2);
        Window::FillRectangle(tx1,ty1,tx2,ty2,background,border);
	Window::OutTextXY(tx1+4,ty1+2,cbuf,border);
        while (GetEvent()!=ButtonRelease);
	RestoreImage(tx1,ty1);
      }
      break;
  }
  return Continue;
}//--- GraphicBox::WindowCB ---//


Status GraphicBox::ChildCB(Window *w, CBTag tag){
  if (w==hbar){
    switch (tag){
      case DecT:     DecOneXUnit();  return Parent()->ChildCB(this,HDecT);
      case IncT:     IncOneXUnit();  return Parent()->ChildCB(this,HIncT);
      case PageDecT: DecOneXPage();  return Parent()->ChildCB(this,HPageDecT);
      case PageIncT: IncOneXPage();  return Parent()->ChildCB(this,HPageIncT);
      case ScrollT:  ScrollXWorld(); return Parent()->ChildCB(this,HScrollT);
      case ResizeT:  SetHBar(False); break;
    }
  }
  else if (w==vbar){
    switch (tag){
      case IncT:     DecOneYUnit();  return Parent()->ChildCB(this,VDecT);
      case DecT:     IncOneYUnit();  return Parent()->ChildCB(this,VIncT);
      case PageIncT: DecOneYPage();  return Parent()->ChildCB(this,VPageDecT);
      case PageDecT: IncOneYPage();  return Parent()->ChildCB(this,VPageIncT);
      case ScrollT:  ScrollYWorld(); return Parent()->ChildCB(this,VScrollT);
      case ResizeT:  SetVBar(False); break;
    }
  }
  else if (w==barW){
    switch (tag){
      case ExposeT:  barW->Rectangle(3,3,barW->Width()-3,barW->Height()-3);
		     break;
    }
  }
  return Continue;
}//--- GraphicBox::ChildCB ---//


fstream &GraphicBox::IO(fstream &fs){
  if (!(fs|xo|yo|dx|dy|Xmin|Ymin|Xmax|Ymax))
    cerr << "GraphicBox::IO Error!" << endl;
  if (iostate_==ios::in) SetWorld(Xmin,Ymin,Xmax,Ymax,True);
  return fs;
}//--- GraphicBox::IO ---//


void GraphicBox::SetMargin(int lM, int tM, int rM, int bM){
  leftM=lM; topM=tM;  rightM=rM; bottomM=bM;
  Expose();
}//--- GraphicBox::SetMargin ---//


void GraphicBox::SetWorld(float xmin,float ymin,float xmax,float ymax,
			  Boolean move_resize){
  if (hbar){
    Xmin=Min(xo,xmin);
    Ymin=Min(yo,ymin);
    Xmax=Max(xo+W()*dx,xmax);
    Ymax=Max(yo+H()*dy,ymax);
    SetHBar(move_resize);
    SetVBar(move_resize);
  }
  else{
    xo=Xmin=xmin;  yo=Ymin=ymin;  Xmax=xmax;  Ymax=ymax;
    dx=(Xmax-Xmin)/W();
    dy=(Ymax-Ymin)/H();
    if (move_resize) Refresh();
  }
}//--- GraphicBox::SetWorld ---//


void GraphicBox::SetDrawW(float Xo, float Yo, float Dx, float Dy){
  if (hbar){
    xo=Max(Xo,Xmin);  yo=Max(Yo,Ymin);  dx=Dx;  dy=Dy;
    Refresh();
  }
  else
    SetWorld(Xo,Yo,Xo+Dx*W(),Yo+Dy*H());
}//--- GraphicBox::SetDrawW ---//


void GraphicBox::SetHBar(Boolean move_resize){
  if (hbar){
  float wx=Xmax-Xmin,p1=(xo-Xmin)/wx,p2=dx*(Width()-(leftM+rightM))/wx+p1;
    hbar->SetP1P2(p1,p2,move_resize);
  }
}//--- GraphicBox::SetHBar ---//


void GraphicBox::SetVBar(Boolean move_resize){
  if (vbar){
  float wy=Ymin-Ymax,p2=(yo-Ymax)/wy,p1=p2+dy*(Height()-(topM+bottomM))/wy;
    vbar->SetP1P2(p1,p2,move_resize);
  }
}//--- GraphicBox::SetVBar ---//


void GraphicBox::MapGadgets(){
  if (hbar){
    hbar->Map();
    vbar->Map();
    barW->Map();
  }
}//--- GraphicBox::MapGadgets ---//


void GraphicBox::UnmapGadgets(){
  if (hbar){
    hbar->Unmap();
    vbar->Unmap();
    barW->Unmap();
  }
}//--- GraphicBox::UnmapGadgets ---//
    

void GraphicBox::Point(float x, float y, Color c){
int xi,yi;
  WorldToWindow(x,y,xi,yi);
  Window::Point(xi,yi,c);
}//--- GraphicBox::Point ---//


void GraphicBox::Points(int n, float *xy, Color c){
int imax=2*n,*xyi=new int[imax],ix,iy;
  for(ix=0,iy=1;ix<imax;ix+=2,iy+=2)
    WorldToWindow(xy[ix],xy[iy],xyi[ix],xyi[iy]);
  Window::Points(n,xyi,c);
  delete xyi;
}//--- GraphicBox::Point ---//


void GraphicBox::Line(float x1,float y1,float x2,float y2,Color c){
int xi,yi,xj,yj;
  WorldToWindow(x1,y1,xi,yi);
  WorldToWindow(x2,y2,xj,yj);
  Window::Line(xi,yi,xj,yj,c);
}//--- GraphicBox::Line ---// 


void GraphicBox::Lines(int n, float *xy, Color c){
int imax=4*n,*xyi=new int[imax],i0,i1,i2,i3;
  for(i0=0,i1=1,i2=2,i3=3;i0<imax;i0+=4,i1+=4,i2+=4,i3+=4){
    WorldToWindow(xy[i0],xy[i1],xyi[i0],xyi[i1]);
    WorldToWindow(xy[i2],xy[i3],xyi[i2],xyi[i3]);
  }
  Window::Lines(n,xyi,c);
  delete xyi;
}//--- GraphicBox::Lines ---//


void GraphicBox::FillRectangle(float x1,float y1,float x2,float y2,
				Color fc, Color dc){
int xi,yi,xj,yj;
  WorldToWindow(x1,y1,xi,yi);
  WorldToWindow(x2,y2,xj,yj);
  Window::FillRectangle(xi,yi,xj,yj,fc,dc);
}//---GraphicBox::FillRectangle ---//


void GraphicBox::Rectangle(float x1,float y1,float x2,float y2,Color c){
int xi,yi,xj,yj;
  WorldToWindow(x1,y1,xi,yi);
  WorldToWindow(x2,y2,xj,yj);
  Window::Rectangle(xi,yi,xj,yj,c);
}//---GraphicBox::Rectangle ---//


void GraphicBox::Polygon(int n, float *fpoint, Color c){
int i,j,imax=2*n,*ipoint=new int[imax];
  for (i=0,j=1;i<imax;i+=2,j+=2)
    WorldToWindow(fpoint[i],fpoint[j],ipoint[i],ipoint[j]);
  Window::Polygon(n,ipoint,c);
  delete ipoint;
}//---GraphicBox::Polygon ---//


void GraphicBox::FillPolygon(int n, float *fpoint, Color fc, Color dc){
int i,j,imax=2*n,*ipoint=new int[imax];
  for (i=0,j=1;i<imax;i+=2,j+=2)
    WorldToWindow(fpoint[i],fpoint[j],ipoint[i],ipoint[j]);
  Window::FillPolygon(n,ipoint,fc,dc);
  delete ipoint;
}//---GraphicBox::FillPolygon ---//


void GraphicBox::Ellipse(float x,float y,float xr, float yr, Color c){
int xi,yi,xri,yri;
  WorldToWindow(x,y,xi,yi);
  WindowLength(xr,yr,xri,yri);
  Window::Ellipse(xi,yi,xri,yri,c);
}//---GraphicBox::Ellipse  ---//


void GraphicBox::FillEllipse(float x,float y,float xr,float yr,
			     Color fc, Color dc){
int xi,yi,xri,yri;
  WorldToWindow(x,y,xi,yi);
  WindowLength(xr,yr,xri,yri);
  Window::FillEllipse(xi,yi,xri,yri,fc,dc);
}//--- GraphicBox::FillEllipse ---//


void GraphicBox::Surface(int nx,int ny,float *xy,Color dc){
int i0,i1,imax=nx*ny*2,*xyi=new int[imax];
  for(i0=0,i1=1;i0<imax;i0+=2,i1+=2)
    WorldToWindow(xy[i0],xy[i1],xyi[i0],xyi[i1]);
  Window::Surface(nx,ny,xyi,dc);
  delete xyi;
}//--- GraphicBox::Surface ---//


void GraphicBox::FillSurface(int nx,int ny,float *xy,int *corner_depth,
			     Color fc,Color dc){
int i0,i1,imax=nx*ny*2,*xyi=new int[imax];
  for(i0=0,i1=1;i0<imax;i0+=2,i1+=2)
    WorldToWindow(xy[i0],xy[i1],xyi[i0],xyi[i1]);
  Window::FillSurface(nx,ny,xyi,corner_depth,fc,dc);
  delete xyi;
}//--- GraphicBox::FillSurface ---//



void GraphicBox::OutTextXY(float x,float y, const char *s,Color c){
int xi,yi;
  WorldToWindow(x,y,xi,yi);
  Window::OutTextXY(xi,yi,s,c);
}//--- GraphicBox::OutTextXY ---//



//---------------------------------------------------------------------------
//
//	GraphBox Class
//
//---------------------------------------------------------------------------
void GraphBox::DrawTexts(){
int x=Width()/2,y=Height();
  hjustify=CenterJ;
  vjustify=TopJ;
  textdir =Horizontal;
  OutTextXY(x,0,title,Black);
  OutTextXY(x,y-TextHeight(hlabel)-3,hlabel,Black);
  textdir =Vertical;
  hjustify=LeftJ;
  vjustify=CenterJ;
  OutTextXY(0,y/2,vlabel);
}//--- GraphBox::DrawTexts ---//


void GraphBox::DrawHAxis(){
char   s[80];
int    i,imax,x,y1,y2;
double x1,dx,dv;
float  v,Xmin,Ymin,Xmax,Ymax;
  GB.GetWorld(Xmin,Ymin,Xmax,Ymax);
  y1=GB.Y2()+1;
  y2=y1+4;
  x1=GB.X1()+0.5;
  imax=num_hticks-1;
  dx=double(GB.Width())/imax;
  dv=(Xmax-Xmin)/imax;
  textdir=Horizontal;
  vjustify=TopJ;
  for (i=0;i<num_hticks;i++){
    x=int(i*dx+x1);
    v=i*dv+Xmin;
    sprintf(s,"%g",v);
    Window::Line(x,y1,x,y2,Black);
    hjustify=(i==0)?LeftJ:(i<imax)?CenterJ:RightJ;
    if (i==imax) x+=2;
    OutTextXY(x,y2,s,Black);
  }
}//--- GraphBox::DrawHAxis ---//


void GraphBox::DrawVAxis(){
char   s[80];
int    i,imax,y,xL,x1,x2;
double y2,dy,dv;
float  v,Xmin,Ymin,Xmax,Ymax;
  GB.GetWorld(Xmin,Ymin,Xmax,Ymax);
  x2=GB.X1()-1;
  x1=x2-4;
  xL=x1-3;
  y2=GB.Y2();
  imax=num_vticks-1;
  dy=double(GB.Height())/imax;
  dv=(Ymax-Ymin)/imax;
  textdir=Vertical;
  hjustify=RightJ;
  for (i=0;i<num_vticks;i++){
    y=int(y2-i*dy);
    v=i*dv+Ymin;
    sprintf(s,"%g",v);
    Window::Line(x1,y,x2,y,Black);
    vjustify=(i==0)?BottomJ:(i<imax)?CenterJ:TopJ;
    if (i==imax) y-=2;
    OutTextXY(xL,y,s,Black);
  }
}//--- GraphBox::DrawVAxis ---//


void GraphBox::ExposeCB(){
  DrawTexts();
  DrawHAxis();
  DrawVAxis();
}//--- GraphBox::ExposeCB ---//


void GraphBox::SelectCB(){
static char *prompts[]={"Title%s",
		 "X Label%s",
		 "Y Label%s",
		 "X min%g",
		 "X max%g",
		 "X ticks%d",
		 "Y min%g",
		 "Y max%g",
		 "Y ticks%d",
		 "Bar <Y/N>%c",
		 ""
		};
int	xo=20,yo=20;
char	oldbg,bg=(bargraph)?'Y':'N';
float	Xmin,Ymin,Xmax,Ymax,x1,y1,x2,y2;
Boolean refreshgraph=False;
Window	*wp=this,*RP=Root();
  do{
    xo += wp->X1();
    yo += wp->Y1();
    wp  = wp->Parent();
  } while (wp!=RP);
  GB.GetWorld(Xmin,Ymin,Xmax,Ymax);
  x1=Xmin; y1=Ymin, x2=Xmax, y2=Ymax; oldbg=bg;
  if (GetData(*RP,0,xo,yo,xo+300,yo+450,"Graph Characteristic",
	      prompts,title,hlabel,vlabel,&Xmin,&Xmax,&num_hticks,
	      &Ymin,&Ymax,&num_vticks,&bg)==Okay){
    if (x1!=Xmin || y1!=Ymin || x2!=Xmin || y2!=Ymin || oldbg!=bg){
      bargraph=(bg=='Y' || bg=='y')?True:False;
      GB.SetWorld(Xmin,Ymin,Xmax,Ymax,False);
      refreshgraph=True;
    }
    num_hticks=Max(2,num_hticks);
    num_vticks=Max(2,num_vticks);
    Refresh(True,refreshgraph);
    Parent()->ChildCB(this,DataChangeT);
  }
}//--- GraphBox::SelectCB ---//


void GraphBox::GB_ExposeCB(){
int axis[6];
  axis[0]=axis[2]=GB.X1();  axis[1]=GB.Y1();
  axis[3]=axis[5]=GB.Y2();  axis[4]=GB.X2();
  DrawOverSW();
  Polygon(3,axis,Black);
  DrawOverSW(False);
  if (xy)
    DrawGraph();
  else
    Parent()->ChildCB(this,ExposeT);
}//--- GraphBox::GB_ExposeCB ---//


GraphBox::GraphBox(Window &w, int x1, int y1, int x2, int y2):
Window(w,x1,y1,x2,y2),
GB(*this,40,20,Width()-10,Height()-40,0,0,0,1,1){
  GB.SetRubber(AllSideR);
  GB.SetWorld(0,0,100,100);
  bargraph=False;
  title[0]=hlabel[0]=vlabel[0]=0;
  title[LB_SIZE-1]=hlabel[LB_SIZE-1]=vlabel[LB_SIZE-1]=0;
  xy=0;
  num_points=0;
  num_hticks=num_vticks=3;
  Map();
}//--- GraphBox::GraphBox ---//


Status GraphBox::WindowCB(CBTag tag){
  switch (tag){
    case ExposeT: ExposeCB(); break;
    case SelectT: SelectCB(); break;
  }
  return Continue;
}//--- GraphBox::WindowCB ---//


Status GraphBox::ChildCB(Window *w, CBTag tag){
  if (w==&GB){
    switch (tag){
      case ExposeT: GB_ExposeCB(); break;
      case SelectT: Parent()->ChildCB(this,SelectT); break;
    }
  }
  return Continue;
}//--- GraphBox::ChildCB ---//


fstream &GraphBox::IO(fstream &fs){
int HL,VL,TL;
  if (iostate_==ios::out){
    HL=strlen(hlabel);
    VL=strlen(vlabel);
    TL=strlen(title);
  }
  if (!(fs|bargraph|num_hticks|num_vticks|HL|VL|TL))
    cerr << "GraphBox::IO Error!" << endl;
  DataIO(fs,hlabel,HL); hlabel[HL]=0;
  DataIO(fs,vlabel,VL); vlabel[VL]=0;
  DataIO(fs, title,TL);  title[TL]=0;
  return GB.IO(fs);
}//--- GraphBox::IO ---//

void GraphBox::ToBarGraph(Boolean refresh){
  bargraph=True;
  Refresh(False,refresh);
}//--- GraphBox::ToBarGraph ---//


void GraphBox::ToLineGraph(Boolean refresh){
  bargraph=False;
  Refresh(False,refresh);
}//--- GraphBox::ToLineGraph ---//

void GraphBox::Clear(){
int h=GB.Height(),axis[6];
  axis[0]=0;  axis[1]=0;           axis[2]=0;
  axis[3]=h;  axis[4]=GB.Width();  axis[5]=h;
  GB.Clear();
  GB.Window::Polygon(3,axis,Black);
}//--- GraphBox::Clear ---//


void GraphBox::SetWorld(float x1,float y1,float x2,float y2,Boolean refresh){
  GB.SetWorld(x1,y1,x2,y2,False);
  Refresh(refresh,refresh);
}//--- GraphBox::SetWorld ---//


void GraphBox::SetTitle(char *s, Boolean refresh){
  strncpy(title,s,LB_SIZE);
  title[LB_SIZE-1]=0;
  Refresh(refresh,False);
}//--- GraphBox::SetTitle ---//


void GraphBox::GetTitle(char *s){
  strcpy(s,title);
}//--- GraphBox::GetTitle ---//


void GraphBox::SetHLabel(char *s, Boolean refresh){
  strncpy(hlabel,s,LB_SIZE);
  hlabel[LB_SIZE-1]=0;
  Refresh(refresh,False);
}//--- GraphBox::SetHLabel ---//


void GraphBox::GetHLabel(char *s){
  strcpy(s,hlabel);
}//--- GraphBox::GetHLabel ---//

void GraphBox::SetVLabel(char *s, Boolean refresh){
  strncpy(vlabel,s,LB_SIZE);
  vlabel[LB_SIZE-1]=0;
  Refresh(refresh,False);
}//--- GraphBox::SetVLabel ---//


void GraphBox::GetVLabel(char *s){
  strcpy(s,vlabel);
}//--- GraphBox::GetVLabel ---//


void GraphBox::SetTicks(int num_ht, int num_vt, Boolean refresh){
  num_hticks=Max(2,num_ht);  num_vticks=Max(2,num_vt);
  Refresh(refresh,False);
}//--- GraphBox::SetTicks ---//


void GraphBox::SetData(int num_pt, float *d, Boolean refresh){
  num_points=num_pt;  xy=d;
  Refresh(False,refresh);
}//--- GraphBox::SetData ---//


void GraphBox::Refresh(Boolean refreshLabels, Boolean refreshGraph){
  if (refreshLabels){
    Window::Clear();
    DrawTexts();
    DrawHAxis();
    DrawVAxis();
  }
  if (refreshGraph){
    GB.Expose();
  }
}//--- GraphBox::Refresh ---//


void GraphBox::DrawGraph(Boolean clearbackground){
  if (clearbackground) Clear();
  if (bargraph){
  int i,imax=num_points*2;
    for (i=0;i<imax;i+=2)
      GB.Line(xy[i],0,xy[i],xy[i+1],Black);
  }
  else{
  float *p=xy;
  int N=num_points-1,DN=100,n=Min(DN,N);
    while(n>0){
      GB.Polygon(n+1,p,Black);
      p+=2*n;
      N-=DN;
      n =Min(DN,N);
    }
  }
}//--- GraphBox::DrawGraph ---//


void GraphBox::DrawOnePoint(int i, Color c){
  if (i>0){
  float x1,y1,x2,y2;
  int   n=i*2,m=n+1;
  Color co;
    if (bargraph){
      x1=x2=xy[n];  y1=0;  y2=xy[m];
    }
    else{
      x1=xy[n-2];  x2=xy[n];   y1=xy[m-2];  y2=xy[m];
    }
    if (c==DefaultColor) co=Black; else co=c;
    GB.Line(x1,y1,x2,y2,co);
  }
}//--- GraphBox::DrawOnePoint ---//


void GraphBox::AddOnePoint(float x, float y, Boolean draw_it){
int i=2*num_points;
  xy[i]=x;  xy[i+1]=y;
  if (draw_it) DrawOnePoint(num_points++);
}//--- GraphBox::AddOnePoint ---//


void GraphBox::SubOnePoint(Boolean erase_it){
  if (num_points>0 && erase_it)
    DrawOnePoint(--num_points,background);
}//--- GraphBox::SubOnePoint ---//



//---------------------------------------------------------------------------
//
//	DrawingBoard Class
//
//---------------------------------------------------------------------------
int DrawingBoard::Mx;
int DrawingBoard::My;
int DrawingBoard::tx1;
int DrawingBoard::ty1;
int DrawingBoard::tx2;
int DrawingBoard::ty2;
int DrawingBoard::tw;
int DrawingBoard::th;
Button DrawingBoard::ButtonFirstPressed;
Cursor DrawingBoard::oldcursor;

DrawingBoard::DrawingBoard(Window &w, int x1, int y1, int x2, int y2,
			   float xo, float yo, float dx, float dy):
GraphicBox(w,x1,y1,x2,y2,1,xo,yo,dx,dy,True){
  Dxo=10*dx;
  Dyo=10*dy;
  Dx =20*dx;
  Dy =20*dy;
  Dx1=Dy1=Dx2=Dy2=0;
  hjustify=LeftJ;
  vjustify=TopJ;
}//---  DrawingBoard::DrawingBoard ---//

void DrawingBoard::FeedBack(Boolean showmeter){
Color XorColor;
  if(Mono())
    XorColor=White;
  else
    if (background!=Black)
      XorColor=background;
    else
      XorColor=White;
  RestoreImage(tx1,ty1);
  SetLineStyle(DottedL);
  SetWriteMode(XorMode);
  Rectangle(Dx1,Dy1,Dx2,Dy2,XorColor);
  SetLineStyle(SolidL);
  SetWriteMode(SetMode);
  if (showmeter){
  char cbuf[80];
  const char *format;
    format = (Dx1==Dx2 && Dy1==Dy2)?"(%g,%g)":"%g,%g";
    if (MouseButton()==LeftB)
      sprintf(cbuf,format,Dx2,Dy2);
    else
      sprintf(cbuf,format,Dx2-Dx1,Dy2-Dy1);
    tw = TextWidth(cbuf)+5;
    th = TextHeight(cbuf)+5;
    tx2= Min(Mx+tw+5,Width());
    tx1= Max(0,tx2-tw-2);
    ty1= Max(My-(th+5),0);
    ty2= Min(ty1+th,Height());
    SaveImage(tx1,ty1,tx2,ty2);
    Window::FillRectangle(tx1,ty1,tx2,ty2,background,border);
    Window::OutTextXY(tx1+4,ty1+2,cbuf,border);
  }
}//--- DrawingBoard::FeedBack ---//


void DrawingBoard::ActivateCB(){
int leftM,topM,rightM,bottomM;
  if (cursor == Cross) SetCursor(Arrow);
  oldcursor = SetCursor(Cross);
  GetMargin(leftM,topM,rightM,bottomM);
  GrabPointer(0);
  MouseAt(Mx,My);
  WindowToWorld(Mx-1,My-1,Dx1,Dy1);
  DrawXY(Dx1,Dy1);
  Dx2=Dx1;  Dy2=Dy1;
  FeedBack();
  Parent()->ChildCB(this,ActivateT);
}//--- DrawingBoard::ActivatedCB ---//


void DrawingBoard::ButtonPressCB(){
float x,y;
  if (ActivatedWindow()==this){
    MouseAt(Mx,My);
    WindowToWorld(Mx-1,My-1,x,y);
    x = Max(x,Dxo);
    y = Max(y,Dyo);
    DrawXY(x,y);
    if (x!=Dx2 || y!=Dy2){
      FeedBack(False);
      Parent()->ChildCB(this,ButtonPressT);
      Dx2=x;  Dy2=y;
      FeedBack();
      Parent()->ChildCB(this,ButtonPressT);
    }
  }
}//--- DrawingBoard::ButtonPressCB ---//

void DrawingBoard::ButtonReleaseCB(){
  if (ActivatedWindow()==this){
    UngrabPointer();
    MapGadgets();
    SetCursor(oldcursor);
    FeedBack(False);
    Parent()->ChildCB(this,ButtonReleaseT);
  }
}//--- DrawingBoard::ButtonReleaseCB ---//


Status DrawingBoard::WindowCB(CBTag tag){
  switch (tag){
    case ActivateT:      ActivateCB();      break;
    case ButtonPressT:   ButtonPressCB();   break;
    case ButtonReleaseT: ButtonReleaseCB(); break;
    default: return GraphicBox::WindowCB(tag);
  }
  return Continue;
}//--- DrawingBoard::WindowCB ---//


fstream &DrawingBoard::IO(fstream &fs){
  if (!(fs|Mx|My|tx1|ty1|tx2|ty2|tw|th|Dxo|Dyo|Dx|Dy|Dx1|Dy1|Dx2|Dy2))
    cerr << "DrawingBoard::IO Error!" << endl;
  GraphicBox::IO(fs);
  return fs;
}//--- DrawingBoard::IO ---//


void DrawingBoard::GetData(float &dx1,float &dy1,float &dx2,float &dy2){
  dx1=Dx1;  dy1=Dy1; dx2=Dx2;  dy2=Dy2;
}//--- DrawingBoard::GetData ---//


void DrawingBoard::SetDrawXY(float dxo, float dyo, float dx, float dy){
  Dxo=dxo;  Dyo=dyo;  Dx=dx;  Dy=dy;
}//--- DrawingBoard::SetDrawXY ---//


void DrawingBoard::GetDrawXY(float &dxo, float &dyo, float &dx, float &dy){
  dxo=Dxo;  dyo=Dyo;  dx=Dx;  dy=Dy;
}//--- DrawingBoard::GetDrawXY ---//


void DrawingBoard::DrawXY(float &x, float &y){
  x = int((x-Dxo)/Dx+0.5)*Dx+Dxo;
  y = int((y-Dyo)/Dy+0.5)*Dy+Dyo;
}//--- DrawingBoard::DrawXY ---//


void DrawingBoard::DrawGrid(){
int   W=Width(),H=Height(),i,I;
float x,y,xmin,xmax,ymin,ymax,dxo=Dxo,dyo=Dyo,*p;
  Dxo=Dyo=0;
  WindowToWorld(0,H,xmin,ymin);
  WindowToWorld(W,0,xmax,ymax);
  DrawXY(xmin,ymin);  xmin-=Dx;  ymin-=Dy;
  DrawXY(xmax,ymax);  xmax+=Dx;  ymax+=Dy;
  I = 2*int((ymax-ymin)/Dy+0.5);
  p = new float[I];
  for (x=xmin;x<xmax;x+=Dx){
    for (y=ymin,i=0;i<I;i+=2,y+=Dy){
      p[i]=x;  p[i+1]=y;
    }
    Points(I/2,p,Black);
  }
  delete p;
  Dxo=dxo;
  Dyo=dyo;
}//--- DrawingBoard::DrawGrid ---//



//---------------------------------------------------------------------------
//
//	Graphic3DBox Class
//
//---------------------------------------------------------------------------
float Graphic3DBox::va;
float Graphic3DBox::vb;
float Graphic3DBox::vc;
float Graphic3DBox::vd;
	
static char *title    ="3D Graphic Box Data",
	    *prompts[]={"Projection Point%g",
			"Magnification%g",
			"Rotation Center - X%g",
			"               - Y%g",
			"               - Z%g",
			"Rotation Angle  - X%g",
			"               - Y%g",
			"               - Z%g",
			"Projection Plane - X1%g",
			"                - Y1%g",
			"                - X2%g",
			"                - Y2%g",
			""
		       };

void Graphic3DBox::DataChangedCB(){
  Projection(ProjP);
  Magnification(Mf);
  SetOrigin(Xo,Yo,Zo);
  SetRotAngles(Rx,Ry,Rz);
  SetWorld(va,vb,vc,vd);
  Clear();
  WindowCB(ExposeT);
}//--- Graphic3DBox::DataChangedCB ---//

void Graphic3DBox::SelectCB(){
float	v1=ProjP,v2=Mf,v3=Xo,v4=Yo,v5=Zo,v6=Rx,v7=Ry,v8=Rz,Va,Vb,Vc,Vd;
int	x1=20,y1=20,x2,y2;
Window	*wp=this,*RP=Root();
  do{
    x1 += wp->X1();
    y1 += wp->Y1();
    wp  = wp->Parent();
  } while (wp!=RP);
  x2=x1+360;
  y2=y1+450;
  GetWorld(va,vb,vc,vd);
  Va=va; Vb=vb; Vc=vc; Vd=vd;
  if (GetData(*RP,this,x1,y1,x2,y2,title,prompts,&ProjP,&Mf,&Xo,&Yo,&Zo,
  	      &Rx,&Ry,&Rz,&va,&vb,&vc,&vd)==Cancel){
    ProjP=v1; Mf=v2; Xo=v3; Yo=v4; Zo=v5; Rx=v6; Ry=v7; Rz=v8;
    va=Va; vb=Vb; vc=Vc; vd=Vd;
    DataChangedCB();
  }
  else
    Parent()->ChildCB(this,DataChangeT);
}//--- Graphic3DBox::SelectCB ---//


Graphic3DBox::Graphic3DBox(Window &w,int x1,int y1,int x2,int y2,float projp,
			   float mf,float xo,float yo,float zo,float rx,
			   float ry,float rz,Boolean scrollbar):
GraphicBox(w,x1,y1,x2,y2,1,0,0,0.1,0.1,scrollbar){
  ProjP=projp; Mf=mf; Xo=xo; Yo=yo; Zo=zo; Rx=rx; Ry=ry; Rz=rz;
  RotationMatrix();
}//--- Graphic3DBox::Graphic3DBox ---//


Status Graphic3DBox::WindowCB(CBTag tag){
  switch (tag){
    case SelectT:	SelectCB();		break;
    case DataChangeT:	DataChangedCB();	break;
    default:		return GraphicBox::WindowCB(tag);
  }
  return Continue;
}//--- Graphic3DBox::WindowCB ---//


fstream &Graphic3DBox::IO(fstream &fs){
  if (!(fs|ProjP|Mf|Xo|Yo|Zo|Rx|Ry|Rz))
    cerr << "Graphic3DBox::IO Error!" << endl;
  GraphicBox::IO(fs);
  if (iostate_==ios::in){
    SetRotAngles(Rx,Ry,Rz);
    Clear();
    WindowCB(ExposeT);
  }
  return fs;
}//--- Graphic3DBox::IO ---//


void Graphic3DBox::Projection(float p){
  ProjP=p;
}//--- Graphic3DBox::Projection ---//


void Graphic3DBox::Magnification(float mf){
  Mf=mf;
}//--- Graphic3DBox::Magnification ---//


void Graphic3DBox::SetRotAngles(float rx, float ry, float rz){
  Rx=rx;  Ry=ry;  Rz=rz;  RotationMatrix();
}//--- Graphic3DBox::SetRotAngles ---//		


void Graphic3DBox::IncRotAngles(float rx, float ry, float rz){
  Rx+=rx;  Ry+=ry;  Rz+=rz;  RotationMatrix();
}//--- Graphic3DBox::IncRotAngles ---//		


void Graphic3DBox::RotationMatrix(){
double t,
  temp = 3.141592653/180.0,
  cosx = cos(t=temp*Rx),
  sinx = sin(t),
  cosy = cos(t=temp*Ry),
  siny = sin(t),
  cosz = cos(t=temp*Rz),
  sinz = sin(t),
  sinx_siny = sinx*siny,
  cosx_siny = cosx*siny;
  Rmatrix[0][0] = cosy*cosz;
  Rmatrix[0][1] = -cosy*sinz;
  Rmatrix[0][2] = -siny;
  Rmatrix[1][0] = sinx_siny*cosz-cosx*sinz;
  Rmatrix[1][1] = -(sinx_siny*sinz+cosx*cosz);
  Rmatrix[1][2] = sinx*cosy;
  Rmatrix[2][0] = cosx_siny*cosz+sinx*sinz;
  Rmatrix[2][1] = sinx*cosz-cosx_siny*sinz;
  Rmatrix[2][2] = cosx*cosy;
}//--- Graphic3DBox::RotationMatrix ---//		



void Graphic3DBox::Transform(float &x, float &y, float &z){
float xt,yt;
  x=Mf*(x-Xo); y=Mf*(y-Yo);  z=Mf*(z-Zo);
  xt=Rmatrix[0][0]*x+Rmatrix[0][1]*y+Rmatrix[0][2]*z;
  yt=Rmatrix[1][0]*x+Rmatrix[1][1]*y+Rmatrix[1][2]*z;
  z =Rmatrix[2][0]*x+Rmatrix[2][1]*y+Rmatrix[2][2]*z;
  x = xt+Xo;
  y = yt+Yo;
  z+= Zo;
}//--- Graphic3DBox::Transform ---//


void Graphic3DBox::Project(float X3,float Y3,float Z3,float &x2,float &y2){
float a,x3=X3,y3=Y3,z3=Z3;
  Transform(x3,y3,z3);
  a=ProjP-y3;
  a  =(a==0)?1:ProjP/a;
  x2 = x3*a;
  y2 = z3*a;
}//--- Graphic3DBox::Project ---//


void Graphic3DBox::Project(int nx,int ny,float *x,float *y,float *z,
			   float *xy){
int i,j,k=0,K=0;
  for (i=0;i<nx;i++)
    for (j=0; j<ny; j++,k++,K+=2)
      Project(x[i],y[j],z[k],xy[K],xy[K+1]);
}//--- Graphic3DBox::Project ---//



void Graphic3DBox::Project(int n, float *xyz, float *xy){
int i,j,k=3*n;
  for(i=j=0;i<k;i+=3,j+=2)
    Project(xyz[i],xyz[i+1],xyz[i+2],xy[j],xy[j+1]);
}//--- Graphic3DBox::Project ---//



void Graphic3DBox::Box(float x1,float y1,float z1,float x2,float y2,float z2,
			Color co){
float xy[72];
  xy[ 0]=x1; xy[ 1]=y1; xy[ 2]=z1;  xy[ 3]=x1; xy[ 4]=y1; xy[ 5]=z2;
  xy[ 6]=x1; xy[ 7]=y1; xy[ 8]=z2;  xy[ 9]=x1; xy[10]=y2; xy[11]=z2;
  xy[12]=x1; xy[13]=y2; xy[14]=z2;  xy[15]=x1; xy[16]=y2; xy[17]=z1;
  xy[18]=x1; xy[19]=y2; xy[20]=z1;  xy[21]=x1; xy[22]=y1; xy[23]=z1;
  xy[24]=x2; xy[25]=y1; xy[26]=z1;  xy[27]=x2; xy[28]=y1; xy[29]=z2;
  xy[30]=x2; xy[31]=y1; xy[32]=z2;  xy[33]=x2; xy[34]=y2; xy[35]=z2;
  xy[36]=x2; xy[37]=y2; xy[38]=z2;  xy[39]=x2; xy[40]=y2; xy[41]=z1;
  xy[42]=x2; xy[43]=y2; xy[44]=z1;  xy[45]=x2; xy[46]=y1; xy[47]=z1;
  xy[48]=x1; xy[49]=y1; xy[50]=z1;  xy[51]=x2; xy[52]=y1; xy[53]=z1;
  xy[54]=x1; xy[55]=y2; xy[56]=z1;  xy[57]=x2; xy[58]=y2; xy[59]=z1;
  xy[60]=x1; xy[61]=y1; xy[62]=z2;  xy[63]=x2; xy[64]=y1; xy[65]=z2;
  xy[66]=x1; xy[67]=y2; xy[68]=z2;  xy[69]=x2; xy[70]=y2; xy[71]=z2;
  Lines(12,xy,co);
}//--- Graphic3DBox::Box ---//				


void Graphic3DBox::Line(float x1,float y1,float z1,float x2,float y2,float z2,
			Color co){
float X1,Y1,X2,Y2;
  Project(x1,y1,z1,X1,Y1);
  Project(x2,y2,z2,X2,Y2);
  GraphicBox::Line(X1,Y1,X2,Y2,co);
}//--- Graphic3DBox::Line ---//				


void Graphic3DBox::Lines(int n, float *xyz, Color co){
float *xy=new float[n*4];
  Project(2*n,xyz,xy);
  GraphicBox::Lines(n,xy,co);
  delete xy;
}//--- Graphic3DBox::Lines ---//				


void Graphic3DBox::Polygon(int n, float *xyz, Color co){
float *xy=new float[n*2];
  Project(n,xyz,xy);
  GraphicBox::Polygon(n,xy,co);
  delete xy;
}//--- Graphic3DBox::Polygon ---//


void Graphic3DBox::Surface(int nx,int ny,float *x,float *y,float *z,Color co){
float *xy=new float[nx*ny*2];
  Project(nx,ny,x,y,z,xy);
  GraphicBox::Surface(nx,ny,xy,co);
  delete xy;
}//--- Graphic3DBox::Surface ---//


void Graphic3DBox::FillSurface(int nx, int ny, float *x, float *y, float *z,
			       Color fc, Color bc){
float xt[4],yt[4],zt[4],*xy;
int   i,j,corner_depth[4];
  //
  //	z
  //	|
  //	|    y
  //	|  1/------- 2
  //	|  /	   /
  //	| /	  /
  //	|/	 /
  //	|------->x
  //   0	  3
  //
  zt[0]=zt[1]=zt[2]=zt[3]=0;
  corner_depth[0]=0; corner_depth[1]=1;
  corner_depth[2]=2; corner_depth[3]=3;
  xy=new float[nx*ny*2];
  xt[0]=x[0];  xt[1]=x[0];    xt[2]=x[nx-1];  xt[3]=x[nx-1];
  yt[0]=y[0];  yt[1]=y[ny-1]; yt[2]=y[ny-1];  yt[3]=y[0];
  Project(nx,ny,x,y,z,xy);
  for(i=0;i<4;i++) Transform(xt[i],yt[i],zt[i]);
  for(i=0;i<4;i++){
    corner_depth[i]=0;
    for(j=0;j<4;j++) if(yt[i]<yt[j])corner_depth[i]++;
  }
  GraphicBox::FillSurface(nx,ny,xy,corner_depth,fc,bc);
  delete xy;
}//--- Graphic3DBox::FillSurface ---//



//---------------------------------------------------------------------------
//
//	Appication Functions
//
//---------------------------------------------------------------------------
void Message(char *message, Boolean beep){
Window *w=Window::DispatchingWindow();
  if (w) Message(*w,message,beep);
}//--- Message ---//


void Message(Window &w, char *message, Boolean beep){
int	   x1=w.Width()/3,y1=w.Height()/3,x2=2*x1,y2=2*y1;
MessageBox *MB=new MessageBox(w,x1,y1,x2,y2,"Message Box");
  MB->Message(message,0,beep);
  delete MB;
}//--- Message ---//


char *Message(char *message, char *b1, char *b2, Boolean beep){
Window *w=Window::DispatchingWindow();
  return (w)?Message(*w,message,b1,b2,beep):0;
}//--- Message ---//


char *Message(Window &w, char *message, char *b1, char *b2, Boolean beep){
int	   x1=w.Width()/3,y1=w.Height()/3,x2=2*x1,y2=2*y1;
MessageBox *MB=new MessageBox(w,x1,y1,x2,y2,"Message Box");
char	   *button=MB->Message(message,b1,b2,beep);
  delete MB;
  return button;
}//--- Message ---//


char *Message(char *message, char *b1, char *b2, char *b3, Boolean beep){
Window *w=Window::DispatchingWindow();
  return (w)?Message(*w,message,b1,b2,b3,beep):0;
}//--- Message ---//


char *Message(Window &w, char *message, char *b1, char *b2, char *b3,
	      Boolean beep){
int	   x1=w.Width()/4,y1=w.Height()/4,x2=3*x1,y2=3*y1;
MessageBox *MB=new MessageBox(w,x1,y1,x2,y2,"Message Box");
char	   *button=MB->Message(message,b1,b2,b3,beep);
  delete MB;
  return button;
}//--- Message ---//


Status GetFile(char *filename, char *ext, Boolean ToOpen){
  return GetFile(*Window::DispatchingWindow(),100,50,filename,ext,ToOpen);
}//--- GetFile ---// 


Status GetFile(Window &w,int x,int y,char *filename,char *ext, Boolean OP){
FileBox *FB=new FileBox(w,x,y);
Status  status=FB->GetFile(filename,ext,OP);
  delete FB;
  return status;
}//--- GetFile ---//


static Status getdata(Window &pw,Window *CR_CB,int x1,int y1,int x2,int y2,
		      char *title, char **prompts, Boolean *active,
		      void **data){
DataPrompt *DP;
Status	   status;
int	   w=x2-x1,h=y2-y1,W=pw.Width(),H=pw.Height();
  if (x2>W){x2=W; x1=x2-w;}
  if (y2>H){y2=H; y1=y2-h;}
  if (x1<0){x1=0; x2=x1+w;}
  if (y1<0){y1=0; y2=y1+h;}
  DP=new DataPrompt(pw,CR_CB,x1,y1,x2,y2,title,prompts,active,data);
  status=DP->DispatchEvents();
  delete DP;
  return status;
}//--- getdata ---//


Status GetData(Window &pw,Window *CR_CB,int x1,int y1,int x2,int y2,
	       char *title,char **prompts,...){
void	**data;
Status	status;
int	i,num_arg;
va_list	arg_ptr;
Boolean	*active;
  for(num_arg=0;prompts[num_arg][0]!='\0';num_arg++);
  active = new Boolean[num_arg];
  data   = new void*[num_arg];
  va_start(arg_ptr,prompts);
  for(i=0;i<num_arg;i++){
    data[i]   = va_arg(arg_ptr,void*);
    active[i] = True;
  }
  va_end(arg_ptr);
  status=getdata(pw,CR_CB,x1,y1,x2,y2,title,prompts,active,data);
  delete data;
  delete active;
  return status;
}//--- GetData ---//


Status GetActiveData(Window &pw, Window *CR_CB, int x1, int y1, int x2, int y2,
        char *title, char **prompts, Boolean *active, ...){
    void	**data;
    Status	status;
    int	i, num_arg;
    va_list	arg_ptr;
    for(num_arg=0;prompts[num_arg][0];num_arg++);
    data   = new void*[num_arg];
    va_start(arg_ptr, active);
    for(i=0;i<num_arg;i++){
        data[i]   = va_arg(arg_ptr, void*);
    }
    va_end(arg_ptr);
    status=getdata(pw, CR_CB, x1, y1, x2, y2, title, prompts, active, data);
    delete data;
    return status;
}//--- GetActiveData ---//


//---------------------------------------------------------------------------
//
//	DTaskWindow Class
//
//---------------------------------------------------------------------------
Boolean DTaskWindow::DeleteCB(){
char *Yes="Yes",*No="No",*string="Do you really want to delete \"",*message;
Boolean returnvalue;
  message=new char[strlen(titlebar.CenterTitle())+strlen(string)+4];
  strcpy(message,string);
  strcat(message,titlebar.CenterTitle());
  strcat(message,"\"?");
  returnvalue=(Message(message,Yes,No,True)==Yes)?True:False;
  delete message;
  return returnvalue;
}//--- DTaskWindow::DeleteCB ---//


void DTaskWindow::AppendToDTaskList(DTaskWindow *DTW){
  if (DTW){
    previous = DTW;
    previous->next=this;
  }
}//--- DTaskWindow::InsertToDTaskList ---//


void DTaskWindow::DeleteFromDTaskList(){
  if (previous)
    previous->next=next;
  if (next)
    next->previous=previous;
  previous=next=0;
}//--- DTaskWindow::DeleteFromDTaskList ---//


DTaskWindow::DTaskWindow(Window &w,int x1,int y1,int x2,int y2,char *title,
			 DTaskWindow *DTW):
TaskWindow(w,x1,y1,x2,y2,True,True,"",title,""),
deletebutton(titlebar,2,2,titlebar.Height()-4,titlebar.Height()-4,LeftTopR){
  previous=next=0;
  if (DTW){
  DTaskWindow *LDTW=DTW->LastDTaskWindow();
    x1=LDTW->X1()+20;  y1=LDTW->Y1()+20;
    Move(x1,y1);
    AppendToDTaskList(LDTW);
  }
}//--- DTaskWindow::DTaskWindow ---//


DTaskWindow::~DTaskWindow(){
  Parent()->ChildCB(this,DestroyT);
  DeleteFromDTaskList();
}//--- DTaskWindow::~DTaskWindow ---//


DTaskWindow *DTaskWindow::LastDTaskWindow(){
DTaskWindow *DTW;
  for (DTW=this; DTW->next; DTW=DTW->next);
  return DTW;
}//--- DTaskWindow::LastDTaskWindow ---//


//===========================================================================
//
// End of GUI.C
//
//===========================================================================
