//===========================================================================
//
// MainWind.h
//
// By Poman So
//
//===========================================================================


#ifndef _MainWind_h_
#define _MainWind_h_

#include "Editor.h"
#include "Generator.h"
#include "Analyzer.h"
#include "Animator.h"
#include "Scope.h"
#include "3d-tlm.h"

//---------------------------------------------------------------------------
//
//     MainWindow Class
//
//---------------------------------------------------------------------------
class MainWindow: public Window{
	TitleBar     titlebar;
	MenuBar      menubar;
	PulldownMenu System;
	PulldownItem   S_Editor,S_Help,S_About;
	PulldownMenu File;
	PulldownItem   F_New,F_Open,F_Save,F_SaveAs,F_FreqRes,F_TimeRes,F_Quit;
	PulldownMenu T_Generator;
	PulldownItem   T_Generator1,T_Generator2,T_Generator3;
	PulldownMenu T_Scope;
	PulldownItem   T_Scope1,T_Scope2,T_Scope3;
	PulldownMenu T_Analyzer;
	PulldownItem   T_S11,T_S21;
	PulldownMenu T_Animator;
	PulldownItem   T_Animator1,T_Animator2,T_Animator3;
	PulldownMenu Simulate;
	PulldownItem   Si_Data,Si_Init,Si_Forward,Si_Backward;
	Editor       MeshEditor;
	Generator    *GeneratorPtr[3];
	Scope        *ScopePtr[3];
	Animator     *AnimatorPtr[3];
	Analyzer     *AnalyzerPtr[2];
	fstream	fs;
	int	StkPtr,StkSize,num_walls,num_ErUrs,num_generators,num_scopes,
		num_animators;
	StkElm	*ElementStk;
	Float7	*WallStack;
	Float8	*ErUrStack;
	Int8	*GeneratorStack;
	Int4	*ScopeStack;
	Int7	AnimatorStack[3];
	char	binfile[DosFileNameSize];
	Boolean FileSaved;
	Boolean Go_Ahead(const char *command);
	void	DeleteTools();
	void	GetStacks();
	void	GetAnimators();
	void	FreeStacks();
	void	FreeAnimators();
	void	ToolDestroyCB(Window *w);
	void	DataChangedCB(Window *w);
	void    AnalyzerChangedCB(Analyzer *S);
	void	AnimatorChangedCB();
	void	InjectGenerator(Boolean forward=True);
	Boolean BinFileIO(int mode);
	void	S_EditorCB();
	void	S_HelpCB();
        void    S_AboutCB();
        void	F_NewCB();
	void	F_OpenCB();
        void	F_SaveCB();
	void	F_SaveAsCB();
	void	F_FreqResCB();
	void	F_TimeResCB();
	Status	F_QuitCB();
	void	T_GeneratorCB(int n);
        void	T_ScopeCB(int n);
	void	T_AnimatorCB(int n);
	void    T_AnalyzerCB(int n);
	void	Si_DataCB();
	Boolean	Si_InitCB(Boolean beep=True);
	void	Si_ForwardCB();
	void	Si_BackwardCB();
public:
	MainWindow(Display &display);
	~MainWindow();
	Status	WindowCB(CBTag tag);
	Status	ChildCB(Window *w, CBTag tag);
};//--- MainWindow ---//



#endif

//===========================================================================
//
// End of MainWind.h
//
//===========================================================================
