//===========================================================================
//
// Node.h
//
// By Poman So
//
//===========================================================================

#ifndef _Node_h_
#define _Node_h_

#include "Real.h"
#include "Const.h"
#include "Vector.h"


class Node{
protected:
    static Real u, v, w;	// unit length in the x,y & z-direction
    Real er, ur, dt;	// material constants and time step
    Real yx, yy, yz,	// normalized y and z values of loading stubs
         zx, zy, zz,
         gx, gy, gz;
    Real s1, s2, s3,  // scattering coefficient
         s4, s5, s6,
         s7, s8, s9;
public:
    Node(VecR vr, Real er=1, Real ur=1);
    // (1) u=vr.x; v=vr.y; w=vr.z;
    // (2) assigns input values to er and ur
    // (3) execute SMatrix()
    Node();
    // (1) er=ur=1;
    // (2) execute SMatrix();
    Real Dt(){return dt;}
    Real Cal_Dt();
    // computes dt base on u,v,w,er and ur, so that all loading
    // stubs have either ZERO or POSITIVE values.
    void SMatrix(Real idt=0);
    // (1) dt=(idt>0)?idt:Cal_Dt()
    // (2) compute values for loading stubs
    // (3) compute values for scattering coefficients
    void SetErUr(Real er, Real ur);
    void GetUVW(Real &U, Real &V, Real &W){U=u; V=v; W=w;}
};//---Node---//


#endif

//===========================================================================
//
// End of Node.h
//
//===========================================================================
