//===========================================================================
//
// Window.C
//
// By Poman So
//
//===========================================================================
extern "C"{
#include "XWindow.h"
}
#include "Window.h"


//---------------------------------------------------------------------------
//
//	Display Class
//
//---------------------------------------------------------------------------
Display::Display(int i_argc, char **i_argv){
  argc = i_argc;
  argv = i_argv;
  x_root_window = X_OpenDisplay();
  width = X_DisplayWidth();
  height = X_DisplayHeight();
  Window::MonoChrome(False);
  Window::Initialize();
}//--- Display::Display ---//


Display::~Display(){
  X_CloseDisplay();
}//--- Display::~Display ---//


void Display::Flush(){
  X_FlushDisplay();
}//--- Display::Flush ---//


void Display::Sync(Boolean discard){
  X_Sync(discard);
}//--- Display::Sync ---//


//---------------------------------------------------------------------------
//
//	Window Class
//
//---------------------------------------------------------------------------
//
//	--- Some Constants ---
//


#define EGA_BLACK	  1
#define EGA_BLUE	  3
#define EGA_GREEN	  0
#define EGA_CYAN	  0
#define EGA_RED		  2
#define EGA_MAGENTA	  0
#define EGA_BROWN	  0
#define EGA_DARKGRAY	  0
#define EGA_LIGHTGRAY	  0
#define EGA_LIGHTBLUE	  0
#define EGA_LIGHTGREEN	  0
#define EGA_LIGHTCYAN	  0
#define EGA_LIGHTRED	  0
#define EGA_LIGHTMAGENTA  0
#define EGA_YELLOW	  0
#define EGA_WHITE	  0
#define EGA_BACKGROUND    0

#define EMPTY_FILL	  0
#define SOLID_FILL	  1
#define LINE_FILL	  2
#define LTSLASH_FILL	  3
#define SLASH_FILL	  4
#define BKSLASH_FILL	  5
#define LTBKSLASH_FILL	  6
#define HATCH_FILL	  7
#define XHATCH_FILL	  8
#define INTERLEAVE_FILL	  9
#define WIDE_DOT_FILL	 10
#define CLOSE_DOT_FILL	 11

#define SOLID_LINE	  0
#define DOTTED_LINE	  1
#define CENTER_LINE 	  2
#define DASHED_LINE	  3

#define NORM_WIDTH	  0
#define THICK_WIDTH	  1

#define LEFT_TEXT	  0
#define RIGHT_TEXT	  1
#define CENTER_TEXT	  2
#define TOP_TEXT	  3
#define BOTTOM_TEXT	  4

#define GOTHIC_FONT	  2
#define SANS_SERIF_FONT	  2
#define SMALL_FONT	  1
#define DEFAULT_FONT	  2
#define TRIPLEX_FONT	  2

#define HORIZ_DIR	  0
#define VERT_DIR	  1

#define COPY_PUT	  0
#define XOR_PUT		  1

Window	*Window::ActivatedWindow_;
Window	*Window::EnteredWindow_;
Window	*Window::ExposedWindow_;
Window	*Window::NextWindow_;
Window	*Window::NoEventWindow_;
Window	*Window::DispatchingWindow_;
Event	Window::WindowEvent_;
unsigned	Window::EventClock_;
int	Window::clip_;
int	Window::maxx_;
int	Window::maxy_;
int	Window::MouseX_;
int	Window::MouseY_;
int	Window::KeyPressed_;
Button	Window::MouseButton_;
Color	Window::drawcolor_;
Color	Window::fillcolor_;
Color	Window::drawC_;
Color	Window::fillC_;
Cursor	Window::cursor_;
FillStyle	Window::fillstyle_;
Font		Window::textfont_;
Direction	Window::textdir_;
Justify		Window::hjustify_;
Justify		Window::vjustify_;
int	        Window::charsize_;
LineStyle	Window::linestyle_;
LineWidth	Window::linewidth_;
WriteMode	Window::writemode_;
Boolean		Window::monochrome_;
Boolean		Window::visible_;

//
//	--- Protected Functions ---
//
int Window::ColorNum(Color c){
  if (!monochrome_)
  switch (c){
    case Black:		return EGA_BLACK;
    case Blue:		return EGA_BLUE;
    case Green:		return EGA_GREEN;
    case Cyan:		return EGA_CYAN;
    case Red:		return EGA_RED;
    case Magenta:	return EGA_MAGENTA;
    case Brown:		return EGA_BROWN;
    case DarkGray:	return EGA_DARKGRAY;
    case LightGray:	return EGA_LIGHTGRAY;
    case LightBlue:	return EGA_LIGHTBLUE;
    case LightGreen:	return EGA_GREEN;
    case LightCyan:	return EGA_LIGHTCYAN;
    case LightRed:	return EGA_LIGHTRED;
    case LightMagenta:	return EGA_LIGHTMAGENTA;
    case Yellow:	return EGA_YELLOW;
    case White:		return EGA_WHITE;
    case DefaultColor:
    default:		return EGA_WHITE;
  }
  else
    if (c==Black)
      return EGA_BLACK;
    else
      return EGA_WHITE;
}//--- Window::ColorNum ---//


int Window::DrawColorNum(Color c){
  if (c==DefaultColor)
    c = drawcolor;
  return ColorNum(c);
}//--- Window::DrawColorNum ---//


int Window::FillColorNum(Color c){
  if (c==DefaultColor)
    c = fillcolor;
  return ColorNum(c);
}//--- Window::FillColorNum ---//


int Window::FillStyleNum(){
  switch (fillstyle){
    case EmptyF:      return EMPTY_FILL;
    case SolidF:      return SOLID_FILL;
    case LineF:       return LINE_FILL;
    case LTSlashF:    return LTSLASH_FILL;
    case SlashF:      return SLASH_FILL;
    case BKSlashF:    return BKSLASH_FILL;
    case LTBKSlashF:  return LTBKSLASH_FILL;
    case HatchF:      return HATCH_FILL;
    case XHatchF:     return XHATCH_FILL;
    case InterLeaveF: return INTERLEAVE_FILL;
    case WideDotF:    return WIDE_DOT_FILL;
    case CloseDotF:   return CLOSE_DOT_FILL;
    case DefaultFillStyle:
    default:	      return SOLID_FILL;
  }
}//--- Window::FillStyleNum ---//


int Window::LineStyleNum(){
  switch (linestyle){
    case SolidL:  return SOLID_LINE;
    case DottedL: return DOTTED_LINE;
    case CenterL: return CENTER_LINE;
    case DashedL: return DASHED_LINE;
    case DefaultLineStyle:
    default:      return SOLID_LINE;
  }
}//--- Window::LineStyleNum ---//


int Window::LineWidthNum(){
  switch (linewidth){
    case Normal: return NORM_WIDTH;
    case Thick:  return THICK_WIDTH;
    default:     return NORM_WIDTH;
  }
}//--- Window::LineWidthNum ---//


int Window::HjustifyNum(){
  switch (hjustify){
    case LeftJ:   return LEFT_TEXT;
    case RightJ:  return RIGHT_TEXT;
    case CenterJ: return CENTER_TEXT;
    default:      return LEFT_TEXT;
  }
}//--- Window::HjustifyNum ---//
  

int Window::VjustifyNum(){
  switch (vjustify){
    case TopJ:    return TOP_TEXT;
    case BottomJ: return BOTTOM_TEXT;
    case CenterJ: return CENTER_TEXT;
    default:      return LEFT_TEXT;
  }
}//--- Window::VjustifyNum ---//
  

int Window::TextFontNum(Font font){
Font f;
  if (font==DefaultFont)
    f=textfont;
  else
    f=font;
  switch (f){
    case Gothic:    return GOTHIC_FONT;
    case SansSerif: return SANS_SERIF_FONT;
    case Small:     return SMALL_FONT;
    case Standard:  return DEFAULT_FONT;
    case Triplex:   return TRIPLEX_FONT;
    case DefaultFont:
    default:        return DEFAULT_FONT;
  }
}//--- Window::TextFontNum ---//


int Window::TextDirNum(){
  switch (textdir){
    case Horizontal: return HORIZ_DIR;
    case Vertical:   return VERT_DIR;
    default:	     return HORIZ_DIR;
  }
}//--- Window::TextDirNum ---//


int Window::WriteModeNum(){
  switch (writemode){
    case SetMode: return 0;
    case XorMode: return 1;
    default:	  return 0;
  }
}//--- Window::WriteModeNum ---//


void Window::Initialize(){
  ActivatedWindow_ = EnteredWindow_ = ExposedWindow_ = 0;
  NextWindow_ = NoEventWindow_ = DispatchingWindow_ = 0;
  maxx_      = X_DisplayWidth();
  maxy_      = X_DisplayHeight();
  drawcolor_ = fillcolor_ = DefaultColor;
  cursor_    = DefaultCursor;
  fillstyle_ = DefaultFillStyle;
  textfont_  = DefaultFont;
  textdir_   = DefaultDirection;
  hjustify_  = vjustify_ = DefaultJustify;
  charsize_  = 0;
  linestyle_ = DefaultLineStyle;
  linewidth_ = DefaultLineWidth;
  writemode_ = DefaultWriteMode;
  visible_   = True;
}//--- Window::Initialize ---//


//
//	--- Constructors & Destructor ---
//
Window::Window(Display &Dp,Color bk,Color bd,Color fg,Cursor cr,Font ft,
	       int bw){
  borderwidth = bw;
  x1 = int(0.1*Dp.Width());
  y1 = int(0.1*Dp.Height());
  x2 = int(0.9*Dp.Width());
  y2 = int(0.9*Dp.Height());
  clip_ = 1;
  threeDLook = 0;
  if (bk==DefaultColor) background=White; else background=bk;
  if (bd==DefaultColor) border    =Black; else border    =bd;
  if (fg==DefaultColor) drawcolor =Black; else border    =fg;
  drawC_=drawcolor;
  cursor=cr;
  fillcolor  = fillC_ = background;
  fillstyle  = SolidF;
  if (ft==DefaultFont) textfont=Small; else textfont=ft;
  textdir    = Horizontal;
  hjustify   = LeftJ;
  vjustify   = TopJ;
  charsize   = (textfont==Small)?4:1;
  linestyle  = SolidL;
  linewidth  = Normal;
  writemode  = SetMode;
  parent = Sibling = sibling = Child = child = 0;
  rubber = LeftTopR;
  rm = bm = 0;
  mapped = map_requested = True;
  popup = False;
  x_window =  X_CreateMainWindow(Dp.Root(),Dp.Argc(),Dp.Argv(),x1,y1,
				 x2-x1,y2-y1,borderwidth,cursor,textfont);
}//--- Window::Window ---//


Window::Window(Window &Wp, int X1, int Y1, int X2, int Y2, Color bk,
	       Color bd, Color fg, Cursor cr, Font ft, Rubber rb, int bw){
  x1=X1;  y1=Y1;  x2=X2;  y2=Y2;  threeDLook=0;
  borderwidth = bw;
  background  = (bk==DefaultColor)?Wp.background:bk;
  border      = (bd==DefaultColor)?Wp.border:bd;
  drawcolor   = drawC_ = (fg==DefaultColor)?Wp.drawcolor:fg;
  fillcolor   = fillC_ = (fg==DefaultColor)?Wp.fillcolor:background;
  cursor      = (cr==DefaultCursor)?Wp.cursor:cr;
  textfont    = (ft==DefaultFont)?Wp.textfont:ft;
  textdir     = Wp.textdir;
  charsize    = (textfont==Small)?4:1;
  hjustify    = Wp.hjustify;
  vjustify    = Wp.vjustify;
  fillstyle   = Wp.fillstyle;
  linestyle   = Wp.linestyle;
  linewidth   = Wp.linewidth;
  writemode   = Wp.writemode;
  parent      = &Wp;
  Sibling = sibling = Child = child = 0;
  InsertToParent();
  rubber = LeftTopR;
  if (rb==DefaultRubber) rubber=LeftTopR; else rubber=rb;
  rm = Wp.Width()-x2;
  bm = Wp.Height()-y2;
  mapped = map_requested = popup = False;
  x_window = X_CreateWindow(Wp.x_window,x1,y1,x2-x1,y2-y1,borderwidth,
			    cursor,textfont);
}//--- Window::Window ---//


Window::~Window(){
  if (ActivatedWindow_ == this) ActivatedWindow_ = 0;
  if (EnteredWindow_   == this) EnteredWindow_   = 0;
  if (NoEventWindow_   == this) NoEventWindow_   = 0;
  Unmap();
  DeleteFromParent();
  X_DestroyWindow(x_window);
  x_window = 0;
}//--- Window::~Window ---//


//
//	--- Window Information Functions ---
//
Window *Window::Root(){
  Window *wp;
  for (wp=this;wp->parent;wp=wp->parent);
  return wp;
}//--- Window::Root ---//


float Window::TimeSec(unsigned c_start, unsigned c_end){
  return (c_end-c_start)/1.0e6;
}//--- Window::TimeSec ---//


void Window::MouseAt(int &x, int &y){
  X_MouseAt(x_window,&x,&y);
}//--- Window::MouseAt ---//


//
//	--- Window Manipulation Functions ---
//
void Window::Expose(){
  WindowCB(ExposeT);
  Flush();
}//--- Window::Expose ---//


void Window::RePaint(){
Window *wp;
  WindowCB(ExposeT);
  for (wp=Child;wp;wp=wp->sibling)
    wp->RePaint();
}//--- Window::RePaint ---//

void Window::InsertToParent(){
  if(!parent) return;
  Sibling = parent->child;
  sibling = 0;
  if (Sibling)
    Sibling->sibling=this;
  else
    parent->Child=this;
  parent->child=this;
}//--- Window::InsertToParent ---//


void Window::DeleteFromParent(){
  if (!parent) return;
  if(Sibling)
    Sibling->sibling=sibling;
  else
    parent->Child=sibling;
  if(sibling)
    sibling->Sibling=Sibling;
  else
    parent->child=Sibling;
  parent=Sibling=sibling=0;
}//--- Window::DeleteFromParent ---//


void Window::Map(){
  if (popup){
    SaveBackground();
  }
  mapped = True;
  map_requested = True;
  WindowCB(MapT);
  X_MapWindow(x_window);
}//--- Window::Map ---//


void Window::Unmap(Boolean mr){
  map_requested = mr;
  mapped = False;
  X_UnmapWindow(x_window);
  if (popup){
    X_RemoveExposeEvents();
    RestoreBackground();
  }
  WindowCB(UnmapT);
}//--- Window::Unmap ---//


void Window::Unmap(Cursor wait_cursor){
Window *p=(parent)?parent:this;
Cursor c=p->SetCursor(wait_cursor);
  Unmap();
  p->SetCursor(c);
}//--- Window::Unmap ---//


void Window::Top(){
  if (sibling){
    if (Sibling)
      Sibling->sibling=sibling;
    else
      parent->Child=sibling;
    sibling->Sibling=Sibling;
    Sibling=parent->child;
    parent->child=Sibling->sibling=this;
    sibling=0;
  }
  X_RaiseWindow(x_window);
}//--- Window::Top ---//


void Window::Bottom(){
  if (Sibling){
    if (sibling==0)
      parent->child=Sibling;
    else
      sibling->Sibling=Sibling;
    Sibling->sibling=sibling;
    sibling=parent->Child;
    parent->Child=sibling->Sibling=this;
    Sibling=0;
  }
  X_LowerWindow(x_window);
}//--- Window::Bottom ---//


void Window::Move(int X1, int Y1, Boolean top){
  x2 = X1+Width();
  y2 = Y1+Height();
  x1 = X1;
  y1 = Y1;
  rm = parent->Width()-x2;
  bm = parent->Height()-y2;
  if (top) Top();
  X_MoveWindow(x_window,x1,y1);
}//--- Window::Move ---//


void Window::Resize(int X2, int Y2, Boolean top){
  x2 = X2;
  y2 = Y2;
  rm = parent->Width()-x2;
  bm = parent->Height()-y2;
  if (top) Top();
  X_ResizeWindow(x_window,x2-x1,y2-y1);
}//--- Window::Resize ---//


void Window::MoveResize(int X1, int Y1, int X2, int Y2, Boolean top){
  x1 = X1;
  y1 = Y1;
  x2 = X2;
  y2 = Y2;
  rm = parent->Width()-x2;
  bm = parent->Height()-y2;
  if (top) Top();
  X_MoveResizeWindow(x_window,x1,y1,x2-x1,y2-y1);
}//--- Window::MoveResize ---//


void Window::SetBorder(int bw, int th){
  borderwidth = Max(0,bw);
  threeDLook  = Max(0,th);
  X_SetBorder(x_window,bw);
}//--- Window::SetBorder ---//


void Window::SetX1Y1X2Y2(int X1, int Y1, int X2, int Y2){
  MoveResize(X1,Y1,X2,Y2,False);
}//--- Window::SetX1Y1X2Y2 ---//


Cursor Window::SetCursor(Cursor cur){
Cursor oldcursor=cursor;
  if (cur!=DefaultCursor) cursor=cur;
  X_SetCursor(x_window,cursor);
  return oldcursor;
}//--- Window::SetCursor ---//


void Window::ToNoEventWindow(){
  if (NoEventWindow_) NoEventWindow_->WindowCB(LastNoEventT);
  NoEventWindow_ = this;
}//--- Window::ToNoEventWindow ---//


void Window::DragMove(Direction dir,int cfx1,int cfy1,int cfx2,int cfy2){
  if(parent){
  int tx1,ty1,tx2,ty2,mdx,mdy,mxo,myo,mx,my,W,H,PW,PH,BW;
  LineStyle parentlinestyle=parent->linestyle;
  WriteMode parentwritemode=parent->writemode;
  Color	    XorColor=(monochrome_)?parent->border:parent->background;
  Cursor    parentcursor;
    if (XorColor==Black) XorColor=White;
    parent->SetLineStyle(DottedL);
    parent->SetWriteMode(XorMode);
    parent->GrabPointer();
    parent->MouseAt(MouseX_,MouseY_);
    mxo=mx=MouseX_;
    myo=my=MouseY_;
    ExposedWindow_ = NextWindow_ = parent->child;
    PW = parent->Width()-cfx2;
    PH = parent->Height()-cfy2;
    W = x2-x1+1;
    H = y2-y1+1;
    tx1=x1; ty1=y1;  tx2=tx1+W;  ty2=ty1+H;  BW=(borderwidth==1)?0:borderwidth;
    parent->DrawOverSW();
    parent->Rectangle(tx1,ty1,tx2+BW,ty2+BW,XorColor);
    mdx=mdy=0;
    do{
      GetEvent();
      parent->MouseAt(MouseX_,MouseY_);
      switch (dir){
	case Horizontal: mdx=MouseX_-mxo; break;
	case Vertical  : mdy=MouseY_-myo; break;
	default        : mdx=MouseX_-mxo; mdy=MouseY_-myo;
      }
      if (mx!=MouseX_ || my!=MouseY_){
	mx = MouseX_;    my = MouseY_;
	parent->Rectangle(tx1,ty1,tx2+BW,ty2+BW,XorColor);
	tx1=x1+mdx;   ty1=y1+mdy;
	if (dir==DefaultDirection || dir==Horizontal){
	  if (tx1<cfx1) tx1=cfx1;
	  if (tx1>PW)   tx1=PW;
	}
	if (dir==DefaultDirection || dir==Vertical){
	  if (ty1<cfy1) ty1=cfy1;
	  if (ty1>PH)   ty1=PH;
	}
	tx2=tx1+W;  ty2=ty1+H;
	parent->Rectangle(tx1,ty1,tx2+BW,ty2+BW,XorColor);
      }
    }while(WindowEvent_!=ButtonRelease);
    ActivatedWindow_ = EnteredWindow_ = 0;
    UngrabPointer();
    parentcursor = Parent()->SetCursor(Watch);
    parent->Rectangle(tx1,ty1,tx2+BW,ty2+BW,XorColor);
    parent->DrawOverSW(False);
    parent->SetLineStyle(parentlinestyle);
    parent->SetWriteMode(parentwritemode);
    ExposedWindow_ = NextWindow_ = 0;
    if (x1!=tx1 || y1!=ty1)
      Move(tx1,ty1,True);
    else
      Top();
    Parent()->SetCursor(parentcursor);
  }
}//--- Window::DragMove ---//


void Window::DragResize(int minW, int minH, int maxW, int maxH){
  if(parent){
  int tx1,ty1,tx2,ty2,t1,t2,t3,t4,mx,my,dx,dy,Dx,BW;
  LineStyle parentlinestyle=parent->linestyle;
  WriteMode parentwritemode=parent->writemode;
  Color	    XorColor=(monochrome_)?parent->border:parent->background;
  Cursor    parentcursor;
    if (XorColor==Black && !monochrome_) XorColor=White;
    parent->SetLineStyle(DottedL);
    parent->SetWriteMode(XorMode);
    parent->GrabPointer();
    parent->MouseAt(MouseX_,MouseY_);
    mx=MouseX_;  my=MouseY_;  dx=dy=0;  Dx=x2-x1+1;
    ExposedWindow_ = NextWindow_ = parent->child;
    t1=tx1=x1;  t2=tx2=x2;  t3=ty1=y1;  t4=ty2=y2;  BW=1+borderwidth;
    parent->DrawOverSW();
    parent->Rectangle(tx1,ty1,tx2+BW,ty2+BW,XorColor);
    do{
      GetEvent();
      parent->MouseAt(MouseX_,MouseY_);
      if (mx!=MouseX_ || my!=MouseY_){
	dx+=MouseX_-mx;  dy+=MouseY_-my;  mx=MouseX_;  my=MouseY_;
	parent->Rectangle(tx1,ty1,tx2+BW,ty2+BW,XorColor);
	if (dx>-Dx){
	  tx1=t1;  tx2=t1+Dx+dx;
	  if (tx2-tx1<minW) tx2=tx1+minW;
	  if (tx2-tx1>maxW) tx2=tx1+maxW;
	}
	else{
	  tx1=Max(0,Min(t1,t1+Dx+dx));  tx2=t2;
	  if (tx2-tx1<minW) tx1=tx2-minW; 
	  if (tx2-tx1>maxW) tx1=tx2-maxW;
	}
	if (dy>0){
	  ty1=t3;  ty2=t3+dy;
	  if (ty2-ty1<minH) ty2=ty1+minH;
          if (ty2-ty1>maxH) ty2=ty1+maxH;
	}
	else{
	  ty1=Max(35,Min(t3,t3+dy));  ty2=t4;
	  if (ty2-ty1<minH) ty1=ty2-minH;
	  if (ty2-ty1>maxH) ty1=ty2-maxH;
	}
	parent->Rectangle(tx1,ty1,tx2+BW,ty2+BW,XorColor);
      }
    }while(WindowEvent_!=ButtonRelease);
    ActivatedWindow_ = EnteredWindow_ = 0;
    UngrabPointer();
    parentcursor = Parent()->SetCursor(Watch);
    parent->Rectangle(tx1,ty1,tx2+BW,ty2+BW,XorColor);
    parent->DrawOverSW(False);
    parent->SetLineStyle(parentlinestyle);
    parent->SetWriteMode(parentwritemode);
    ExposedWindow_ = NextWindow_ = 0;
    if (x1!=tx1 || y1!=ty1 || x2!=tx2 || y2!=ty2){
      if (x1==tx1 && y1==ty1)
	Resize(tx2,ty2,True);
      else
	MoveResize(tx1,ty1,tx2,ty2,True);
    }
    else
      Top();
    Parent()->SetCursor(parentcursor);
  }
}//--- Window::DragResize ---//


void Window::GrabPointer(int owner_event){
  X_GrabPointer(x_window, owner_event);
}//--- Window::GrabPointer ---//


void Window::UngrabPointer(){
  X_UngrabPointer();
}//--- Window::UngrabPointer ---//


//
//	--- Graphic Functions ---
//
void Window::Flush(){
  X_FlushDisplay();
}//--- Window::Flush ---//


void Window::Sync(Boolean discard){
  X_Sync(discard);
}//--- Window::Sync ---//


Boolean Window::SaveBackground(){
Boolean status;
  parent->DrawOverSW();
  status = parent->SaveImage(x1,y1,x2,y2);
  parent->DrawOverSW(False);
  return status;
}//--- Window::SaveBackground ---//


Boolean Window::RestoreBackground(){
Boolean status;
  parent->DrawOverSW();
  status = parent->RestoreImage(x1,y1);
  parent->DrawOverSW(False);
  return status;
}//--- Window::RestoreBackground ---//


Boolean Window::SaveImage(int x1, int y1, int x2, int y2){
  X_SaveImage(x_window,x1,y1,x2,y2);
  return True;
}//--- Window::GetImage ---//


Boolean Window::RestoreImage(int x, int y){
  X_RestoreImage(x_window,x,y);
  return True;
}//--- Window::PutImage ---//


void Window::Bell(int percent){
  X_Bell(percent);
}//--- Window::Bell ---//


void Window::Clear(){
  X_ClearWindow(x_window);
}//--- Window::Clear ---//


void Window::Clear(int x1, int y1, int x2, int y2){
  X_ClearArea(x_window,x1,y1,x2,y2,0);
}//--- Window::Clear ---//


#define GXclear                 0x0             /* 0 */
#define GXand                   0x1             /* src AND dst */
#define GXandReverse            0x2             /* src AND NOT dst */
#define GXcopy                  0x3             /* src */
#define GXandInverted           0x4             /* NOT src AND dst */
#define GXnoop                  0x5             /* dst */
#define GXxor                   0x6             /* src XOR dst */
#define GXor                    0x7             /* src OR dst */
#define GXnor                   0x8             /* NOT src AND NOT dst */
#define GXequiv                 0x9             /* NOT src XOR dst */
#define GXinvert                0xa             /* NOT dst */
#define GXorReverse             0xb             /* src OR NOT dst */
#define GXcopyInverted          0xc             /* NOT src */
#define GXorInverted            0xd             /* NOT src OR dst */
#define GXnand                  0xe             /* NOT src OR NOT dst */
#define GXset                   0xf             /* 1 */
void Window::SetWriteMode(WriteMode wm){
int f;
  writemode = wm;
#ifdef __ActiveCC
  if (wm==XorMode) f=GXinvert; else f=GXcopy;
#else
  if (wm==XorMode) f=GXxor; else f=GXcopy;
#endif
  X_SetFunction(x_window,f);
}//--- Window::SetWriteMode ---//


void Window::SetLineStyle(LineStyle ls, LineWidth lw){
int LS,LW;
  linestyle=ls;  linewidth=lw;
  switch(ls){
  case DefaultLineStyle: LS=0; break;
  case SolidL:		 LS=0; break;
  case DottedL:		 LS=1; break;
  case CenterL:		 LS=0; break;
  case DashedL:		 LS=2; break;
  default:		 LS=0; break;
  }
  switch(lw){
  case DefaultLineWidth: LW=0; break;
  case Normal:		 LW=0; break;
  case Thick:		 LW=3; break;
  default:		 LW=0; break;
  }
  X_SetLineAttributes(x_window,LW,LS);
}//--- Window::SetLineStyle ---//


void Window::SetFillStyle(FillStyle fs, Color fc){
  fillstyle=fs;  fillcolor=fc;
}//--- Window::SetFillStyle ---//


void Window::DrawOverSW(Boolean mode){
  if (mode)
    X_SetSubwindowMode(x_window,0);
  else
    X_SetSubwindowMode(x_window,1);
}//--- Window::DrawOverSW ---//


void Window::Point(int x, int y, Color c){
  X_SetForeground(x_window,DrawColorNum(c));
  X_Point(x_window,x,y);
}//--- Window::Point ---//


void Window::Points(int n, int *xy, Color c){
  X_SetForeground(x_window,DrawColorNum(c));
  X_Points(x_window,n,xy);
}//--- Window::Points ---//


void Window::Line(int x1, int y1, int x2, int y2, Color c){
  X_SetForeground(x_window,DrawColorNum(c));
  X_Line(x_window,x1,y1,x2,y2);
}//--- Window::Line ---//


void Window::Lines(int n, int *xy, Color c){
  X_SetForeground(x_window,DrawColorNum(c));
  X_Lines(x_window,n,xy);
}//--- Window::Lines ---//


void Window::Rectangle(int x1, int y1, int x2, int y2, Color c){
  X_SetForeground(x_window,DrawColorNum(c));
  X_Rectangle(x_window,x1,y1,x2,y2);
}//--- Window::Rectangle ---//


void Window::FillRectangle(int x1, int y1, int x2, int y2, Color fc, Color dc){
  X_SetForeground(x_window,FillColorNum(fc));
  X_FillRectangle(x_window,x1,y1,x2,y2);
  if (dc!=DefaultColor && dc!=fc){
    X_SetForeground(x_window,DrawColorNum(dc));
    X_Rectangle(x_window,x1,y1,x2,y2);
  }
}//--- Window::FillRectangle ---//


void Window::Polygon(int n, int *xy, Color c){
  X_SetForeground(x_window,DrawColorNum(c));
  X_Polygon(x_window,n,xy);
}//--- Window::Polygon ---//


void Window::FillPolygon(int n, int *xy, Color fc, Color dc){
  X_SetForeground(x_window,FillColorNum(fc));
  X_FillPolygon(x_window,n,xy);
  if (dc!=DefaultColor && dc!=fc){
    X_SetForeground(x_window,DrawColorNum(dc));
    X_Polygon(x_window,n,xy);
  }
}//--- Window::FillPolygon ---//


void Window::Ellipse(int x, int y, int xr, int yr, Color c){
  X_SetForeground(x_window,DrawColorNum(c));
  X_Ellipse(x_window,x,y,xr,yr);
}//--- Window::Ellipse ---//


void Window::FillEllipse(int x, int y, int xr, int yr, Color fc, Color dc){
  X_SetForeground(x_window,FillColorNum(fc));
  X_FillEllipse(x_window,x,y,xr,yr);
  if (dc!=DefaultColor && dc!=fc){
    X_SetForeground(x_window,DrawColorNum(dc));
    X_Ellipse(x_window,x,y,xr,yr);
  }
}//--- Window::FillEllipse ---//


void Window::Surface(int nx, int ny, int *xy, Color c){
  X_SetForeground(x_window,DrawColorNum(c));
  X_Surface(x_window,nx,ny,xy);
}//--- Window::Surface ---//


void Window::FillSurface(int nx,int ny,int *xy,int *d,Color fc,Color dc){
  X_FillSurface(x_window,nx,ny,xy,d,FillColorNum(fc),DrawColorNum(dc));
}//--- Window::FillSurface ---//


void Window::SetTextStyle(Font f, Direction d, int s){
  if (f!=DefaultFont)      textfont=f;
  if (d!=DefaultDirection) textdir =d;
  if (s<=0)		   charsize=s;
  textfont_ = textfont;
  textdir_  = textdir;
  charsize_ = charsize;
  X_SetFont(x_window,TextFontNum());
}//--- Window::SetTextStyle ---//


void Window::SetTextJustify(Justify h, Justify v){
  hjustify = h;
  vjustify = v;
}//--- Window::SetTextJustify ---//


void Window::OutTextXY(int x, int y, const char *s, Color c){
  if (!s) return;
  X_SetForeground(x_window,DrawColorNum(c));
  if (textdir==Horizontal){
    if (hjustify!=LeftJ)
      x -= TextWidth(s,textfont)/((hjustify==CenterJ)?2:1);
    if (vjustify!=BottomJ)
      y += TextHeight(s,textfont)/((vjustify==CenterJ)?2:1);
    X_DrawHText(x_window, x, y, s);
  }
  else{
    if (hjustify!=LeftJ)
      x -= TextHeight(s,textfont)/((hjustify==CenterJ)?2:1);
    if (vjustify!=BottomJ)
      y += TextWidth(s,textfont)/((vjustify==CenterJ)?2:1);
    X_DrawVText(x_window, x, y, s);
  }
}//--- Window::OutTextXY ---//


int Window::TextWidth(const char *s, Font ft){
  return int(X_FontTextWidth(TextFontNum(ft),s));
}//--- Window::TextWidth ---//


int Window::TextHeight(const char *s, Font ft){
  return int(X_FontTextHeight(TextFontNum(ft),s));
}//--- Window::TextHeight ---//


//
//	--- Event Handling Functions ---
//

Event Window::GetEvent(){
  WindowEvent_ = Event(X_NextEvent(&MouseX_,&MouseY_,(int*)((&MouseButton_)),
			&KeyPressed_,&EventClock_));
  return WindowEvent_;
}//--- Window::GetEvent---//


Event Window::CheckEvents(){
Window *wp;
Event  event;
  do{
    event = GetEvent();
    if (event==ButtonRelease || event==KeyPress){
      return event;
    }
    if (event==::Expose || event==ConfigureE){
      wp = IdentifyEventWindow(event);
      if (wp){
	visible_ = True;
	if (event==ConfigureE)
	  wp->ConfigureCB();
	else{
	  Sync(True);
	  RePaint();
        }
      }
    }
    else if (event==PropertyE){
      visible_ = False;
    }
  } while (event!=NoEvent);
  return NoEvent; 
}//--- Window::CheckEvents ---//


Window* Window::IdentifyEventWindow(Event event){
Window *ChildWindow,*EventWindow;
  if (X_IsEventWindow(x_window))
    return this;
  else{
    if (event==::Expose || event==ConfigureE)
      if (X_IsEventWindow(Root()->x_window))
	return Root();
      else
	ChildWindow=Root()->Child;
    else
      ChildWindow=Child;
    for (;ChildWindow;ChildWindow=ChildWindow->sibling)
      if (X_IsEventWindow(ChildWindow->x_window))
        return ChildWindow;
      else{
        EventWindow = ChildWindow->IdentifyEventWindow();
        if (EventWindow)
          return EventWindow;
      }
    return 0;
  }
}//--- Window::IdentifyEventWindow ---//


void Window::ConfigureCB(){
Window	 *cp;
Rubber   rub;
unsigned width,height;
  X_GetConfiguration(x_window,&x1,&y1,&width,&height);
  x2=x1+width;  y2=y1+height;
  WindowCB(ResizeT);
  for (cp=Child; cp; cp=cp->sibling){
    if ((rub=cp->rubber) != LeftTopR){
      if (rub & RightR){
        int dx=Width()-cp->rm-cp->x2;
        cp->x2 += dx;
        if (!(rub & LeftR)) cp->x1+=dx;
      }
      if (rub & BottomR){
        int dy=Height()-cp->bm-cp->y2;
        cp->y2 += dy;
        if (!(rub & TopR)) cp->y1+=dy;
      }
      X_MoveResizeWindow(cp->x_window,cp->x1,cp->y1,
			 cp->x2-cp->x1,cp->y2-cp->y1);
    }
  }
}//--- Window::ConfigureCB ---//


Status Window::DispatchEvents(){
unsigned old_clock;
Status	status=Continue;
Window	*wp,*old_activated=0,*old_DispatchingWindow=DispatchingWindow_;
  DispatchingWindow_ = this;
  while (GetEvent()!=NoEvent);
  Top();
  Map();
  old_clock = EventClock_;
  do{
    GetEvent();
    if (WindowEvent_==KeyPress){
      if (KeyPressed_==-1){
	status = Quit;
      }
      else if (NoEventWindow_)
        status = NoEventWindow_->WindowCB(KeyPressT);
    }
    else{
      wp = IdentifyEventWindow(WindowEvent_);
      switch(WindowEvent_){
      case Activated:
	if (NoEventWindow_) NoEventWindow_->WindowCB(LastNoEventT);
	ActivatedWindow_ = EnteredWindow_ = NoEventWindow_ = wp;
	if (wp){
	  if (wp==old_activated && (TimeSec(old_clock,EventClock_)<0.25)){
	    status = wp->WindowCB(SelectT);
	    old_activated = ActivatedWindow_ = EnteredWindow_ = 0;
	    X_UngrabPointer();
	  }
	  else{
	    status = wp->WindowCB(ActivateT);
	    wp->WindowCB(EnterT);
	    old_activated = wp;
	  }
	  old_clock = EventClock_;
	}
	break;
      case ButtonPress:
	if (wp==EnteredWindow_){
	  if (wp) status = wp->WindowCB(ButtonPressT);
	}
	else{
	  if (EnteredWindow_) EnteredWindow_->WindowCB(LeaveT);
	  EnteredWindow_=wp;
	  if (EnteredWindow_) EnteredWindow_->WindowCB(EnterT);
	}
	break;
      case ButtonRelease:
	if (wp!=EnteredWindow_){
	  if (EnteredWindow_)
	    EnteredWindow_->WindowCB(LeaveT);
	  if (wp) wp->WindowCB(EnterT);
	}
	if (wp){
	  wp->WindowCB(LeaveT);
	  status = wp->WindowCB(ButtonReleaseT);
	}
	if (ActivatedWindow_) ActivatedWindow_->WindowCB(DeActivateT);
	ActivatedWindow_ = EnteredWindow_ = 0;
	X_UngrabPointer();
	break;
      case ConfigureE:
	if (wp) wp->ConfigureCB();
	break;
      case ::Expose:
        if (wp) wp->WindowCB(ExposeT);
	break;
      case PropertyE:
	if (wp) wp->WindowCB(PropertyT);
	break;
      default:
	if (NoEventWindow_) NoEventWindow_->WindowCB(NoEventT);
      }
    }
  }while(status==Continue);
  DispatchingWindow_ = old_DispatchingWindow;
  return status;
}//--- Window::DispatchEvents ---//


Window *Window::Wait(Event event){
  while (GetEvent()!=event);
  return IdentifyEventWindow();
}//--- Window::Wait ---//


Status Window::WindowCB(CBTag tag){
  if (parent)
    return parent->ChildCB(this,tag);
  else
    return Continue;
}//--- Window::WindowCB ---//


Status Window::ChildCB(Window *w, CBTag tag){
  if (parent)
    return parent->ChildCB(w,tag);
  else
    return Continue;
}//--- Window::ChildCB ---//


fstream &Window::IO(fstream &fs){
  return fs;
}//--- Window::IO ---//


//===========================================================================
//
// End of Window.C
//
//===========================================================================
