/*==========================================================================*/
/*									    */
/*	XWindow.c							    */
/*									    */
/*	By Poman So							    */
/*									    */
/*==========================================================================*/

#define max(x,y) (((x)>(y))?(x):(y))
#define min(x,y) (((x)<(y))?(x):(y))


#include <sys/dirent.h>
#include <sys/types.h>
#include <stddef.h>
#include <malloc.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/cursorfont.h>
#include <X11/keysym.h>
#include <X11/keysymdef.h>
#include "XWindow.h"


#define EventMask (KeyPressMask|KeyReleaseMask|ButtonPressMask|\
		   ButtonReleaseMask|ButtonMotionMask|ExposureMask|\
		   EnterWindowMask|LeaveWindowMask|StructureNotifyMask)
#define NoEventE	 0
#define ActivatedE	 1
#define ButtonPressE	 2
#define ButtonReleaseE	 3
#define KeyPressE	 4
#define KeyReleaseE	 5
#define EnterE		 6
#define LeaveE		 7
#define ExposeE		 8
#define MapE		 9
#define UnmapE		10
#define MoveE		11
#define MotionE		12
#define ResizeE		13
#define MoveResizeE	14
#define ConfigureE	15
#define PropertyE	16
#define UnknownE	17

#define x_black          1
#define x_background     0
#define x_white		 0
#define x_red            2
#define x_blue           3
#define x_numPixels      16

unsigned long thePixels[x_numPixels];
char *theColorNames[x_numPixels] = {
    "black",        /*  0 */
    "red",          /*  1 */
    "blue",         /*  2 */
    "green",        /*  3 */
    "DarkGreen",    /*  4 */
    "yellow",       /*  5 */
    "cyna",         /*  6 */
    "violet",       /*  7 */
    "DeepSkyBlue",  /*  8 */
    "coral",        /*  9 */
    "brown",        /* 10 */
    "DarkSeaGreen", /* 11 */
    "DarkSlateBlue", /* 12 */
    "DeepPink",     /* 13 */
    "Firebrick",    /* 14 */
    "white",        /* 15 */
};

typedef struct{
    Window	x_window;
    GC		x_gc;
    Pixmap	x_pixmap;
    int		x_pixmap_width, x_pixmap_height;
    XGCValues	x_gc_values;
    Cursor	x_cursor;
    XFontStruct	*x_font;
} XWindow;

enum CURSOR{Cdefault, Arrow, Cross, Hand1, Pencil, Watch};
enum FONT{Fdefault, CourierBold12, CourierBold14};

static Display		*x_display;
static int		x_screen;
static unsigned		x_display_width, x_display_height, x_display_depth;
static Visual		*x_visual;
static Colormap		x_colormap;
static int		x_numcells, x_avilcells;
static unsigned long	*x_pixel;
static XFontStruct	*x_courier_bold12, *x_courier_bold14;
static Cursor		x_arrow, x_cross, x_hand1, x_pencil, x_watch;
static XEvent		x_event;

static int  RedColor(int i);
static int  GreenColor(int i);
static int  BlueColor(int i);
static void ComputePixels(void);
static void ShowColor(Drawable d);

static int RedColor(int i){
    int    j, v;
    double cf;
    j = (int)(2.0*x_avilcells/3.0);
    if (i<j)
        return 0;
    else{
        cf = ((double)(i-j))/(x_avilcells-j);
        v  = (int)(65535*cf);
        return (v>65535)?65536:v;
    }
}/*--- RedColor ---*/

static int GreenColor(int i){
    int    j, k, v;
    double cf;
    j = x_avilcells/3;
    k = (int)(2.0*x_avilcells/3.0);
    if (i>65536 || i<j)
        return 0;
    if (i<k)
        cf = ((double)(i-j))/(k-j);
    else
        cf = ((double)(x_avilcells-i))/(x_avilcells-k);
    v  = (int)(65535*cf);
    return (v>65535)?65536:v;
}/*--- GreenColor ---*/

static int BlueColor(int i){
    int    j, k, v;
    double cf;
    k = (int)(2.0*x_avilcells/3.0);
    if (i>k || i<0)
        return 0;
    j = x_avilcells/3;
    if (i<j)
        cf = ((double)i)/j;
    else
        cf = ((double)(k-i))/(k-j);
    v  = (int)(65535*cf);
    return (v>65535)?65536:v;
}/*--- BlueColor ---*/


static void PrintData(void){
#ifdef DEBUG
    printf("x_screen  = %d\n", x_screen);
    printf("x_display_width  = %d\n", x_display_width);
    printf("x_display_height = %d\n", x_display_height);
    printf("x_display_depth  = %d\n", x_display_depth);
    printf("x_numcells       = %d\n", x_numcells);
    printf("x_avilcells      = %d\n", x_avilcells);
#endif
}/*--- PrintData ---*/

static void ShowColor(Drawable d){
    int i;
    for(i=0;i<x_avilcells;i++){
        XSetWindowBackground(x_display, d, x_pixel[i]);
        XClearWindow(x_display, d);
        XFlush(x_display);
    }
    PrintData();
}/*--- ShowColor ---*/


static int TheKey(XKeyEvent *theEvent){
    int		length;
    int		theKeyBufferMaxLen=64;
    char		theKeyBuffer[65];
    KeySym		theKeySym;
    XComposeStatus	theComposeStatus;
    length=XLookupString(theEvent, theKeyBuffer, theKeyBufferMaxLen,
            &theKeySym, &theComposeStatus);
    if (theEvent->state & Mod1Mask){
        return 0;
    }
    if ((theKeySym>=' ') &&	/*--- ASCII  32 is a space ---*/
            (theKeySym<='~') &&	/*--- ASCII 126 is a tilde ---*/
            (length>0)){		/*--- We have char input   ---*/
        return theKeySym;
    }
    else{
        switch (theKeySym){
            case XK_Return:
            case XK_Down:
            case XK_KP_Enter:		return '\r';
            case XK_BackSpace:
            case XK_Delete:
            case XK_Left:		return '\b';
            case XK_Escape:
            case XK_F1:		return -1;
            default:			return 0;
        }
    }
}/*--- TheKey ---*/

static void ComputePixels(void){
    XColor *x_color;
    int	i;
    char	flags;
    while(x_avilcells>0 &&
    !XAllocColorCells(x_display, x_colormap, False, NULL, 0, x_pixel,
            x_avilcells)) {
        // fprintf(stderr, "num of color cell is %d\n", x_avilcells);
        x_avilcells--;
    }
    if (x_avilcells>0){
        flags = DoRed|DoGreen|DoBlue;
        x_color = (XColor*)malloc(sizeof(XColor)*x_avilcells);
        for(i=0;i<x_avilcells;i++){
            x_color[i].pixel = x_pixel[i];
            x_color[i].red   = RedColor(i);
            x_color[i].green = GreenColor(i);
            x_color[i].blue  = BlueColor(i);
            x_color[i].flags = flags;
        }
        x_color[0].red   = 65535;
        x_color[0].green = 65535;
        x_color[0].blue  = 65535;
        x_color[1].red   = 0;
        x_color[1].green = 0;
        x_color[1].blue  = 0;
        x_color[2].red   = 65535;
        x_color[2].green = 0;
        x_color[2].blue  = 0;
        x_color[3].red   = 0;
        x_color[3].green = 0;
        x_color[3].blue  = 65535;
        XStoreColors(x_display, x_colormap, x_color, x_avilcells);
        free(x_color);
    }
    else{
        PrintData();//fprintf(stderr, "No free color cells!\n");
        x_avilcells = x_numcells;//-4;
        free(x_pixel);
        x_pixel = (unsigned long*)malloc(sizeof(unsigned long)*x_avilcells);
        unsigned long black_pixel = BlackPixel(x_display, x_screen);
        unsigned long white_pixel = WhitePixel(x_display, x_screen);
        for (i=0;i<x_avilcells;i++) x_pixel[i]=black_pixel;
        x_pixel[x_background] = white_pixel;
        
        // to be comppleted
        XColor theRGBColor, theHardwareColor;
        int theDepth = DefaultDepth(x_display,x_screen);
        if(theDepth>1)
            for(i=0;i<x_numPixels;i++){
                int theStatus = XLookupColor(x_display, x_colormap,
                        theColorNames[i], &theRGBColor, &theHardwareColor);
                if(theStatus != 0){
                    theStatus = XAllocColor(x_display, x_colormap, &theHardwareColor);
                    if(theStatus != 0)
                        thePixels[i] = theHardwareColor.pixel;
                    else
                        thePixels[i] = black_pixel;
                }
            }
        else
        {
            for(i=0;i<x_numPixels;i++)
            {
                if(strcmp("white", theColorNames[i])==0)
                    thePixels[i] = white_pixel;
                else
                    thePixels[i] = black_pixel;
            }
        }
        
        x_pixel[x_red ] = thePixels[1];
        x_pixel[x_blue] = thePixels[2];
    }
}/*--- ComputePixels ---*/


/*----------------------------------------------------------------------*/
/*				Global functions			*/
/*----------------------------------------------------------------------*/

/*------------------------------*/
/*	Display Functions	*/
/*------------------------------*/
char *X_OpenDisplay(void){
    XWindow	*Xw;
    Xw = (XWindow*)malloc(sizeof(XWindow));
    x_display = 0;
    /*-------------------------------------------*/
    /* Open display and get display information. */
    /*-------------------------------------------*/
    x_display	    = XOpenDisplay(NULL);
    x_screen = XDefaultScreen(x_display);
    Xw->x_window    = XDefaultRootWindow(x_display);
    Xw->x_gc	      = 0;
    x_display_width  = XDisplayWidth(x_display, x_screen);  // 1340;
    x_display_height = XDisplayHeight(x_display, x_screen); // 900;
    x_display_depth  = XDefaultDepth(x_display, x_screen);
    x_visual	   = XDefaultVisual(x_display, x_screen);
    x_colormap	   = XDefaultColormap(x_display, x_screen);
    x_numcells	   = XDisplayCells(x_display, x_screen);
    if (x_numcells>4) x_numcells=4;
    x_pixel	   = (unsigned long*)malloc(sizeof(unsigned long)*x_numcells);
    x_avilcells	   = x_numcells;
    ComputePixels();
    x_courier_bold12 = XLoadQueryFont(x_display, "variable");
    x_courier_bold14 = XLoadQueryFont(x_display, "variable");
    if (!x_courier_bold12 || !x_courier_bold14){
        fprintf(stderr, "Unable to load font!\n" );
        exit(1);
    }
    x_arrow  = XCreateFontCursor(x_display, XC_left_ptr);
    x_cross  = XCreateFontCursor(x_display, XC_cross);
    x_hand1  = XCreateFontCursor(x_display, XC_hand1);
    x_pencil = XCreateFontCursor(x_display, XC_pencil);
    x_watch  = XCreateFontCursor(x_display, XC_watch);
    if (!x_arrow || !x_cross || !x_hand1 || !x_pencil || !x_watch){
        fprintf(stderr, "Unable to load cursor!\n" );
        exit(1);
    }
    return (char*)Xw;
}/* --- X_OpenDisplay --- */

void X_CloseDisplay(void){
    XFreeCursor(x_display, x_watch);
    XFreeCursor(x_display, x_pencil);
    XFreeCursor(x_display, x_hand1);
    XFreeCursor(x_display, x_cross);
    XFreeCursor(x_display, x_arrow);
    XFreeFont(x_display, x_courier_bold14);
    XFreeFont(x_display, x_courier_bold12);
    free(x_pixel);
    XCloseDisplay(x_display);
}/* --- X_CloseDisplay --- */

unsigned X_DisplayWidth(void){
    return x_display_width;
}/* --- X_DisplayWidth --- */

unsigned X_DisplayHeight(void){
    return x_display_height;
}/* --- X_DisplayHeight --- */

void X_Bell(int percent){
    if (percent<-100)  percent = -100;
    if (percent> 100)  percent =  100;
    XBell(x_display, percent);
}/*--- X_Bell ---*/

void X_FlushDisplay(void){
    XFlush(x_display);
    XSync(x_display, False);
}/*--- X_FlushDisplay ---*/

void X_Sync(int discard){
    XSync(x_display, (discard)?True:False);
}/*--- X_Sync ---*/

/*------------------------------*/
/*	Window Functions	*/
/*------------------------------*/
char *X_CreateMainWindow(char *parent, int argc, char **argv,
        int x, int y, int w, int h, int b,
        int cursor, int font){
    XSetWindowAttributes xwa;
    XSizeHints	     x_size_hints;
    unsigned	     W, H, B;
    unsigned long        x_gc_valuemask, x_window_valuemask;
    XWindow		     *Xw;
    Window               x_root_window;
    Xw = (XWindow*)malloc(sizeof(XWindow));
    x_root_window = ((XWindow*)parent)->x_window;
    W=(w<1)?1:w;
    H=(h<1)?1:h;
    B=(b<0)?0:b;
    /*---------------*/
    /* Create window */
    /*---------------*/
    x_window_valuemask   = CWBackPixel;
    xwa.background_pixel = x_pixel[x_background];
    x_window_valuemask |= CWColormap;
    xwa.colormap = x_colormap;
    x_window_valuemask |= CWEventMask;
    xwa.event_mask = EventMask|PropertyChangeMask;
    x_window_valuemask |= CWBackingStore;
    xwa.backing_store = NotUseful;
    x_window_valuemask |= CWSaveUnder;
    xwa.save_under = False;
    Xw->x_window = XCreateWindow(x_display, x_root_window, x, y, W, H, B,
            x_display_depth, InputOutput, x_visual, x_window_valuemask, &xwa);
    Xw->x_pixmap = 0;
    Xw->x_pixmap_width = Xw->x_pixmap_height = 0;
    X_SetCursor((char*)Xw, cursor);
    /*--------------------------------------*/
    /* Create graphic context and set font. */
    /*--------------------------------------*/
    x_gc_valuemask = 0;
    Xw->x_gc = XCreateGC(x_display, Xw->x_window, x_gc_valuemask,
            &(Xw->x_gc_values));
    X_SetFont((char*)Xw, font);
    /*-----------------------------------------------------*/
    /* Set size hints and communicate with window manager. */
    /*-----------------------------------------------------*/
    x_size_hints.flags	  = PPosition|PSize|PMinSize|PMaxSize;
    x_size_hints.x	  = x;
    x_size_hints.y	  = y;
    x_size_hints.width	  = W;
    x_size_hints.height	  = H;
    x_size_hints.min_width  = 500;
    x_size_hints.min_height = 400;
    x_size_hints.max_width  = x_display_width;
    x_size_hints.max_height = x_display_height;
    XSetStandardProperties(x_display, Xw->x_window, argv[0], argv[0], None,
            argv, argc, &x_size_hints);
    return (char*)Xw;
}/* --- X_CreateMainWindow --- */

char *X_CreateWindow(char *parent, int x, int y, int w, int h,
        int b, int cursor, int font){
    XSetWindowAttributes xwa;
    unsigned	     W, H, B;
    unsigned long        x_gc_valuemask, x_window_valuemask;
    XWindow		     *Xw;
    Window		     Pw;
    Xw = (XWindow*)malloc(sizeof(XWindow));
    Pw = ((XWindow*)parent)->x_window;
    W=(w<1)?1:w;
    H=(h<1)?1:h;
    B=(b<0)?0:b;
    /*---------------*/
    /* Create window */
    /*---------------*/
    x_window_valuemask   = CWBackPixel;
    xwa.background_pixel = x_pixel[x_background];
    x_window_valuemask |= CWColormap;
    xwa.colormap = x_colormap;
    x_window_valuemask |= CWEventMask;
    xwa.override_redirect = True;
    x_window_valuemask |= CWOverrideRedirect;
    xwa.event_mask = EventMask;
    x_window_valuemask |= CWBackingStore;
    xwa.backing_store = NotUseful;
    x_window_valuemask |= CWSaveUnder;
    xwa.save_under = False;
    Xw->x_window = XCreateWindow(x_display, Pw, x, y, W, H, B, x_display_depth,
            InputOutput, x_visual, x_window_valuemask, &xwa);
    Xw->x_pixmap = 0;
    Xw->x_pixmap_width = Xw->x_pixmap_height = 0;
    X_SetCursor((char*)Xw, cursor);
    /*--------------------------------------*/
    /* Create graphic context and set font. */
    /*--------------------------------------*/
    x_gc_valuemask = 0;
    Xw->x_gc = XCreateGC(x_display, Xw->x_window, x_gc_valuemask,
            &(Xw->x_gc_values));
    X_SetFont((char*)Xw, font);
    return (char*)Xw;
}/* --- X_CreateWindow --- */

void X_DestroyWindow(char *window){
    XFreeGC(x_display, ((XWindow*)window)->x_gc);
    XDestroyWindow(x_display, ((XWindow*)window)->x_window);
    free(window);
}/* --- X_DestroyWindow --- */

void X_MapWindow(char *window){
    Window x_window;
    x_window = ((XWindow*)window)->x_window;
    XMapWindow(x_display, x_window);
}/*--- X_MapWindow --- */

void X_RaiseWindow(char *window){
    Window x_window;
    x_window = ((XWindow*)window)->x_window;
    XRaiseWindow(x_display, x_window);
}/*--- X_RaiseWindow --- */

void X_LowerWindow(char *window){
    Window x_window;
    x_window = ((XWindow*)window)->x_window;
    XLowerWindow(x_display, x_window);
}/*--- X_LowerWindow --- */

void X_MapRaiseWindow(char *window){
    Window x_window;
    x_window = ((XWindow*)window)->x_window;
    XMapRaised(x_display, x_window);
}/*--- X_MapRaiseWindow --- */

void X_UnmapWindow(char *window){
    Window x_window;
    x_window = ((XWindow*)window)->x_window;
    XUnmapWindow(x_display, x_window);
}/*--- X_UnmapWindow --- */

void X_GetConfiguration(char *window, int *x, int *y, unsigned *w, unsigned *h){
    XWindowAttributes wa;
    Window x_window;
    x_window = ((XWindow*)window)->x_window;
    XGetWindowAttributes(x_display, x_window, &wa);
    *x = wa.x;
    *y = wa.y;
    *w = wa.width;
    *h = wa.height;
}/*--- X_GetConfiguration ---*/


void X_MoveWindow(char *window, int x, int y){
    XMoveWindow(x_display, ((XWindow*)window)->x_window, x, y);
}/*--- X_MoveWindow ---*/

void X_ResizeWindow(char *window, unsigned w, unsigned h){
    XResizeWindow(x_display, ((XWindow*)window)->x_window, w, h);
}/*--- X_ResizeWindow ---*/

void X_MoveResizeWindow(char *window, int x, int y, unsigned w, unsigned h){
    XMoveResizeWindow(x_display, ((XWindow*)window)->x_window, x, y, w, h);
}/*--- X_MoveResizeWindow ---*/

void X_SetBorder(char *window, int bw){
    XWindowChanges wc;
    wc.border_width = bw;
    XConfigureWindow(x_display, ((XWindow*)window)->x_window, CWBorderWidth, &wc);
}/*--- X_SetBorder ---*/

void X_SetSubwindowMode(char *window, int wm){
    GC  x_gc=((XWindow*)window)->x_gc;
    XSetSubwindowMode(x_display, x_gc, (wm)?ClipByChildren:IncludeInferiors);
}/*--- X_SetSubWindowMode ---*/

void X_SetCursor(char *window, int cursor){
    XWindow *Xw;
    Xw = (XWindow*)window;
    switch (cursor){
        case Arrow: Xw->x_cursor=x_arrow;	break;
        case Cross: Xw->x_cursor=x_cross;	break;
        case Hand1: Xw->x_cursor=x_hand1;	break;
        case Pencil:Xw->x_cursor=x_pencil;	break;
        case Watch: Xw->x_cursor=x_watch;	break;
        default: Xw->x_cursor=None;
    }
    if (Xw->x_cursor!=None)
        XDefineCursor(x_display, Xw->x_window, Xw->x_cursor);
    else
        XUndefineCursor(x_display, Xw->x_window);
}/*--- X_SetCursor ---*/

/*--------------------------------------*/
/*	Event Handling Functions	*/
/*--------------------------------------*/
int X_NextEvent(int *x, int *y, int *b, int *k, unsigned *t) {
    int event=NoEventE, state=0;
    *t=(unsigned)clock();
    if (XPending(x_display)==0) return NoEventE;
    XNextEvent(x_display, &x_event);
    switch (x_event.type) {
        case ButtonPress:
            event = ActivatedE;
            *x	= x_event.xbutton.x;
            *y	= x_event.xbutton.y;
            state = x_event.xbutton.state;
            break;
        case ButtonRelease:
            event = ButtonReleaseE;
            *x	= x_event.xbutton.x;
            *y	= x_event.xbutton.y;
            state = x_event.xbutton.state;
            break;
        case KeyPress:
            event = KeyPressE;
            *k	= TheKey(&x_event.xkey);
            break;
        case KeyRelease:
            event = KeyReleaseE;
            *k	= TheKey(&x_event.xkey);
            break;
        case EnterNotify:
            event = EnterE;
            *x	= x_event.xcrossing.x;
            *y	= x_event.xcrossing.y;
            break;
        case LeaveNotify:
            event = LeaveE;
            *x	= x_event.xcrossing.x;
            *y	= x_event.xcrossing.y;
            break;
        case Expose:
            event = (x_event.xexpose.count==0)?ExposeE:NoEventE;
            break;
        case MapNotify:
            event = MapE;
            break;
        case UnmapNotify:
            event = UnmapE;
            break;
        case MotionNotify:
            event = ButtonPressE;
            *x	= x_event.xmotion.x;
            *y	= x_event.xmotion.y;
            state = x_event.xmotion.state;
            break;
        case ConfigureNotify:
            event = ConfigureE;
            break;
        case PropertyNotify:
            event = PropertyE;
            break;
        default:
            event = UnknownE;
    }
    if (state & Button1Mask)
        if ((state & Button2Mask) || (state & Button3Mask))
            *b = 2; /* Both */
        else
            *b = 0; /* Left */
    else
        if ((state & Button2Mask) || (state & Button3Mask))
            *b = 1; /* Right */
        else
            *b = 0; /* Left */
    return event;
}/*--- X_NextEvent ---*/

int X_IsEventWindow(char *window){
    return (x_event.xany.window==((XWindow*)window)->x_window)?1:0;
}/*--- X_IsEventWindow ---*/

int X_MouseAt(char *window, int *x, int *y){
    Window rw, cw;
    int    rx, ry;
    unsigned mask;
    return XQueryPointer(x_display, ((XWindow*)window)->x_window, &rw, &cw, &rx, &ry,
            x, y, &mask);
}/*--- X_MouseAt ---*/

int X_EventXY(int *x, int *y){
    switch (x_event.type){
        case ButtonPress:
        case ButtonRelease:
            *x = x_event.xbutton.x;
            *y = x_event.xbutton.y;
            return 1;
        case MotionNotify:
            *x = x_event.xmotion.x;
            *y = x_event.xmotion.y;
            return 1;
        case KeyPress:
        case KeyRelease:
            *x = x_event.xkey.x;
            *y = x_event.xkey.y;
            return 1;
        default:
            return 0;
    }
}/*--- X_EventXY ---*/

void X_GrabPointer(char *window, int usual){
    Window x_window=((XWindow*)window)->x_window;
    Bool state;
    unsigned int EM;
    state = (usual==0)?False:True;
    EM = (unsigned int)(ButtonPressMask|ButtonReleaseMask|ButtonMotionMask|\
            EnterWindowMask|LeaveWindowMask);
    XGrabPointer(x_display, x_window, state, EM, GrabModeAsync, GrabModeAsync,
            x_window, None, CurrentTime);
}/*--- X_GrabPoiner ---*/

void X_UngrabPointer(void){
    XUngrabPointer(x_display, CurrentTime);
}/*--- X_UngrabPointer ---*/

void X_RemoveExposeEvents(void){
    XSync(x_display, False);
    while (XCheckMaskEvent(x_display, ExposureMask, &x_event));
}/*--- X_RemoveExposeEvents ---*/

/*------------------------------*/
/*	Graphic Functions	*/
/*------------------------------*/
unsigned X_FontTextWidth(int f, const char *string){
    XFontStruct *x_font;
    switch (f){
        case CourierBold12: x_font = x_courier_bold12;	break;
        case CourierBold14: x_font = x_courier_bold14;	break;
        default:		x_font = x_courier_bold12;	break;
    }
    return XTextWidth(x_font, string, strlen(string));
}/*--- X_FontTextWidth --- */

unsigned X_TextWidth(char *window, const char *string){
    XFontStruct *x_font=((XWindow*)window)->x_font;
    return XTextWidth(x_font, string, strlen(string));
}/*--- X_TextWidth ---*/

unsigned X_FontTextHeight(int f, const char *string){
    XFontStruct *x_font;
    switch (f){
        case CourierBold12: x_font = x_courier_bold12;	break;
        case CourierBold14: x_font = x_courier_bold14;	break;
        default:		x_font = x_courier_bold12;	break;
    }
    return (x_font->ascent + x_font->descent);
}/*--- X_FontTextHeight --- */

unsigned X_TextHeight(char *window, const char *string){
    XFontStruct *x_font=((XWindow*)window)->x_font;
    return (x_font->ascent + x_font->descent);
}/*--- X_TextHeight ---*/

void X_SetFont(char *window, int font){
    XWindow	*Xw=((XWindow*)window);
    switch (font){
        case CourierBold12: Xw->x_font=x_courier_bold12;	break;
        case CourierBold14: Xw->x_font=x_courier_bold14;	break;
        default:		Xw->x_font=x_courier_bold12;	break;
    }
    XSetFont(x_display, Xw->x_gc, Xw->x_font->fid);
}/*--- X_SetFont ---*/

void X_SetFunction(char *window, int f){
    GC x_gc=((XWindow*)window)->x_gc;
    XSetFunction(x_display, x_gc, f);
}/*--- X_SetFunction ---*/

void X_SetLineAttributes(char *window, int lw, int ls){
    GC x_gc=((XWindow*)window)->x_gc;
    if (lw<0) lw=0;
    if (ls!=LineOnOffDash && ls!=LineDoubleDash) ls=LineSolid;
    XSetLineAttributes(x_display, x_gc, lw, ls, CapNotLast, JoinMiter);
}/*--- X_SetLineAttributes ---*/

void X_DrawHText(char *window, int x, int y, const char *string){
    Window	x_window=((XWindow*)window)->x_window;
    GC	x_gc	=((XWindow*)window)->x_gc;
    XFontStruct *x_font=((XWindow*)window)->x_font;
    y -= x_font->descent-1;
    XDrawString(x_display, x_window, x_gc, x, y, string, strlen(string));
}/*--- X_DrawHText ---*/

void X_DrawVText(char *window, int x, int y, const char *string){
    Window	x_window=((XWindow*)window)->x_window;
    GC	x_gc	=((XWindow*)window)->x_gc;
    XFontStruct *x_font=((XWindow*)window)->x_font;
    int	i, j, m, n, width, height;
    Pixmap	Text_Pm1, Text_Pm2;
    if (strlen(string)>0){
        width = X_TextWidth(window, string);
        height = X_TextHeight(window, string);
        Text_Pm1 = XCreatePixmap(x_display, x_window, width, height, x_display_depth);
        Text_Pm2 = XCreatePixmap(x_display, x_window, height, width, x_display_depth);
        X_SetForeground(window, x_background);
        XFillRectangle(x_display, Text_Pm1, x_gc, 0, 0, width, height);
        X_SetForeground(window, x_black);
        XDrawString(x_display, Text_Pm1, x_gc, 0, x_font->ascent,
                string, strlen(string));
        XSetGraphicsExposures(x_display, x_gc, False);
        for (i=0, n=width-1;i<width;i++, n--)
            for (j=0, m=0;j<height;j++, m++)
                XCopyArea(x_display, Text_Pm1, Text_Pm2, x_gc, i, j, 1, 1, m, n);
        y -= width;
        x += 2;
        XCopyArea(x_display, Text_Pm2, x_window, x_gc, 0, 0, height, width, x, y);
        XSetGraphicsExposures(x_display, x_gc, True);
        XFreePixmap(x_display, Text_Pm2);
        XFreePixmap(x_display, Text_Pm1);
        XFlush(x_display);
    }
}/*--- X_DrawVText ---*/

void X_SetBackground(char *window, int color){
    XSetWindowBackground(x_display, ((XWindow*)window)->x_window, x_pixel[color]);
    XSetBackground(x_display, ((XWindow*)window)->x_gc, x_pixel[color]);
}/*--- X_SetBackground ---*/

void X_SetForeground(char *window, int color){
    XSetForeground(x_display, ((XWindow*)window)->x_gc, x_pixel[color]);
}/*--- X_SetForeground ---*/

void X_ClearArea(char *window, int x1, int y1, int x2, int y2, int expo){
    XClearArea(x_display, ((XWindow*)window)->x_window, x1, y1, x2-x1+1,
            y2-y1+1, expo);
}/*--- X_ClearArea ---*/

void X_ClearWindow(char *window){
    XClearWindow(x_display, ((XWindow*)window)->x_window);
}/*--- X_ClearWindow ---*/

void X_Point(char *window, int x, int y){
    Window	x_window=((XWindow*)window)->x_window;
    GC	x_gc	=((XWindow*)window)->x_gc;
    /*
  XDrawPoint(x_display,x_window,x_gc,x,y);
     */
    XDrawLine(x_display, x_window, x_gc, x, y, x+1, y+1);
}/*--- X_Point ---*/

void X_Points(char *window, int n, int *xy){
    Window	x_window=((XWindow*)window)->x_window;
    GC	x_gc	=((XWindow*)window)->x_gc;
    /*
XPoint	*xp;
int     i,*p;
  xp = (XPoint*)malloc(sizeof(XPoint)*n);
  for (p=xy,i=0;i<n;i++){
    xp[i].x = *p++; xp[i].y = *p++;
  }
  XDrawPoints(x_display,x_window,x_gc,xp,n,CoordModeOrigin);
  free(xp);
     */
    int i, j, I;
    for (i=0, j=1, I=2*n;i<I;i+=2, j+=2)
        XDrawLine(x_display, x_window, x_gc, xy[i], xy[j], xy[i]+1, xy[j]+1);
}/*--- X_Points ---*/

void X_Line(char *window, int x1, int y1, int x2, int y2){
    Window	x_window=((XWindow*)window)->x_window;
    GC	x_gc	=((XWindow*)window)->x_gc;
    XDrawLine(x_display, x_window, x_gc, x1, y1, x2, y2);
}/*--- X_Line ---*/

void X_Lines(char *window, int n, int *xy){
    Window	x_window=((XWindow*)window)->x_window;
    GC	x_gc	=((XWindow*)window)->x_gc;
    XSegment *x_segments;
    int	i, j, jmax;
    x_segments = (XSegment*)malloc(sizeof(XSegment)*n);
    for(jmax=4*n, i=j=0;j<jmax;j+=4, i++){
        x_segments[i].x1=xy[j];    x_segments[i].y1=xy[j+1];
        x_segments[i].x2=xy[j+2];  x_segments[i].y2=xy[j+3];
    }
    XDrawSegments(x_display, x_window, x_gc, x_segments, n);
    free(x_segments);
}/*--- X_Lines ---*/

void X_Polygon(char *window, int n, int *xy){
    Window	x_window=((XWindow*)window)->x_window;
    GC	x_gc	=((XWindow*)window)->x_gc;
    /*
XPoint  *x_points;
int	i,j,jmax;
  x_points = (XPoint*)malloc(sizeof(XPoint)*n);
  for(jmax=2*n,i=j=0;j<jmax;j+=2,i++){
    x_points[i].x=xy[j];  x_points[i].y=xy[j+1];
  }
  XDrawLines(x_display,x_window,x_gc,x_points,n,CoordModeOrigin);
  free(x_points);
     */
    int k, K;
    K=2*n-2;
    for(k=0;k<K;k+=2)
        XDrawLine(x_display, x_window, x_gc, xy[k], xy[k+1], xy[k+2], xy[k+3]);
}/*--- X_Polygon ---*/

void X_FillPolygon(char *window, int n, int *xy){
    Window	x_window=((XWindow*)window)->x_window;
    GC	x_gc	=((XWindow*)window)->x_gc;
    XPoint  *x_points;
    int	i, j, jmax;
    x_points = (XPoint*)malloc(sizeof(XPoint)*n);
    for(jmax=2*n, i=j=0;j<jmax;j+=2, i++){
        x_points[i].x=xy[j];  x_points[i].y=xy[j+1];
    }
    XFillPolygon(x_display, x_window, x_gc, x_points, n, Complex, CoordModeOrigin);
    free(x_points);
}/*--- X_FillPolygon ---*/

void X_Rectangle(char *window, int x1, int y1, int x2, int y2){
    Window	x_window=((XWindow*)window)->x_window;
    GC	x_gc	=((XWindow*)window)->x_gc;
    int      x, y;
    unsigned w, h;
    x=min(x1, x2);  w=max(x1, x2)-x;
    y=min(y1, y2);  h=max(y1, y2)-y;
    XDrawRectangle(x_display, x_window, x_gc, x, y, w, h);
}/*--- X_Rectangle ---*/

void X_Rectangles(char *window, int n, int *xy){
    Window	x_window=((XWindow*)window)->x_window;
    GC	x_gc	=((XWindow*)window)->x_gc;
    XRectangle *x_rectangles;
    int	i, j, jmax, x, y;
    x_rectangles = (XRectangle*)malloc(sizeof(XRectangle)*n);
    for(jmax=4*n, i=j=0;j<jmax;j+=4, i++){
        x_rectangles[i].x   =x=min(xy[j],  xy[j+2]);
        x_rectangles[i].y   =y=min(xy[j+1], xy[j+3]);
        x_rectangles[i].width =max(xy[j],  xy[j+2])-x;
        x_rectangles[i].height=max(xy[j+1], xy[j+3])-y;
    }
    XDrawRectangles(x_display, x_window, x_gc, x_rectangles, n);
    free(x_rectangles);
}/*--- X_Rectangles ---*/

void X_FillRectangle(char *window, int x1, int y1, int x2, int y2){
    Window	x_window=((XWindow*)window)->x_window;
    GC	x_gc	=((XWindow*)window)->x_gc;
    int      x, y;
    unsigned w, h;
    x=min(x1, x2);  w=max(x1, x2)-x;
    y=min(y1, y2);  h=max(y1, y2)-y;
    XFillRectangle(x_display, x_window, x_gc, x, y, w, h);
}/*--- X_FillRectangle ---*/

void X_FillRectangles(char *window, int n, int *xy){
    Window	x_window=((XWindow*)window)->x_window;
    GC	x_gc	=((XWindow*)window)->x_gc;
    XRectangle *x_rectangles;
    int	i, j, jmax, x, y;
    x_rectangles = (XRectangle*)malloc(sizeof(XRectangle)*n);
    for(jmax=4*n, i=j=0;j<jmax;j+=4, i++){
        x_rectangles[i].x   =x=min(xy[j],  xy[j+2]);
        x_rectangles[i].y   =y=min(xy[j+1], xy[j+3]);
        x_rectangles[i].width =max(xy[j],  xy[j+2])-x;
        x_rectangles[i].height=max(xy[j+1], xy[j+3])-y;
    }
    XFillRectangles(x_display, x_window, x_gc, x_rectangles, n);
    free(x_rectangles);
}/*--- X_FillRectangles ---*/

void X_Ellipse(char *window, int x, int y, int xr, int yr){
    Window	x_window=((XWindow*)window)->x_window;
    GC	x_gc	=((XWindow*)window)->x_gc;
    XDrawArc(x_display, x_window, x_gc, x-xr, y-yr, 2*xr+1, 2*yr+1, 0, 360*64);
}/*--- X_Ellipse ---*/

void X_FillEllipse(char *window, int x, int y, int xr, int yr){
    Window	x_window=((XWindow*)window)->x_window;
    GC	x_gc	=((XWindow*)window)->x_gc;
    XFillArc(x_display, x_window, x_gc, x-xr, y-yr, 2*xr+1, 2*yr+1, 0, 360*64);
}/*--- X_FillEllipse ---*/

void  X_Surface(char *window, int nx, int ny, int *xy){
    Window	x_window=((XWindow*)window)->x_window;
    GC	x_gc	=((XWindow*)window)->x_gc, xp_gc;
    XGCValues xp_gc_values;
    Pixmap  x_pixmap;
    unsigned long xp_gc_valuemask;
    unsigned w, h;
    int i, k, K, *x, *y, *Y, *p, nx2=nx*2, ny2=ny*2;
    X_GetConfiguration(window, &i, &k, &w, &h);
    x_pixmap = XCreatePixmap(x_display, x_window, w, h, x_display_depth);
    xp_gc_valuemask = 0;
    xp_gc = XCreateGC(x_display, x_window, xp_gc_valuemask, &xp_gc_values);
    XSetForeground(x_display, xp_gc, x_pixel[x_background]);
    XFillRectangle(x_display, x_pixmap, xp_gc, 0, 0, w, h);
    XFreeGC(x_display, xp_gc);
    x = (int*)malloc(sizeof(int)*nx2);
    K = ny*2-2;
    for(y=xy, Y=y+nx2*ny;y<Y;y+=ny2)
        for(k=0;k<K;k+=2)
            XDrawLine(x_display, x_pixmap, x_gc, y[k], y[k+1], y[k+2], y[k+3]);
    K = nx*2-2;
    for(i=0;i<ny2;i+=2){
        for(p=xy+i, y=x, Y=y+nx2;y<Y;){
            *y++ = p[0];
            *y++ = p[1];
            p+=ny2;
        }
        for(k=0;k<K;k+=2)
            XDrawLine(x_display, x_pixmap, x_gc, x[k], x[k+1], x[k+2], x[k+3]);
    }
    XSetGraphicsExposures(x_display, x_gc, False);
    XCopyArea(x_display, x_pixmap, x_window, x_gc, 0, 0, w, h, 0, 0);
    XSetGraphicsExposures(x_display, x_gc, True);
    XFreePixmap(x_display, x_pixmap);
    XFlush(x_display);
    free(x);
}/*--- X_Surface ---*/

void  X_FillSurface(char *window, int nx, int ny, int *xy, int *yt,
        int fc, int dc){
    Window	x_window=((XWindow*)window)->x_window;
    GC	x_gc	=((XWindow*)window)->x_gc, xp_gc;
    XGCValues xp_gc_values;
    Pixmap  x_pixmap;
    unsigned long xp_gc_valuemask;
    unsigned w, h;
    XPoint  a[5];
    int  ii, jj, m, n, i, j, k, xct, yct, dx, dy, xfirst, ist, mst, xmax=nx-1, ymax=ny-1,
            ny2=ny*2;
    /*
//	z
//	|
//	|    y
//	|  1/------- 2
//	|  /	   /
//	| /	  /
//	|/	 /
//	|------->x
//     0	  3
     */
    X_GetConfiguration(window, &i, &k, &w, &h);
    x_pixmap = XCreatePixmap(x_display, x_window, w, h, x_display_depth);
    xp_gc_valuemask = 0;
    xp_gc = XCreateGC(x_display, x_window, xp_gc_valuemask, &xp_gc_values);
    XSetBackground(x_display, xp_gc, x_pixel[x_background]);
    XSetForeground(x_display, xp_gc, x_pixel[x_background]);
    XFillRectangle(x_display, x_pixmap, xp_gc, 0, 0, w, h);
    XFreeGC(x_display, xp_gc);
    if (!yt){
        xfirst=ist=mst=0;
        dx=dy=1;
    }
    else{
        float Y_MAX=max(max(max(yt[0], yt[1]), yt[2]), yt[3]);
        if (Y_MAX==yt[0]){
            ist=0;      mst=0;      dx= 1;  dy= 1;  xfirst=(yt[1]<yt[3])?1:0;
        }
        else if (Y_MAX==yt[1]){
            ist=0;      mst=ymax-1; dx= 1;  dy= -1; xfirst=(yt[0]<yt[2])?1:0;
        }
        else if (Y_MAX==yt[2]){
            ist=xmax-1; mst=ymax-1; dx= -1; dy= -1; xfirst=(yt[3]<yt[1])?1:0;
        }
        else /*Y_MAX==yt[3]*/{
            ist=xmax-1; mst=0;      dx= -1; dy= 1;  xfirst=(yt[2]<yt[0])?1:0;
        }
    }
    if (xfirst)
        for(m=mst, n=m+1, yct=0;yct<ymax;m+=dy, n+=dy, ++yct)
            for(i=ist, j=i+1, xct=0;xct<xmax;i+=dx, j+=dx, ++xct){
                ii  =i*ny2;    jj  = j*ny2;
                a[0].x=a[4].x=xy[k=ii+2*m];
                a[0].y=a[4].y=xy[++k];
                a[1].x=xy[k=jj+2*m];
                a[1].y=xy[++k];
                a[2].x=xy[k=jj+2*n];
                a[2].y=xy[++k];
                a[3].x=xy[k=ii+2*n];
                a[3].y=xy[++k];
                XSetForeground(x_display, x_gc, x_pixel[fc]);
                XFillPolygon(x_display, x_pixmap, x_gc, a, 5, Complex, CoordModeOrigin);
                XSetForeground(x_display, x_gc, x_pixel[dc]);
                XDrawLines(x_display, x_pixmap, x_gc, a, 5, CoordModeOrigin);
            }
    else
        for(i=ist, j=i+1, xct=0;xct<xmax;i+=dx, j+=dx, ++xct)
            for(m=mst, n=m+1, yct=0;yct<ymax;m+=dy, n+=dy, ++yct){
                ii  =i*ny2;    jj  = j*ny2;
                a[0].x=a[4].x=xy[k=ii+2*m];
                a[0].y=a[4].y=xy[++k];
                a[1].x=xy[k=jj+2*m];
                a[1].y=xy[++k];
                a[2].x=xy[k=jj+2*n];
                a[2].y=xy[++k];
                a[3].x=xy[k=ii+2*n];
                a[3].y=xy[++k];
                XSetForeground(x_display, x_gc, x_pixel[fc]);
                XFillPolygon(x_display, x_pixmap, x_gc, a, 5, Complex, CoordModeOrigin);
                XSetForeground(x_display, x_gc, x_pixel[dc]);
                XDrawLines(x_display, x_pixmap, x_gc, a, 5, CoordModeOrigin);
            }
    XSetGraphicsExposures(x_display, x_gc, False);
    XCopyArea(x_display, x_pixmap, x_window, x_gc, 0, 0, w, h, 0, 0);
    XSetGraphicsExposures(x_display, x_gc, True);
    XFreePixmap(x_display, x_pixmap);
    XFlush(x_display);
}/*--- X_FillSurface ---*/

void X_SaveImage(char *window, int x1, int y1, int x2, int y2){
    Window	x_window=((XWindow*)window)->x_window;
    GC	x_gc	=((XWindow*)window)->x_gc;
    Pixmap  x_pixmap;
    unsigned w, h;
    ((XWindow*)window)->x_pixmap_width  = w = x2-x1+2;
    ((XWindow*)window)->x_pixmap_height = h = y2-y1+2;
    ((XWindow*)window)->x_pixmap = x_pixmap = XCreatePixmap(x_display, x_window,
            w, h, x_display_depth);
    XCopyArea(x_display, x_window, x_pixmap, x_gc, x1, y1, w, h, 0, 0);
}/*--- X_SaveImage ---*/

void X_RestoreImage(char *window, int x, int y){
    Window	x_window=((XWindow*)window)->x_window;
    GC	x_gc	=((XWindow*)window)->x_gc;
    Pixmap  x_pixmap=((XWindow*)window)->x_pixmap;
    unsigned w = ((XWindow*)window)->x_pixmap_width,
            h = ((XWindow*)window)->x_pixmap_height;
    if (x_pixmap){
        XCopyArea(x_display, x_pixmap, x_window, x_gc, 0, 0, w, h, x, y);
        XFreePixmap(x_display, x_pixmap);
        ((XWindow*)window)->x_pixmap = 0;
        ((XWindow*)window)->x_pixmap_width  = 0;
        ((XWindow*)window)->x_pixmap_height = 0;
    }
}/*--- X_RestoreImage ---*/

void X_ShowColorRuler(char *window){
    Window	 x_window;
    GC	 x_gc;
    int	 i, j, ch, x, y, yp, y2;
    unsigned w, h;
    double	 dh;
    x_window = ((XWindow*)window)->x_window;
    x_gc     = ((XWindow*)window)->x_gc;
    X_GetConfiguration(window, &x, &y, &w, &h);
    x = 0;
    y = 0;
    dh = ((double)(h))/x_avilcells;
    for(y2=y+h, j=x_avilcells-1, i=0;i<x_avilcells;i++, j--){
        yp = y+(int)(dh*j);
        ch = y2-yp;
        XSetForeground(x_display, x_gc, x_pixel[i]);
        XFillRectangle(x_display, x_window, x_gc, x, yp, w, ch);
        y2 = yp;
    }
    XFlush(x_display);
}/*--- X_ShowColorRuler ---*/


/*------------------------------*/
/*	Directory Functions	*/
/*------------------------------*/
int Matched(char *filename, char *ext){
    char *dot;
    dot = strrchr(filename, '.');
    return (dot==0 || strcmp(dot+1, ext)!=0)?0:1;
}/*--- Matched ---*/

void Sort(char **list, int size){
    int  *ip, i, j;
    char **fp;
    ip = (int*)malloc(sizeof(int)*size);
    fp = (char**)malloc(sizeof(char*)*size);
    for (i=0;i<size;i++) fp[i]=list[i];
    for (i=0;i<size;i++)
        for (ip[i]=0, j=0;j<size;j++)
            if (strcmp(fp[i], fp[j])>0) ip[i]++;
    for (i=0;i<size;i++) list[ip[i]]=fp[i];
    free(fp);
    free(ip);
}/*--- Sort ---*/


int GetFileList(char *indir, char *ext, char ***filelist){
    DIR    *dirp;
    struct dirent *dp;
    int    i, numfiles;
    char   **fp, *dir;
    if (indir)
        dir=indir;
    else
        dir=".";
    dirp = opendir(dir);
    numfiles = 0;
    for (dp=readdir(dirp);dp;dp=readdir(dirp))
        if (dp && Matched(dp->d_name, ext)) numfiles++;
    closedir(dirp);
    if (numfiles){
        *filelist = fp = (char**)malloc(sizeof(char*)*(numfiles+1));
        dirp = opendir(dir);
        for (i=0, dp=readdir(dirp);dp;dp=readdir(dirp))
            if (dp && Matched(dp->d_name, ext)){
                fp[i] = (char*)malloc(strlen(dp->d_name)+1);
                strcpy(fp[i], dp->d_name);
                i++;
            }
        fp[numfiles]=NULL;
        Sort(fp, numfiles);
        closedir(dirp);
    }
    else
        *filelist=NULL;
    return numfiles;
}/*--- GetFileList ---*/


void FreeFileList(char ***filelist){
    int i;
    char **fp;
    fp = *filelist;
    if (fp){
        for(i=0;fp[i];i++) free(fp[i]);
        free(fp);
        *filelist=NULL;
    }
}/*--- FreeFileList ---*/


/*==========================================================================*/
/*									    */
/*	End of XWindow.c						    */
/*									    */
/*==========================================================================*/
